/* $Id: client.c,v 0.12 1997/08/28 14:35:52 yoichi v0_70 $
 *
 * dhcpcd - DHCP client daemon -
 * Copyright (C) 1996 - 1997 Yoichi Hariguchi <yoichi@fore.com>
 *
 * dhcpcd is an RFC2131 and RFC1541 compliant DHCP client daemon.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/utsname.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <linux/netdevice.h>
#include <linux/if_arp.h>
#include <linux/if_ether.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "if.h"
#include "dhcp.h"
#include "dhcp-options.h"
#include "socket-if.h"
#include "arp.h"
#include "error-handler.h"
#include "hostinfo.h"
#include "client.h"
#define max(a, b) (((a) > (b)) ? (a) : (b))

extern int BeRFC1541;		/* if 1, Be RFC154 compliant */
extern int changeIP;	        /* if 1, OK to accept new IP */
extern int zeroLeaseTimeOK;	/* if 1, 0 leasetime = infinite. */
extern int acceptBootpPackets;
extern unsigned long orig_ipaddr;
extern char *Hostname;		/* if !NULL, put hostname into DHCP msg */
extern int MSMode;		/* 1 for MS (null terminated) servers */

extern time_t boottime(time_t *);


int (*Fsm[MAX_STATES])();	/* finite state machine */
int CurrState;			/* current state */
int PrevState;			/* previous state */

dhcpOutMessage DhcpMsgSend;	/* DHCP message to send */
dhcpInMessage DhcpMsgRecv;	/* DHCP message received */

int Ssend;			/* socket fd for send */
int Srecv;			/* socket fd for receive */

time_t ReqSentTime;		/* time when DHCPREQUEST message is sent */
u_long SuggestLeaseTime = 0;	/* lease time suggested by the user */

/* DHCP information
*/
u_long ServerInaddr;		/* Server's IP address (network byte order) */
u_long LeaseTime;		/* lease time (network byte order) */
u_long RenewTime;		/* T1 time (network byte order) */
u_long RebindTime;		/* T2 time (network byte order) */

/* Set for all places where options can reside */
static u_char *OptPtr[N_SUPPORT_OPTIONS]; /* array of ptrs to DHCP option element */
int sizeOptBuf = N_SUPPORT_OPTIONS;
static u_char ClassId[MAXIDCHARS];	/* class identifier */
static u_char ClientId[MAXIDCHARS];	/* client identifier */

extern unsigned char Wyse_Options[];	/* List of 128-254 options */
extern int numWyse_Options;
extern int maxWyse_Options;


void
classIDsetup(char *id)
{
#ifdef WYSE
    FILE *fp = fopen("/proc/cpuinfo", "r");
    char buf[BUFSIZ];
    int cpu = 0, vendor = 0; /* 0 = neither, 1 = mars, 2 = leo */

    if ( id != NULL ) {
	strncpy(ClassId, id, sizeof(ClassId));
	return;
    }
    if (fp == NULL)
	logSysExit("classIDSetup (dhcpMsgInit)");
    while(fgets(buf, BUFSIZ, fp) != NULL)
    {
	if (strncmp(buf, "vendor_id", strlen("vendor_id")) == 0)
	{
	    if (strstr(buf, "AuthenticAMD"))
		vendor = 1;
	    else if (strstr(buf, "CyrixInstead"))
		vendor = 2;
	} else if (strncmp(buf, "cpu", strlen("cpu")) == 0)
	{
	    if (strstr(buf, "486"))
		cpu = 1;
	    else if (strstr(buf, "586"))
		cpu = 2;
	}
    }
    fclose(fp);
#ifdef COMPAQ
    sprintf(ClassId, "Compaq T1500");
#else
    if (cpu == 1 && vendor == 1)
	sprintf(ClassId, "Wyse Winterm 531x");
    else if (cpu == 2 && vendor == 2)
	sprintf(ClassId, "Wyse Winterm 535x");
    else
	sprintf(ClassId, "Wyse Winterm 5xxx");
#endif
#else
    struct utsname sname;

    if ( id != NULL ) {
	strncpy(ClassId, id, sizeof(ClassId));
	return;
    }
    /* setup default class identifier if id is NULL
     */
    if ( uname(&sname) < 0 ) {
	logSysExit("uname (dhcpMsgInit)");
    }
    sprintf(ClassId, "%s %s %s", sname.sysname, sname.release, sname.machine);
#endif
}

void
clientIDsetup(char *id, char *ifname)
{
    struct ifinfo ifinfo;
    u_char *s;
    int len;

    s = ClientId;
    *s++ = 61;			/* type value of client identifier */

    if ( id != NULL ) {
	len = strlen(id);
	if ( len > sizeof(ClientId) - 4 ) {
	    /* 4: code, len, type, EOS */
	    logQuit("clientIDsetup: too long client ID string");
	}
	*s++ = len + 1 + (MSMode ? 1 : 0);	/* 1 for the # field */
	*s++ = 0;				/* type: string */
	strcpy(s, id);
	return;
    }
    /* setup default client identifier if id is NULL */
    strcpy(ifinfo.ifname, ifname);
    getIfInfo(&ifinfo);
    *s++ = 7; /* length: 6 (MAC Addr) + 1 (# field) */
    *s++ = ARPHRD_ETHER;		/* type: Ethernet address */
    bcopy(ifinfo.haddr, s, sizeof(ifinfo.haddr));
}

void
dhcpMsgInit(u_char *ifname)
{
    struct sockaddr_in	addr;

#if 0
    /* open and bind socket for receiving
     */
    setSockAddrIn(htons(0), htonl(INADDR_ANY), &addr);
    openRecvSocket(&addr, &Srecv);
#endif

    /* open and bind socket for sending
     */
    setSockAddrIn(htons(DHCP_CLIENT_PORT), htonl(INADDR_ANY), &addr);
    openSendSocket(&addr, &Ssend);
    Srecv = Ssend;
}

void
dhcpClient()
{
    int next_state;

    Fsm[INIT_REBOOT]	= initReboot;
    Fsm[INIT]		= init;
    Fsm[REBOOTING]	= rebooting;
    Fsm[SELECTING]	= selecting;
    Fsm[REQUESTING]	= requesting;
    Fsm[BOUND]		= bound;
    Fsm[RENEWING]	= renewing;
    Fsm[REBINDING]	= rebinding;

    CurrState = PrevState = INIT_REBOOT;

    for(;;) {
	next_state = (*Fsm[CurrState])();
	if ( next_state == EXCEPTION )
	    errQuit("Exception occured in the fsm (dhcpClient)");
	PrevState = CurrState;
	CurrState  = next_state;
    }
}

int
initReboot()
{
    int			fd;
    int			nextState = REBOOTING;
    char		filename[IFNAMSIZ + 128];
    struct sockaddr_in	addr;

    bzero((char *)&addr, sizeof(addr));
    strcpy(filename, DHCP_CACHE_FILE);
    strcat(filename, Ifbuf.ifname);
    if ( (fd = open(filename, O_RDONLY)) < 0 ) {
	if (changeIP == 0)
	{
	    FILE *pfp;
	    char buf[128];
	    pfp = popen("/sbin/ifconfig eth0 | /sbin/getipaddr", "r");
	    if (pfp == NULL)
		return INIT;
	    if (fgets(buf, sizeof(buf), pfp) == NULL)
		return INIT;
	    pclose(pfp);
	    if (inet_aton(buf, &addr.sin_addr) == 0)
		return INIT;
	    orig_ipaddr = addr.sin_addr.s_addr;
	    goto gotipaddr;
	}
	return INIT;
    }
    /* try to open cache file
     */
    if (read(fd, (char*)&addr.sin_addr.s_addr,sizeof(addr.sin_addr.s_addr))<0)
	nextState = INIT;
    orig_ipaddr = addr.sin_addr.s_addr;
    close(fd);
    unlink(filename);
gotipaddr:
    if ( nextState != REBOOTING )
	return nextState;
    /* found cache file */
    mkDhcpDiscoverMsg(Ifbuf.haddr, &DhcpMsgSend); /* set up MAC adddr, etc. */
    /* TODO: cache file should have the lease time previously used */
    mkDhcpRequestMsg(REBOOTING, 0, 0, rand(), orig_ipaddr, &DhcpMsgSend);
    setSockAddrIn(htons(DHCP_SERVER_PORT), htonl(INADDR_BROADCAST), &addr);
    boottime(&ReqSentTime);
#ifdef NEED_BCAST_RESPONSE
    DhcpMsgSend.flags = htons(F_BROADCAST);
#endif
    if (DebugFlag)
	printf("About to send size %d\n", sizeof(DhcpMsgSend));
    if ( sendto(Ssend, (char *)&DhcpMsgSend, sizeof(DhcpMsgSend), 0,
			    (struct sockaddr *)&addr, sizeof(addr)) < 0 ) {
	logSysExit("sendto (initReboot)");
    }
    return REBOOTING;
}

int
init()
{
    char ifname[IFNAMSIZ];
    int waitTime = 0;
    struct sockaddr_in	addr;

    if ( PrevState == RENEWING || PrevState == REBINDING ) {
	strcpy(ifname, Ifbuf.ifname);
	ifReset(ifname);
	dhcpMsgInit(Ifbuf.ifname);
    }
    mkDhcpDiscoverMsg(Ifbuf.haddr, &DhcpMsgSend);
    setSockAddrIn(htons(DHCP_SERVER_PORT), htonl(INADDR_BROADCAST), &addr);

    /* Note, the RFC says 1 - 10 secs to desync the use of DHCP at startup.
     * Booting a terminal takes > 10 seconds total to this point, and is
     * not exactly the same amount of time in any case.  Also, we are
     * using a timer with more resolution than seconds, so a smaller
     * range should be acceptible.
     */
    waitTime = random() % (3*1000); /* 0 - 3 seconds in millisecs */
    usleep(waitTime * 1000); /* millisecs to microsecs */
    boottime(&ReqSentTime);	/* record local time
			         * not @ REQUEST but @ DISCOVER (RFC1541) */
    if (DebugFlag)
	printf("About to send size %d\n", sizeof(DhcpMsgSend));
    if ( sendto(Ssend, (char *)&DhcpMsgSend, sizeof(DhcpMsgSend), 0,
			    (struct sockaddr *)&addr, sizeof(addr)) < 0 ) {
	logSysExit("sendto (init)");
    }
    return SELECTING;
}

int
rebooting()
{
    int			 nextState;
    struct sockaddr_in	 addr;
    u_int		 waitMsgType;

    setSockAddrIn(htons(DHCP_SERVER_PORT), htonl(INADDR_BROADCAST), &addr);
    waitMsgType = 0;
    setWaitMsgType(DHCP_ACK, &waitMsgType);
    setWaitMsgType(DHCP_NAK, &waitMsgType);

    /* DhcpMsgSend should contain DHCPREQUEST message (created in initReboot).
     */
    if ( !waitChkReXmitMsg(Srecv, &DhcpMsgRecv, Ssend, &DhcpMsgSend,
			   &addr, waitMsgType, OptPtr, N_REXMIT_REQ_REBOOT) ) {
	logRet("REBOOTING: timeout. Fall back to INIT");
	return INIT;
    }
    nextState = setDhcpInfo(OptPtr, &DhcpMsgRecv);
    if ( nextState == BOUND ) {
	/* check if yiaddr is already used */
	if ( !arpCheck(DhcpMsgRecv.yiaddr, &Ifbuf, ARP_REPLY_TIMEOUT) ) {
	    /* Reworked to allow bootstrap to provide data in a network boot environment */
	    unsigned long servaddr = ServerInaddr;
	    if (ServerInaddr == 0 && OptPtr[OserverInaddr] != NULL)
		    servaddr = *((u_int *)(OptPtr[OserverInaddr]+1));
	    sendDhcpDecline(DHCP_DECLINE, servaddr,
			DhcpMsgRecv.yiaddr);
	    sleep(10);
	    logRet("REBOOTING: %s is already used. Fall back to INIT",
		   inet_ntoa(*((struct in_addr *)&DhcpMsgRecv.yiaddr)));
	    return INIT;
	}
	/* now the client is initialized */

	/* initHost uses the values of Ifbuf.bcast and Ifbuf.mask as is
	 * under the following condition:
	 *   1. the DHCP ACK msg does not includes bcast or subnetmask option
	 *   2. Ifbuf.bcast or Ifbuf.mask is not 0.
	 * So we have to set these values to 0 not to confuse initHost.
	 */
	Ifbuf.bcast = Ifbuf.mask = 0;
	initHost(&Ifbuf, DhcpMsgRecv.yiaddr);
	execCommandFile();
	if ( LeaseTime == INFINITE_LEASE_TIME ) {
	    logRet("got the INFINITE lease time. exit(0).");
	    exit(0);
	}
    } else if ( nextState == INIT ) {
	logRet("REBOOTING: got DHCPNAK. Fall back to INIT");
    }
    return nextState;
}

static int
ConfigureHostFromResponse(void) {
    /* check if yiaddr is already used */
    Ifbuf.addr = htonl(0);	/* sender's IP address must be 0 */
    if ( !arpCheck(DhcpMsgRecv.yiaddr, &Ifbuf, ARP_REPLY_TIMEOUT) ) {
	/* Reworked to allow bootstrap to provide data in a network boot environment */
	unsigned long servaddr = ServerInaddr;
	if (ServerInaddr == 0 && OptPtr[OserverInaddr] != NULL)
		servaddr = *((u_int *)(OptPtr[OserverInaddr]+1));
	sendDhcpDecline(DHCP_DECLINE, servaddr,
			DhcpMsgRecv.yiaddr);
	logRet("REQUESTING: %s is already used. Fall back to INIT",
	       inet_ntoa(*((struct in_addr *) &DhcpMsgRecv.yiaddr)));
	sleep(10);
	return INIT;
    }
    /* now the client is initialized */
    initHost(&Ifbuf, DhcpMsgRecv.yiaddr);
    sendArpReply(MAC_BCAST_ADDR, Ifbuf.bcast, &Ifbuf);
    execCommandFile();
    if ( LeaseTime == INFINITE_LEASE_TIME ) {
	logRet("got the INFINITE lease time. exit(0).");
	exit(0);
    }
    return BOUND;
}

int
selecting()
{
    time_t		 prevTime;
    struct sockaddr_in	 addr;
    u_int		 waitMsgType;

    boottime(&prevTime);

    setSockAddrIn(htons(DHCP_SERVER_PORT), htonl(INADDR_BROADCAST), &addr);

    /* DhcpMsgSend shoud contain DHCPDISCOVER message (created in 'init') */
    waitMsgType = 0;
    setWaitMsgType(DHCP_OFFER, &waitMsgType);
    if (acceptBootpPackets)
	setWaitMsgType(0, &waitMsgType);
    if ( waitChkReXmitMsg(Srecv, &DhcpMsgRecv, Ssend, &DhcpMsgSend, &addr, waitMsgType, OptPtr, 0) ) /* 0 retries is infinite */
    {
	if (DhcpMsgRecv.yiaddr != orig_ipaddr && !changeIP)
	    return SELECTING;
	/* send DHCPREQUEST msessage */
	if (acceptBootpPackets &&
	    (*(OptPtr[OmsgType]+1) == 0 || OptPtr[OleaseTime] == NULL)) {
	      LeaseTime = INFINITE_LEASE_TIME;
	      ConfigureHostFromResponse();
	      /* If we got here, ConfiugreHostFromResponse failed. */
	      return SELECTING;
	}
	{
	    /* Reworked to allow bootstrap to provide data in a network boot environment */
	    if (ServerInaddr == 0 && OptPtr[OserverInaddr] != NULL)
		    ServerInaddr = *((u_int *)(OptPtr[OserverInaddr]+1));
	}
	if (OptPtr[OleaseTime] != NULL)
	    LeaseTime = *((u_int *)(OptPtr[OleaseTime]+1));
	if (LeaseTime == 0 && zeroLeaseTimeOK)
	  LeaseTime = INFINITE_LEASE_TIME;
	mkDhcpRequestMsg(SELECTING, ServerInaddr, LeaseTime, DhcpMsgRecv.xid,
			DhcpMsgRecv.yiaddr, &DhcpMsgSend);
#ifdef NEED_BCAST_RESPONSE
	DhcpMsgSend.flags = htons(F_BROADCAST);
#endif
	if (DebugFlag)
	    printf("About to send size %d\n", sizeof(DhcpMsgSend));
	if ( sendto(Ssend, (char *)&DhcpMsgSend, sizeof(DhcpMsgSend), 0,
		    (struct sockaddr *)&addr, sizeof(addr)) < 0 ) {
	    logSysExit("sendto (selecting)");
	}
	/* record option values in the DHCPOFFER message because NT server
	   does not include those values in the DHCPACK message
	 */
	setupIfInfo(&Ifbuf, DhcpMsgRecv.yiaddr, OptPtr);
	return REQUESTING;
    }
    ifDown(&Ifbuf);
    logQuit("no DHCPOFFER messages");
    return EXCEPTION;			/* should not happen */
}

int
requesting()
{
    struct sockaddr_in	 addr;
    int			 nextState;
    u_int		 waitMsgType;

    /* wait DHCPACK/DHCPNAK, rexmit RHCPREQUEST if necessary */
    setSockAddrIn(htons(DHCP_SERVER_PORT), htonl(INADDR_BROADCAST), &addr);
    waitMsgType = 0;
    setWaitMsgType(DHCP_ACK, &waitMsgType);
    setWaitMsgType(DHCP_NAK, &waitMsgType);
    if ( waitChkReXmitMsg(Srecv, &DhcpMsgRecv, Ssend, &DhcpMsgSend,
			  &addr, waitMsgType, OptPtr, N_REXMIT_REQUEST) ) {
	if ( (nextState = setDhcpInfo(OptPtr, &DhcpMsgRecv)) != EXCEPTION ) {
	    if ( nextState == BOUND ) {
		nextState = ConfigureHostFromResponse();
	    }
	    return nextState;
	}
	else
	{
	    logRet("requesting: Got inappropriate message back as DHCP response\n");
	}
    }
    logRet("no response to DHCPREEQUEST message. move to INIT state");
    return INIT;
}

int
bound()
{
    //logRet("got in BOUND state");
    sleep(ntohl(RenewTime));
    return RENEWING;
}

int
renewing()
{
    struct sockaddr_in	 addr;
    int			 nextState;
    time_t		 sendTime;
    time_t		 prevTime;
    long		 timeout;
    long		 tm;
    u_int		 waitMsgType;

    //logRet("got in RENEWING state");

    /* setup for sending unicast DHCPREQUEST message */
    setSockAddrIn(htons(DHCP_SERVER_PORT), ServerInaddr, &addr);
    if (DebugFlag)
	printf("Renew: Using addr of 0x%08.8lX\n", (long) ServerInaddr);
    mkDhcpRequestMsg(RENEWING, ServerInaddr, LeaseTime,
		     rand(), Ifbuf.addr, &DhcpMsgSend);
    /* send DHCPREQUESTvia unicast, and
     * wait server's response (DHCPACK/DHCPNAK)
     */
    timeout = ReqSentTime + ntohl(RebindTime) - boottime(NULL);
    /* timeout     = ntohl(RebindTime) - ntohl(RenewTime); */
    nextState   = REBINDING;
    waitMsgType = 0;
    setWaitMsgType(DHCP_ACK, &waitMsgType);	/* wait DHCPACK, DHCPNAK */
    setWaitMsgType(DHCP_NAK, &waitMsgType);
    tm = getNextTimeout(INIT_TIMEOUT);
    boottime(&prevTime);
    while ( timeout > 0 ) {
	/* send DHCPREQUEST via unicast */
	boottime(&sendTime);
	if (DebugFlag)
	    printf("About to send size %d\n", sizeof(DhcpMsgSend));
	if ( sendto(Ssend, (char *)&DhcpMsgSend, sizeof(DhcpMsgSend), 0,
			    (struct sockaddr *)&addr, sizeof(addr)) < 0 ) {
	    logSysRet("sendto (renewing) failure");
	    //sleep(tm/1000000);
	} else {
	    /* wait server's response */
	    if ( rcvAndCheckDhcpMsg(Srecv, &DhcpMsgRecv, waitMsgType, OptPtr,
				    tm) ) {
		nextState = setDhcpInfo(OptPtr, &DhcpMsgRecv);
		if ( nextState == EXCEPTION ) {
		    return nextState;
		}
		if ( nextState == INIT )
		    logSysRet("renew: got NAK, resetting to INIT state");
		break;
	    }
	}
	/* rcvAndCheckDhcpMsg will fail (with message) if Dhcp server
	 * service is shut down (NT) the receive from will fail, and we
	 * will immediately send out another request.  Instead, we
	 * should wait like normal to send it out again.  To do this,
	 * we need to ALWAYS do the sleep if we don't get a correct
	 * response (or an exception).
	 */
	sleep(tm/1000000);

	tm = getNextTimeout(NEXT_TIMEOUT); /* renew response timeout value */
	timeout -= boottime(NULL) - prevTime;
	boottime(&prevTime);
    }
    if ( nextState == BOUND ) {
#ifdef WYSE
	/* Wyse enhancement is to resave cache, /etc/dhcpc/hostinfo-*, and
	 * /etc/dhcpc/resolv.conf as appropriate because the information
	 * provided from the DHCP server MAY have changed.
	 */
	saveIfInfo(&Ifbuf);	/* save interface information onto file */
	saveHostInfo((const u_char **)OptPtr);
#endif /* WYSE */
	ReqSentTime = sendTime;		/* renew time when DHCPREQ is sent */
    } else if ( nextState == INIT ) {	/* got DHCPNAK */
	logRet("RENEWING: got DHCPNAK. Fall back to INIT");
	close(Ssend);
	close(Srecv);
	ifDown(&Ifbuf);
    }
    return nextState;
}

int
rebinding()
{
    struct sockaddr_in	 addr;
    int			 nextState;
    time_t		 sendTime;
    time_t		 prevTime;
    long		 timeout;
    long		 tm;
    u_int		 waitMsgType;

    //logRet("got in REBINDING state");

    nextState = INIT;			/* init nextState */

    /* setup for sending broadcast DHCPREQUEST message */
    setSockAddrIn(htons(DHCP_SERVER_PORT), Ifbuf.bcast, &addr);
    if (DebugFlag)
	printf("Rebind: Using addr of 0x%08.8lX\n", (long) Ifbuf.bcast);
    mkDhcpRequestMsg(REBINDING, ServerInaddr, LeaseTime,
		     rand(), Ifbuf.addr, &DhcpMsgSend);

    /* send DHCPREQUEST via broadcast, and
     * wait server's response (DHCPACK/DHCPNAK)
     */
    timeout = ReqSentTime + ntohl(LeaseTime) - boottime(NULL);
    /* timeout     = ntohl(LeaseTime) - ntohl(RebindTime); */
    waitMsgType = 0;
    setWaitMsgType(DHCP_ACK, &waitMsgType);	/* wait DHCPACK, DHCPNAK */
    setWaitMsgType(DHCP_NAK, &waitMsgType);
    tm = getNextTimeout(INIT_TIMEOUT);
    boottime(&prevTime);
    while ( timeout > 0) {
	boottime(&sendTime);
	if (DebugFlag)
	    printf("About to send size %d\n", sizeof(DhcpMsgSend));
	if ( sendto(Ssend, (char *)&DhcpMsgSend, sizeof(DhcpMsgSend), 0,
		    (struct sockaddr *)&addr, sizeof(addr)) < 0 ) {
	    logSysRet("sendto (rebinding) failure");
	    //sleep(tm/1000000);
	} else {
	    if ( rcvAndCheckDhcpMsg(Srecv, &DhcpMsgRecv,
				    waitMsgType, OptPtr, tm) ) {
		nextState = setDhcpInfo(OptPtr, &DhcpMsgRecv);
		if ( nextState == EXCEPTION ) {
		    return nextState;
		}
		if ( nextState == INIT )
		    logSysRet("rebind: got NAK, resetting to INIT state");
		break;
	    }
	}
	/* rcvAndCheckDhcpMsg will fail (with message) if Dhcp server
	 * service is shut down (NT) the receive from will fail, and we
	 * will immediately send out another request.  Instead, we
	 * should wait like normal to send it out again.  To do this,
	 * we need to ALWAYS do the sleep if we don't get a correct
	 * response (or an exception).
	 */
	sleep(tm/1000000);

	tm = getNextTimeout(NEXT_TIMEOUT);
	timeout -= boottime(NULL) - prevTime;
	boottime(&prevTime);
    }
    if ( nextState == BOUND ) {
#ifdef WYSE
	/* Wyse enhancement is to resave cache, /etc/dhcpc/hostinfo-*, and
	 * /etc/dhcpc/resolv.conf as appropriate because the information
	 * provided from the DHCP server MAY have changed.
	 */
	saveIfInfo(&Ifbuf);	/* save interface information onto file */
	saveHostInfo((const u_char **)OptPtr);
#endif /* WYSE */
	ReqSentTime = sendTime;	/* renew time when DHCPREQ is sent */
	return nextState;
    }
    /* Lease expired. halt network
     */
    logRet("REBINDING: Lease time expired. Fall back to INIT");
    close(Ssend);
    close(Srecv);
    ifDown(&Ifbuf);
    return INIT;
}

void
sub_mkDhcpMsg(u_char *p)
{
    *p++ = dhcpParamRequest;	/* Parameter Request List */
    *p++ = 19 + numWyse_Options;		/* number of requests */
    *p++ = subnetMask;
    *p++ = routersOnSubnet;
    *p++ = dns;
    *p++ = domainName;
    *p++ = broadcastAddr;
    *p++ = timeServer;
    *p++ = timerOffset; /* 7 */

    *p++ = dhcpT1value;
    *p++ = dhcpT2value;

    *p++ = dhcpOptionOverload; /* 10 */
    *p++ = extentionsPath;
    *p++ = vendorSpecificInfo;
    *p++ = xFontServer;
    *p++ = xDisplayManager;

    /* Not used yet */
    *p++ = nisDomainName; /* 15 */
    *p++ = nisServers;
    *p++ = lprServer;
    *p++ = SMTPServer;
    *p++ = POP3Server; /* 19 */
    {	/* Add all site options to the list */
	int val;
	for (val = 0; val < maxWyse_Options; val++)
	    if (Wyse_Options[val])
		*p++ = val + 128;
    }

    if ( Hostname != NULL ) {
	int len;

	len = strlen(Hostname); /* +1 to force NULL (for now) */
	*p++ = hostName;
	if (MSMode)
	    *p++ = len+1;
	else
	    *p++ = len;
	strncpy(p, Hostname, len);
	p += len;
	if (MSMode) *p++ = '\0';
    }
    if (ClassId[0] != '\0')
    {
	*p++ = dhcpClassIdentifier;	/* class identifier */
	if (MSMode)
	    *p++ = strlen(ClassId)+1;
	else
	    *p++ = strlen(ClassId);
	strcpy(p, ClassId);
	p += strlen(ClassId);
	if (MSMode) *p++ = '\0';
    }
    bcopy(ClientId, p, ClientId[1]+2); /* client identifier */
    p += ClientId[1] + 2;
    *p = endOption;				/* end */
}

void
mkDhcpDiscoverMsg(u_char *haddr, dhcpOutMessage *msg)
{
    u_char *p =	msg->options + 4; /* just after the magic cookie */

    bzero((char *)msg, sizeof(*msg));
    msg->htype = HTYPE_ETHER;	/* supports Etherenet only */
    msg->hlen  = 6;
    bcopy(haddr, msg->chaddr, 6);
    msg->op	   = BOOTREQUEST;
#ifdef NEED_BCAST_RESPONSE
    msg->flags = htons(F_BROADCAST);
#endif
    msg->xid   = htonl(random());

    /* make DHCP message option field
     */
    *((u_long *)msg->options) = htonl(MAGIC_COOKIE);
    *p++ = dhcpMessageType;		/* DHCP message type */
    *p++ = 1;
    *p++ = DHCP_DISCOVER;
    *p++ = dhcpMaxMsgSize;		/* Maximum DHCP message size */
    *p++ = 2;
    *((u_short *)p) = htons(INMAXNOPT);
    p += sizeof(u_short);
    if ( SuggestLeaseTime ) {
	*p++ = dhcpIPaddrLeaseTime;	/* IP address lease time */
	*p++ = 4;
	*((u_int *)p) = htonl(SuggestLeaseTime);
	p += sizeof(long);
    }
    sub_mkDhcpMsg(p);
}

void
mkDhcpRequestMsg(int flag, u_long serverInaddr, u_long leaseTime,
		 u_long xid, u_long ciaddr, dhcpOutMessage *msg)
{
    u_char *p =	msg->options + 4; /* just after the magic cookie */

    msg->xid	= xid;
    msg->ciaddr = ciaddr;
    if (flag == REBINDING && msg->giaddr == 0 && msg->ciaddr != 0)
	msg->flags = htons(0); /* do not set the broadcast flag here (RFC2131, 4.1) */
    else if (flag == REBINDING)	/* Rebinding should be broadcast, not unicast */
	msg->flags = htons(F_BROADCAST);
    else
	msg->flags = htons(0);	/* do not set the broadcast flag here */
    bzero((char *)p, sizeof(msg->options) - 4);	/* clear DHCP option field */

    /* 1. Requested IP address must not be in the DHCPREQUEST message
     *    under the RFC1541 mode.
     * 2. Requested IP address must be in the DHCPREQUEST message
     *    sent in the SELECTING or INIT-REBOOT state under the Internet
     *    Draft mode
     */
    if ( !BeRFC1541 ) {
	if ( flag == REBOOTING || flag == SELECTING ) {
	    /* ciaddr must be 0 in REBOOTING & SELECTING */
	    msg->ciaddr = htonl(0);
	    *p++ = dhcpRequestedIPaddr;
	    *p++ = 4;
	    *((u_int *)p) = ciaddr;
	    p += sizeof(u_int);
	}
    }
    *p++ = dhcpMessageType;				/* DHCP message type */
    *p++ = 1;
    *p++ = DHCP_REQUEST;
    *p++ = dhcpMaxMsgSize;				/* Maximum DHCP message size */
    *p++ = 2;
    *((u_short *)p) = htons(INMAXNOPT);
    p += sizeof(u_short);
    if ( flag == REBOOTING && SuggestLeaseTime ) {
	*p++ = dhcpIPaddrLeaseTime;	/* IP address lease time */
	*p++ = 4;
	*((u_int *)p) = htonl(SuggestLeaseTime);
	p += sizeof(long);
    }
    if ( flag == SELECTING || flag == REQUESTING ) {
	*p++ = dhcpServerIdentifier;	/* server identifier */
	*p++ = 4;
	*((u_int *)p) = serverInaddr;
	p += sizeof(u_int);
    }
    if ( leaseTime != 0 ) {
	*p++ = dhcpIPaddrLeaseTime;		/* IP address lease time */
	*p++ = 4;
	*((u_int *)p) = leaseTime;
	p += sizeof(u_int);
    }
    sub_mkDhcpMsg(p);
    if (DebugFlag)
    {
	if (flag == REBINDING || flag == RENEWING)
	{
	    printf("%s mode:\n", flag == REBINDING ? "Rebind" : "Renew");
	    printf("  op %d\n", msg->op);
	    printf("  htype %d\n", msg->htype);
	    printf("  hlen %d\n", msg->hlen);
	    printf("  hops %d\n", msg->hops);
	    printf("  xid 0x%08.8X\n", msg->xid);
	    printf("  secs %d\n", msg->secs);
	    printf("  flags 0x%08.8X\n", msg->flags);
	    printf("  ciaddr 0x%08.8X\n", msg->ciaddr);
	    printf("  yiaddr 0x%08.8X\n", msg->yiaddr);
	    printf("  siaddr 0x%08.8X\n", msg->siaddr);
	    printf("  giaddr 0x%08.8X\n", msg->giaddr);
	}
    }
}

void
mkDhcpDeclineMsg(int flag, u_long serverInaddr, u_long ciaddr,
		 dhcpOutMessage *msg)
{
    u_char *p =	msg->options + 4; /* just after the magic cookie */

    msg->xid	= rand();
    if ( flag == DHCP_RELEASE ) {	/* make a DHCPRELEASE msg */
	msg->ciaddr = ciaddr;
    } else {				/* make a DHCPDECLINE msg */
	if ( BeRFC1541 ) {	/* use ciaddr in RFC1541 compliant mode */
	    msg->ciaddr = ciaddr;
	} else {
	    msg->ciaddr = 0;
	}
    }

    bzero((char *)p, sizeof(msg->options) - 4);
    *p++ = dhcpMessageType;		/* DHCP message type */
    *p++ = 1;
    *p++ = (u_char)flag;
    *p++ = dhcpServerIdentifier; /* server identifier */
    *p++ = 4;
    *((u_int *)p) = serverInaddr;
    p += sizeof(long);
    if ( flag == DHCP_DECLINE && !BeRFC1541 ) {
	/* use the requested IP address option
	 * in the Internet Draft compliant mode
	 */
	*p++ = dhcpRequestedIPaddr;
	*p++ = 4;
	*((u_int *)p) = ciaddr;
	p += sizeof(u_int);
    }
#if 0
/* Code removed 7/13/99 because RFC2131 Page 37 says that "All Others"
 * MUST NOT be sent at DHCPRELEASE or DHCPDECLINE time.
 */
    if ( Hostname != NULL ) {
	int len;

	len = strlen(Hostname); /* +1 to force NULL (for now) */
	*p++ = hostName;
	if (MSMode)
	    *p++ = len+1;
	else
	    *p++ = len;
	strncpy(p, Hostname, len);
	p += len;
	if (MSMode) *p++ = '\0';
    }
#endif /* 0 */
    bcopy(ClientId, p, ClientId[1]+2); /* client identifier */
    p += ClientId[1] + 2;
    *p = endOption;				/* end */
}

void
sendDhcpDecline(int flag, u_long serverInaddr, u_long ciaddr)
{
    struct sockaddr_in	 addr;

    bzero((char *)&addr, sizeof(addr));
    addr.sin_family  = AF_INET;
    addr.sin_port = htons(DHCP_SERVER_PORT);
    switch ( flag ) {
      case DHCP_DECLINE:
	addr.sin_addr.s_addr = htonl(INADDR_BROADCAST);
	break;
      case DHCP_RELEASE:
	addr.sin_addr.s_addr = serverInaddr;
	break;
      default:
	logQuit("illegal flag value (sendDhcpDecline)");
	break;
    }
    mkDhcpDeclineMsg(flag, serverInaddr, ciaddr, &DhcpMsgSend);
    if (DebugFlag)
	printf("About to send size %d\n", sizeof(DhcpMsgSend));
    if ( sendto(Ssend, (char *)&DhcpMsgSend, sizeof(DhcpMsgSend), 0,
		(struct sockaddr *)&addr, sizeof(addr)) < 0 ) {
	logSysExit("sendto (sendDhcpDecline)");
    }
}

int
setDhcpInfo(u_char *optp[], dhcpInMessage *msg)
{
    switch ( *(optp[OmsgType]+1) ) {
      case DHCP_NAK:
	return INIT;
      case DHCP_ACK:
	if ( optp[OserverInaddr] != NULL )
	    ServerInaddr = *((u_long *)(optp[OserverInaddr]+1));
	if ( optp[OleaseTime] == NULL )
		LeaseTime = 0;
	else
		LeaseTime = *((u_long *)(optp[OleaseTime]+1));
	if (LeaseTime == 0 && zeroLeaseTimeOK)
	    LeaseTime = INFINITE_LEASE_TIME;
	if ( optp[OrenewalTime] == NULL ) {
	    RenewTime = htonl(ntohl(LeaseTime) / 2); /* default T1 time */
	} else {
	    RenewTime = *((u_long *)(optp[OrenewalTime]+1));
	}
	if ( optp[OrebindTime] == NULL ) {
	    RebindTime = htonl(ntohl(LeaseTime) / 8 * 7); /* default T2 time */
	} else {
	    RebindTime = *((u_long *)(optp[OrebindTime]+1));
	}

	/* Fix to handle HP/UX BOOTP server which gives a sixty second lease and
	 * 0 renewal time (but sets the T1 option).  If we detect this condition
	 * assume infinite lease time or 0 lease time.
	 */
	if (LeaseTime && optp[OrenewalTime] != NULL && RenewTime == 0)
	    LeaseTime = INFINITE_LEASE_TIME;
	return BOUND;
      default:
	return EXCEPTION;		/* should not happen */
    }
}

/* NOTE: this function is called from 'selecting()', and sets up the network
 *       interface information. It sets subnetmask and broadcast to 0 unless
 *       it receives these values. These values are set correctly in
 *       'initHost()'. This is for the following case: these values were in
 *       the DHCPOFFER message, and were not in the DHCPACK message. In this
 *       case, dhcpcd must not override these values with the 'natural'
 *       subenetmask and broadcast.
 */
void
setupIfInfo(struct ifinfo *ifbuf, u_long yiaddr, u_char *optp[])
{
    ifbuf->addr = yiaddr;
    if ( optp[Onetmask] != NULL ) {
	ifbuf->mask = *((u_int *)(optp[Onetmask]+1));
    } else {
	ifbuf->mask = 0;
    }
    if ( optp[ObcastInaddr] != NULL ) {
	ifbuf->bcast = *((u_int *)(optp[ObcastInaddr]+1));
    } else {
	/* Broadcast = netmask | (255.255.255.255 ^ subnet)
	 * netmask = subnet ^ IP;
	 */
	//ifbuf->bcast = 0;
	ifbuf->bcast = ifbuf->mask & ifbuf->addr;
	ifbuf->bcast = ifbuf->bcast | ((~0) ^ ifbuf->mask);
    }
}

void
initHost(struct ifinfo *ifbuf, u_long yiaddr)
{
    /* configure interface */
    ifbuf->addr  = yiaddr;
    logRet("assigned IP address:        %s",
	   inet_ntoa(*((struct in_addr *)&ifbuf->addr)));
    if ( OptPtr[Onetmask] == NULL ) {
	if ( ifbuf->mask == 0 ) {
	    /* if ifbuf->mask != 0, subnetmask info. was included in the
	     *  DHCPOFFER message and not included in the DHCPACK message.
	     *  In this case, subnetmask value must not been overwritten.
	     */
	    ifbuf->mask = getNaturalMask(ifbuf->addr);
	}
    } else {
	ifbuf->mask  = *((u_int *)(OptPtr[Onetmask]+1));
	logRet("assigned subnetmask:        %s",
	       inet_ntoa(*((struct in_addr *)&ifbuf->mask)));
    }
    if ( OptPtr[ObcastInaddr] == NULL ) {
	/* if the server responds only subnetmask, I should presume
	 * the broadcast address from the subnetmask instead of
	 * using the 'natural' broadcast address.
	 */
	if ( ifbuf->bcast == 0 ) {
	    /* if ifbuf->bcast != 0, broadcast info. was included in the
	     *  DHCPOFFER message and not included in the DHCPACK message.
	     *  In this case, broadcast addr. must not been overwritten.
	     */
	    ifbuf->bcast = (ifbuf->addr & ifbuf->mask) | ~ifbuf->mask;
	}
    } else {
	ifbuf->bcast = *((u_int *)(OptPtr[ObcastInaddr]+1));
	logRet("assigned broadcast address: %s",
	       inet_ntoa(*((struct in_addr *)&ifbuf->bcast)));
    }
    ifbuf->flags =
	IFF_UP | IFF_BROADCAST | IFF_NOTRAILERS | IFF_RUNNING | IFF_MULTICAST;
    ifConfig(ifbuf);

    saveIfInfo(ifbuf);		/* save interface information onto file */
    saveHostInfo((const u_char **)OptPtr);
}

long
getNextTimeout(int flag)
{
    static long  prevTimeout;
    long	 rv;			/* return value */

    if ( flag == INIT_TIMEOUT ) {
	prevTimeout = 4;		/* 4 seconds */
    } else if ( prevTimeout > 64 ) { /*New default set to 64 after we peg out*/
	//prevTimeout = 4;
	prevTimeout = 64;
    }
    rv = (prevTimeout - 1) * 1000000 + rand()/2000000;
    if (prevTimeout < 64)
	prevTimeout *= 2;
    return rv;
}
