/* $Id: dhcp-options.h,v 0.2 1997/08/30 18:00:29 yoichi v0_70 $
 *
 * dhcpcd - DHCP client daemon -
 * Copyright (C) 1996 - 1997 Yoichi Hariguchi <yoichi@fore.com>
 *
 * Dhcpcd is an RFC2131 and RFC1541 compliant DHCP client daemon.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* DHCP option type
 */
enum {
    OT_STRING = 1,	/* string */
    OT_ADDR   = 2,	/* addresses */
    OT_NUMBER = 3	/* 32 bit number */
};


/* DHCP option and value (cf. RFC1533)
 */
enum {
    padOption			=  0,
    endOption			= 255,
    subnetMask			=  1,
    timerOffset			=  2,
    routersOnSubnet		=  3,
    timeServer			=  4,
    nameServer			=  5,
    dns				=  6,
    logServer			=  7,
    cookieServer		=  8,
    lprServer			=  9,
    impressServer		= 10,
    resourceLocationServer	= 11,
    hostName			= 12,
    bootFileSize		= 13,
    meritDumpFile		= 14,
    domainName			= 15,
    swapServer			= 16,
    rootPath			= 17,
    extentionsPath		= 18,
    IPforwarding		= 19,
    nonLocalSourceRouting	= 20,
    policyFilter		= 21,
    maxDgramReasmSize		= 22,
    defaultIPTTL		= 23,
    pathMTUagingTimeout		= 24,
    pathMTUplateauTable		= 25,
    ifMTU			= 26,
    allSubnetsLocal		= 27,
    broadcastAddr		= 28,
    performMaskDiscovery	= 29,
    maskSupplier		= 30,
    performRouterDiscovery	= 31,
    routerSolicitationAddr	= 32,
    staticRoute			= 33,
    trailerEncapsulation	= 34,
    arpCacheTimeout		= 35,
    ethernetEncapsulation	= 36,
    tcpDefaultTTL		= 37,
    tcpKeepaliveInterval	= 38,
    tcpKeepaliveGarbage		= 39,
    nisDomainName		= 40,
    nisServers			= 41,
    ntpServers			= 42,
    vendorSpecificInfo		= 43, /* Format is described in enum below */
    netBIOSnameServer		= 44,
    netBIOSdgramDistServer	= 45,
    netBIOSnodeType		= 46,
    netBIOSscope		= 47,
    xFontServer			= 48,
    xDisplayManager		= 49,
    dhcpRequestedIPaddr		= 50,
    dhcpIPaddrLeaseTime		= 51,
    dhcpOptionOverload		= 52,
    dhcpMessageType		= 53,
    dhcpServerIdentifier	= 54,
    dhcpParamRequest		= 55,
    dhcpMsg			= 56,
    dhcpMaxMsgSize		= 57,
    dhcpT1value			= 58,
    dhcpT2value			= 59,
    dhcpClassIdentifier		= 60,
    dhcpClientIdentifier	= 61,

    /* New options 8/27/98, per RFC 2132 */
    nisplusDomain		= 64,
    nisplusServer		= 65,
    MobileIPHomeAgent		= 68,
    SMTPServer			= 69,
    POP3Server			= 70,
    NNTPServer			= 71,
    WWWServer			= 72,
    FingerServer		= 73,
    IRCServer			= 74,
    StreetTalkServer		= 75,
    STDAServer			= 76,
    TFTPServer			= 66,
    Bootfile			= 67
    /* User fields are 128-254 */
};

/* Wyse Vendor Options (option 43 or 128+) */
/* Refer to sub_getVendorOptions in options.c for conversion to normal (above)
 * option space.  Note that Q080_ is replaced with 0080_ at file save time.
 * At file save time, Q080_ strings will also need to be free'ed and NULLed.
 */
enum {
    WyseHeader		= 1,
    WyseNetSvc		= 2, /* 0080_NETSVC= */
    WyseNoSwap		= 3, /* 0080_NOSWAP= */
    WyseReflash		= 4, /* 0080_REFLASH= */
    WyseMfgDir		= 5, /* 0080_MF_DIR= */
    WyseMfgCfg		= 6, /* 0080_MF_CFG= */
    WyseXFS		= 7, /* 0080_XFS= */
    WyseXDMCP		= 8, /* 0080_XDMCP= */
    WyseBuddyBoot	= 9  /* 0080_BUDDY_BOOT= */
};


/* function prototypes
 */

void	getOptions(u_char *optp[], dhcpInMessage *msg);
/*  requires: 'optp' pointing to the area to which pointers to DHCP option
 *            fields are saved. 'optp' must have at least 312 elements.
 *            'msg' pointing to the DHCP message to be processed.
 *  effects:  it scans the option field of '*msg', and assigns pointers to
 *            the options in '*msg' to optp[i]. optp[] elements are not sorted.
 *  modifies: 'optp[i]'
 *  return:   Nothing
 */

int	parseDhcpMsg(u_char *optp[], dhcpInMessage *msg);
/*  requires: 'optp' pointing to the area to which pointers to DHCP option
 *            fields are saved. 'optp' must have at least 312 elements.
 *            'msg' pointing to the DHCP message to be processed.
 *  effects:  it checks '*msg' is a correct DHCP message or not. it also
 *            sorts 'optp[i]' according to 'Oxxx' (enum in dhcp.h)
 *  modifies: 'optp[i]'
 *  return:   1 if '*msg' is good, 0 if '*msg' is bad
 */
