/* $Id$
 *
 * dhcpcd - DHCP client daemon -
 * Copyright (C) 1996 - 1997 Yoichi Hariguchi <yoichi@fore.com>
 *
 * Dhcpcd is an RFC2131 and RFC1541 compliant DHCP client daemon.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <net/if.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include "dhcp.h"
#include "dhcp-options.h"
#include "error-handler.h"
#include "signal-handler.h"
#include "hostinfo.h"
#include "if.h"
#include "memory.h"

#define MAXLEN 256
#define max(a,b) ((a) > (b) ? (a) : (b))

extern char *CommandFile;
extern int sizeOptBuf;

extern void wyse_tftp(char *, char *, char *);
extern int sub_getOptions(u_char *optp[], u_char *, u_char *, int *, int);

#include "fix_hosts.h"

void
saveHostInfo(const u_char *optp[])
{
    int fd, i;
    char *appendfile = NULL;
    char path[MAXLEN];
    char buf[MAXLEN];
    extern int next_128plus_option;

    /* hostname
     */
    if ( optp[OhostName] != NULL ) {
	strncpy(path, optp[OhostName]+1, *optp[OhostName]);
	if ( sethostname(path, *optp[OhostName]) < 0 ) {
	    logSysRet("sethostname (setupHostInfo)");
	} else
	{
	    /* Fix /etc/hosts and /etc/HOSTNAME based upon new info */
	    fix_hosts(0);
	}
    }
    /* NIS domain name
     */
    if ( optp[OnisDomName] != NULL ) {
	strncpy(path, optp[OnisDomName]+1, *optp[OnisDomName]);
	if ( setdomainname(path, *optp[OnisDomName]) < 0 ) {
	    logSysRet("setdomainname (setupHostInfo)");
	}
    }
    /* default route (routers)
     */
    if ( optp[Orouter] != NULL ) {
	setDefRoute(optp[Orouter], &Ifbuf);
    }
    /* make directory for resolv.conf and  hostinfo file
     */
    if ( !setupHostInfoDir(HOST_INFO_DIR) ) {
	return;
    }
    /* ntp.conf
     */
    if ( optp[OntpServer] != NULL ) {
	mkNTPconf(optp[OntpServer]);
    }
    /* resolv.conf
     */
    if ( optp[Odns] != NULL || optp[OdomainName] != NULL ) {
	mkResolvConf(optp[Odns], optp[OdomainName]);
    }
    /* hostinfo file
     */
    strncpy(path, HOST_INFO_DIR, MAXLEN);
    strncat(path, "/", MAXLEN);
    strncat(path, HOST_INFO_FILE, MAXLEN);
    strncat(path, "-", MAXLEN);
    strncat(path, Ifbuf.ifname, MAXLEN);
    if ( (fd = creat(path,  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0 ) {
	logSysRet("creat (setupHostInfo)");
	return;
    }
    sprintf(buf, "LEASETIME=%ld\n", ntohl(LeaseTime));
    sprintf(buf+strlen(buf), "RENEWALTIME=%ld\n", ntohl(RenewTime));
    sprintf(buf+strlen(buf), "REBINDTIME=%ld\n", ntohl(RebindTime));
    sprintf(buf+strlen(buf), "IPADDR=%s\n",
	inet_ntoa(*((struct in_addr *)&Ifbuf.addr)));
    sprintf(buf+strlen(buf), "NETMASK=%s\n",
	inet_ntoa(*((struct in_addr *)&Ifbuf.mask)));
    sprintf(buf+strlen(buf), "BROADCAST=%s\n",
	inet_ntoa(*((struct in_addr *)&Ifbuf.bcast)));
    if ( write(fd, buf, strlen(buf)) < 0 ) {
	logSysRet("write (setupHostInfo)");
    }
    
    /* TFTP stuff -- requires routes, etc. to already be configured (above) */
    if ( optp[OextensionsPath] != NULL)
    {
	u_char *p = (u_char *) optp[OextensionsPath];
	u_char *optptr = (u_char *) optp[OextensionsPath];
	char *q, *r;

	optptr++;

	while(*p && *p != ':')
	    p++;
	if (*p != ':')
	{
	    logSysRet("Invalid format for extensionsPath option");
	    goto remainder;
	}
	q = strrchr(p, '/');
	*p++ = '\0';
	if (q == NULL)
	    q = p;
	else
	    q++;
	sprintf(path, "/tmp/%s", q);
	wyse_tftp(optptr, p, path);
	for (r = q; *r; r++)
	    if (strcasecmp(r, ".txt") == 0)
		break;
	if (strcasecmp(r, ".txt") == 0)
	{
	    appendfile = strdup(path);
	} else
	{
	    struct stat stbuf;
	    char *tbuf;
	    u_char **optbuf = NULL;
	    int fd2;

	    if (stat(path, &stbuf) < 0)
	    {
		logSysRet("Unable to read TFTP option set");
		goto remainder;
	    }
	    tbuf = malloc(stbuf.st_size);
	    if (tbuf == NULL)
		goto remainder;
	    if ( (fd2 = open(path,  O_RDONLY)) < 0 ) {
		logSysRet("read (setupHostInfo)");
		goto remainder;
	    }
	    read(fd2, tbuf, stbuf.st_size);
	    close(fd2);

	    /* Now, pull the options apart into the optp array */
	    optbuf = malloc(stbuf.st_size * sizeof(u_char *));
	    if (optbuf != NULL)
	    {
		int overload;
		u_char **p  = optbuf;
		bzero((u_char *)optbuf, stbuf.st_size * sizeof(u_char *));
		i = sub_getOptions(optbuf, tbuf, tbuf + stbuf.st_size, &overload, 0);
		/* At this point, optbuf contains all the options.  i is one past
		 * the last option.  Now perform the equivalent of the main loop
		 * from parseDhcpMsg, included below.
		 */
		for (; *p != NULL; ++p )
		{
		    switch (**p)
		    {
		      case dhcpMessageType:
			if (optp[OmsgType] == NULL)
			    optp[OmsgType] = *p + 1;
			break;
		      case dhcpServerIdentifier:
			if (optp[OserverInaddr] == NULL)
			    optp[OserverInaddr] = *p + 1;
			break;
		      case dhcpIPaddrLeaseTime:
			if (optp[OleaseTime] == NULL)
			    optp[OleaseTime] = *p + 1;
			break;
		      case dhcpT1value:
			if (optp[OrenewalTime] == NULL)
			    optp[OrenewalTime] = *p + 1;
			break;
		      case dhcpT2value:
			if (optp[OrebindTime] == NULL)
			    optp[OrebindTime] = *p + 1;
			break;
		      case subnetMask:
			if (optp[Onetmask] == NULL)
			    optp[Onetmask] = *p + 1;
			break;
		      case broadcastAddr:
			if (optp[ObcastInaddr] == NULL)
			    optp[ObcastInaddr] = *p + 1;
			break;
		      case timerOffset:
			if (optp[OtimerOffset] == NULL)
			    optp[OtimerOffset] = *p + 1;
			break;
		      case timeServer:
			if (optp[OtimeServer] == NULL)
			    optp[OtimeServer] = ( (*p)[1]%4 ) ? NULL : *p + 1;
			break;
		      case dns:
			if (optp[Odns] == NULL)
			    optp[Odns] = ( (*p)[1]%4 ) ? NULL : *p + 1;
			break;
		      case xFontServer:
			if (optp[OxFontServer] == NULL)
			    optp[OxFontServer] = ( (*p)[1]%4 ) ? NULL : *p + 1;
			break;
		      case xDisplayManager:
			if (optp[OxDisplayManager] == NULL)
			    optp[OxDisplayManager] = ( (*p)[1]%4 ) ? NULL : *p + 1;
			break;
		      case lprServer:
			if (optp[OlprServer] == NULL)
			    optp[OlprServer] = ( (*p)[1]%4 ) ? NULL : *p + 1;
			break;
		      case hostName:
			if (optp[OhostName] == NULL)
			    optp[OhostName] = *p + 1;
			break;
		      case domainName:
			if (optp[OdomainName] == NULL)
			    optp[OdomainName] = *p + 1;
			break;
		      case nisDomainName:
			if (optp[OnisDomName] == NULL)
			    optp[OnisDomName] = *p + 1;
			break;
		      case ntpServers:
			if (optp[OntpServer] == NULL)
			    optp[OntpServer] = ( (*p)[1]%4 ) ? NULL : *p + 1;
			break;
		      case dhcpMsg:
			strcpy(buf, "msg from the DHCP server: ");
			strncat(buf, *p + 2, (*p)[1]);
			logRet(buf);
			if (optp[OdhcpMessage] == NULL)
			    optp[OdhcpMessage] = *p + 1;
			break;
		      case dhcpClassIdentifier:
			if (optp[OdhcpClassId] == NULL)
			    optp[OdhcpClassId] = ( (*p)[1]%4 ) ? NULL : *p + 1;
			break;
		      case routersOnSubnet:
			if (optp[Orouter] == NULL)
			    optp[Orouter] = ( (*p)[1]%4 ) ? NULL : *p + 1;
			break;
		      case nisServers:
			if (optp[OnisServer] == NULL)
			    optp[OnisServer] = ( (*p)[1]%4 ) ? NULL : *p + 1;
			break;
		      case nisplusDomain:
			if (optp[OnisplusDomain] == NULL)
			    optp[OnisplusDomain] = *p + 1;
			break;
		      case nisplusServer:
			if (optp[OnisplusServer] == NULL)
			    optp[OnisplusServer] = ( (*p)[1]%4 ) ? NULL : *p + 1;
			break;
		      case SMTPServer:
			if (optp[OSMTPServer] == NULL)
			    optp[OSMTPServer] = ( (*p)[1]%4 ) ? NULL : *p + 1;
			break;
		      case POP3Server:
			if (optp[OPOP3Server] == NULL)
			    optp[OPOP3Server] = ( (*p)[1]%4 ) ? NULL : *p + 1;
			break;
		      default:
			/* 128-254 */
			if (**p >= 128 && **p <= 254)
			{
			    /* If it is a Wyse option, save it, otherwise ignore it */
			    if (strncmp("0080_", *p + 2, 5) == 0)
			    {
				optp[(next_128plus_option - 128) + OBaseUserOptions] = *p + 1;
				next_128plus_option++;
			    }
			}
			break;
		    }
		}
		free(optbuf);
	    }

	    /* Finally, free up resources */
	    free(tbuf);
	}
    }
remainder:

    if ( optp[OhostName] != NULL ) {
	addHostInfo(fd, OT_STRING, "HOSTNAME", optp[OhostName]);
    }
    if ( optp[OnisDomName] != NULL ) {
	addHostInfo(fd, OT_STRING, "NISDOMAINNAME", optp[OnisDomName]);
    }
    if ( optp[OlprServer] != NULL ) {
	addHostInfo(fd, OT_ADDR, "LPRSERVER", optp[OlprServer]);
    }
    if ( optp[OntpServer] != NULL ) {
	addHostInfo(fd, OT_ADDR, "NTPSERVER", optp[OntpServer]);
    }
    if ( optp[OtimeServer] != NULL ) {
	addHostInfo(fd, OT_ADDR, "TIMESERVER", optp[OtimeServer]);
    }
    if ( optp[Orouter] != NULL ) {
	addHostInfo(fd, OT_ADDR, "ROUTER", optp[Orouter]);
    }
    if ( optp[OnisServer] != NULL ) {
	addHostInfo(fd, OT_ADDR, "NISSERVER", optp[OnisServer]);
    }
    if ( optp[OnisplusDomain] != NULL ) {
	addHostInfo(fd, OT_STRING, "NISPLUSDOMAIN", optp[OnisplusDomain]);
    }
    if ( optp[OnisplusServer] != NULL ) {
	addHostInfo(fd, OT_ADDR, "NISPLUSSERVER", optp[OnisplusServer]);
    }
    if ( optp[OSMTPServer] != NULL ) {
	addHostInfo(fd, OT_ADDR, "SMTPSERVER", optp[OSMTPServer]);
    }
    if ( optp[OPOP3Server] != NULL ) {
	addHostInfo(fd, OT_ADDR, "POP3SERVER", optp[OPOP3Server]);
    }
    if ( optp[OserverInaddr] != NULL ) {
	addHostInfo(fd, OT_ADDR, "DHCPSERVER", optp[OserverInaddr]);
    }
    if ( optp[OxDisplayManager] != NULL ) {
	addHostInfo(fd, OT_ADDR, "XDMCP", optp[OxDisplayManager]);
    }
    if ( optp[OxFontServer] != NULL ) {
	addHostInfo(fd, OT_ADDR, "XFONTSERVER", optp[OxFontServer]);
    }
    if ( optp[OtimerOffset] != NULL ) {
	addHostInfo(fd, OT_NUMBER, "TIMEOFFSET", optp[OtimerOffset]);
    }
    /* Note that all Wyse specific tags are strings */
    for (i = OBaseUserOptions; i < N_SUPPORT_OPTIONS; i++)
    {
	if ( optp[i] != NULL )
	{
	    if (strncmp(&optp[i][1], "Q080_", 5) == 0)
	    {
		((u_char) (optp[i][1])) = '0';
		addHostInfo(fd, OT_STRING, NULL, optp[i]);
		free((void *) optp[i]); /* Note this was strdup'ed when this prefix is given */
		optp[i] = NULL;
	    } else
		if (strncmp(&optp[i][1], "0080_", 5) == 0)
			addHostInfo(fd, OT_STRING, NULL, optp[i]);
	}
    }
    if (appendfile)
    {
	/* Append the new file after the old one */
	int fd2, count;
	if ( (fd2 = open(appendfile,  O_RDONLY)) < 0 ) {
	    logSysRet("append (setupHostInfo)");
	    return;
	}
	while((count = read(fd2, path, sizeof(path))) > 0)
	{
	    write(fd, path, count);
	}
	close(fd2);
	unlink(appendfile);
	free(appendfile);
    }
    close(fd);
}


int
setupHostInfoDir(const char *dir)
{
    struct stat stbuf;

    if ( stat(dir, &stbuf) < 0 ) {
	if ( errno == ENOENT ) {
	    if ( mkdir(dir, S_IRUSR | S_IWUSR | S_IXUSR |
			   S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) < 0 ) {
		logSysRet("mkdir (setupHostInfoDir)");
		return 0;
	    }
	    return 1;
	} else {
	  logSysRet("stat (setupHostInfoDir)");
	  return 0;
	}
    }
    if ( !S_ISDIR(stbuf.st_mode) ) {
	chmod(dir, S_IRUSR | S_IWUSR);
	if ( unlink(dir) < 0 ) {
	    logSysRet("unlink (setupHostInfoDir)");
	    return 0;
	}
	if ( mkdir(dir, S_IRUSR | S_IWUSR | S_IXUSR |
		       S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) < 0 ) {
	    logSysRet("mkdir (setupHostInfoDir)");
	    return 0;
	}
    }
    if ( chmod(dir, S_IRUSR | S_IWUSR | S_IXUSR |
	       S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) < 0 ) {
	logSysRet("chmod (setupHostInfoDir)");
	return 0;
    }
    return 1;
}

void
addHostInfo(int fd, const int flag, const char *name, const u_char *optp)
{
    char    buf[MAXLEN];
    char    env[MAXLEN];
    int	    i;
    u_int  *p;

    switch ( flag ) {
      case OT_STRING:
	strncpy(env, optp+1, *optp);
	buf[0] = '\0';
	if (name != NULL)
	{
	    if ( setenv(name, env, 1) < 0 ) {
		logRet("setenv (addHostInfo): insufficient space");
	    }
	    strcat(buf, name);
	    strcat(buf, "=");
	} /* Otherwise, our tag and = is part of the option string */
	strncat(buf, optp+1, *optp);
	strcat(buf, "\n");
	break;
      case OT_ADDR:
	p = (u_int *)(optp + 1);
	if ( setenv(name, inet_ntoa(*((struct in_addr *)p)), 1) < 0 ) {
	    logRet("setenv (addHostInfo): insufficient space");
	}
	strcpy(buf, name);
	strcat(buf, "=");
	strcat(buf, inet_ntoa(*((struct in_addr *)p)));
	++p;
	strcat(buf, "\n");
	for ( i = 1; i < *optp/4; ++i ) {
	    sprintf(env, "%s%d", name, i+1);
	    if ( setenv(env, inet_ntoa(*((struct in_addr *)p)), 1) < 0 ) {
		logRet("setenv (addHostInfo): insufficient space");
	    }
	    sprintf(buf+strlen(buf), "%s%d=%s\n", name, i+1,
			    inet_ntoa(*((struct in_addr *)p)));
	    ++p;
	}
	break;
      case OT_NUMBER:
	p = (u_int *)(optp + 1);
	sprintf(env, "%ld", ntohl(*p));
	if ( setenv(name, env, 1) < 0 ) {
	    logRet("setenv (addHostInfo): insufficient space");
	}
	strcpy(buf, name);
	strcat(buf, "=");
	strcat(buf, env);
	++p;
	strcat(buf, "\n");
	break;
      default:
	return;
    }
    if ( write(fd, buf, strlen(buf)) < 0 ) {
	logSysRet("write (addHostInfo)");
    }
}

void
mkNTPconf(const u_char *addr)
{
    char buf[MAXLEN];
    int  fd;
    int  i;
    u_int *p;

    strcpy(buf, HOST_INFO_DIR);
    strcat(buf, "/");
    strcat(buf, "ntp.conf");
    if ( (fd = creat(buf,  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0 ) {
	logSysRet("creat (mkNTPconf)");
	return;
    }
    /* TODO: check whether NTP service is working on those addresses
     */
    p = (u_int *)(addr + 1);
    *buf = '\0';
    for ( i = 0; i < *addr/4; ++i ) {
	sprintf(buf+strlen(buf), "server %s\n",
		inet_ntoa(*((struct in_addr *)p)));
	++p;
    }
    if ( write(fd, buf, strlen(buf)) < 0 ) {
	logSysRet("write (mkNTPconf)");
    }
    close(fd);
}


void
mkResolvConf(const u_char *addr, const u_char *domName)
{
    char buf[MAXLEN];
    int  fd;
    int  i;
    u_int *p;

    strcpy(buf, HOST_INFO_DIR);
    strcat(buf, "/");
    strcat(buf, "resolv.conf");
    if ( (fd = creat(buf,  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0 ) {
	logSysRet("creat (mkResolvConf)");
	return;
    }
    /* TODO: check whether name server is working on those addresses
     */
    if (domName != NULL) {
	strcpy(buf, "domain ");
	strncat(buf, domName+1, *domName);
	strcat(buf, "\n");
    } else {
	buf[0] = '\0';
    }
    if (addr != NULL) {
	p = (u_int *)(addr + 1);
	for ( i = 0; i < *addr/4; ++i ) {
	    sprintf(buf+strlen(buf), "nameserver %s\n",
		    inet_ntoa(*((struct in_addr *)p)));
	    ++p;
	}
    }
    if ( write(fd, buf, strlen(buf)) < 0 ) {
	logSysRet("write (mkResolvConf)");
    }
    close(fd);
}

void
execCommandFile()
{
    pid_t pid;

    if ( CommandFile == NULL ) {
	return;
    }
    if ( (pid = fork()) < 0 ) {
	logSysRet("first fork (execCommandFile)");
    } else if ( pid == 0 ) {
	if ( (pid = fork()) < 0 ) {
	    logSysRet("second fork (execCommandFile)");
	} else if ( pid > 0 ) {
	    /* we're the 1st child. let's just exit */
	    exit(0);
	}
	/* we are the second child. let's exec the command file */
	if ( execlp(CommandFile, CommandFile, NULL) < 0 ) {
	    logSysRet("execlp (execCommandFile)");
	}
    }
    /* we're the parent (original process).
     * let's wait for the first child
     */
    if ( waitpid(pid, NULL, 0) != pid ) {
	logSysRet("waitpid (execCommandFile)");
    }
}
