/* $Id: main.c,v 0.10 1997/08/28 14:42:26 yoichi v0_70 $
 *
 * dhcpcd - DHCP client daemon -
 * Copyright (C) 1996 - 1997 Yoichi Hariguchi <yoichi@fore.com>
 *
 * dhcpcd is an RFC2131 and RFC1541 compliant DHCP client daemon.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <net/if.h>
#include <arpa/inet.h>
#include "if.h"
#include "dhcp.h"
#include "signal-handler.h"
#include "error-handler.h"
#include "daemon.h"
#include "client.h"
#include "memory.h"

#define DEFAULT_IF	"eth0"

char *CommandFile = NULL;	/* invoked command file name when dhcpcd
				 * succeeds in getting an IP address
				 */
int	  BeRFC1541 = 0;	/* default is InternetDraft mode */
int	  changeIP = 1;		/* OK to change IP address */
int	  zeroLeaseTimeOK = 0;	/* Inerpret zero lease time as infinite  */
unsigned long orig_ipaddr;
char *Hostname = NULL;		/* hostname in the DHCP msg for xmit */
int	  acceptBootpPackets = 0;
int MSMode = 0;			/* 1 will append NULL characters after string
				 * options on output to the server
				 */

static char VersionStr[] = "dhcpcd 0.70\n";
extern u_long ServerInaddr;

void	usage();

char *option_file = "/etc/dhcp-options"; /* Contains list of 128-254 to allow */
unsigned char Wyse_Options[254-128+1] = { 0 }; /* 128-254 */
int numWyse_Options = 0;
int maxWyse_Options = sizeof(Wyse_Options);

extern unsigned long getIPAddress(char *);

void
main(argc, argv)
int argc;
char *argv[];
{
    char  ifname[10];		/* interface name */
    char  pidfile[128];		/* file name in which pid is stored */
    char *clientId = NULL;	/* ptr to client identifier user specified */
    int   killFlag = 0;		/* if 1: kill the running proc and exit */
    FILE *fp;
    struct timeval tv;
    char *p;

    DebugFlag = 0;		/* default is NON debug mode */
    MSMode = 0;
    gettimeofday(&tv, NULL);
    tv.tv_usec /= 1000; /* convert to millisecs because of clock resolution */
    tv.tv_usec += (tv.tv_sec * 1000);
    srand((u_int)tv.tv_usec); /* more random than current time because no RTC */
    signalSetup();
    umask(0);			/* clear umask */
    classIDsetup(NULL);		/* setup default class identifier */

    if ((p = getenv("DHCPSERVER")) != NULL)
    {
    	ServerInaddr = inet_network(p);
	if (ServerInaddr == ~0)
	    ServerInaddr = 0;
    }

    /* option handling */
    while ( *++argv ) {
	if ( **argv == '-' ) {
	    switch ( argv[0][1] ) {
	      case 'a':
		    changeIP = 0;
		    break;
	      case 'b':
		    acceptBootpPackets = 1;
		    break;
	      case 'c':
		    if ( *++argv == NULL )
			usage();
		    if ( (CommandFile = malloc(strlen(*argv)+1)) == NULL )
			usage();
		    strcpy(CommandFile, *argv);
		    break;
	      case 'd':
		    DebugFlag = 1;
		    break;
	      case 'h':
		    if ( *++argv == NULL )
			usage();
		    Hostname = smalloc(strlen(*argv)+1);
		    strcpy(Hostname, *argv);
		    break;
	      case 'i':
		    if ( *++argv == NULL )
			usage();
		    classIDsetup(*argv); /* overwrite class identifier */
		    break;
	      case 'I':
		    if ( *++argv == NULL )
			usage();
		    clientId = *argv;
		    break;
	      case 'k':
		    killFlag = 1;	/* kill running process and exit */
		    break;
	      case 'l':
		    ++argv;
		    if ( *argv == NULL || **argv == '-' )
			usage();
		    SuggestLeaseTime = atol(*argv);
		    break;
	      case 'M': /* Microsoft complient for output */
		    MSMode = 1;
		    break;
	      case 'r':
		    BeRFC1541 = 1;	/* Be RFC1541 compliant */
		    break;
	      case 'v':
		    fflush(stdout);
		    fputs(VersionStr, stderr);
		    fflush(NULL);
		    exit(0);
	      case 'z':
		    zeroLeaseTimeOK = 1;
		    break;
	      case 'o':
		    if ( *++argv == NULL )
			usage();
		    option_file = *argv;
		    break;
	      default:
		    usage();
	    }
	} else {
	    break;
	}
    }

    if ( getuid() != 0 && geteuid() != 0 )
	errQuit("Must be root");

    if ((fp = fopen(option_file, "r")) == NULL)
    {
	/* Assume all are disabled */
	memset(Wyse_Options, 0, sizeof(Wyse_Options));
	numWyse_Options = 0;
    } else
    {
	char buf[BUFSIZ];
	int i;

	/* Assume the only ones enabled are the ones in the file. */
	memset(Wyse_Options, 0, sizeof(Wyse_Options));
	while(fgets(buf, BUFSIZ, fp) != NULL)
	{
	    i = atoi(buf);
	    if (i < 128 || i > 254)
		fprintf(stderr, "Illegal option '%d' stored in %s, skipping!\n", i, option_file);
	    else
		Wyse_Options[i - 128] = 1;
	}
	fclose(fp);

	/* Account for possible duplicates in file */
	numWyse_Options = 0;
	for (i = 0; i < sizeof(Wyse_Options); i++)
	    if (Wyse_Options[i])
		numWyse_Options++;
    }

    if ( *argv ) {
	strncpy(ifname, *argv, sizeof(ifname));
    } else {
	strncpy(ifname, DEFAULT_IF, sizeof(ifname));
    }
    if ( killFlag ) {
	sprintf(pidfile, PIDFILE, ifname);
	killCurProc(pidfile);
    }
    if ( !DebugFlag ) {
	sprintf(pidfile, PIDFILE, ifname);
	daemonInit(pidfile);
    }
    logOpen("dhcpcd", LOG_PID, LOG_LOCAL0);
    orig_ipaddr = getIPAddress(ifname);
    ifReset(ifname);		/* reset interface, 'Ifbuf' */
    clientIDsetup(clientId, ifname);
    {
	FILE *fp;
	char buf[100];
	sprintf(buf, PIDFILE, DEFAULT_IF);
	fp = fopen(buf, "w");
	fprintf(fp, "%d\n", getpid());
	fclose(fp);
    }
    dhcpMsgInit(ifname);
    dhcpClient();
}

void
usage()
{
    fflush(stdout);
    fputs("Usage: dhcpcd [-c filename] [-d] [-i classIdentifier] [-I clientIdentifier]\n    [-k] [-l leasetime] [-o optionfile] [ifname]\n",
	      stderr);
    fflush(NULL);
    exit(1);
}

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

time_t
boottime(time_t *t)
{
    int fd = open("/proc/uptime", O_RDONLY);
    char buf[64];
    time_t i;

    if (fd < 0)
    {
    	if (t != NULL)
	    *t = 0;
	return(0);
    }
    if (read(fd, buf, sizeof(buf)) < 0)
    {
	close(fd);
    	if (t != NULL)
	    *t = 0;
	return(0);
    }
    close(fd);
    i = atol(buf);
    if (t != NULL)
	*t = i;
    return(i);
}
