
#include "debugger.h"

#define	INT3	0xCC		/* this is a INT3 0xE instruction */

#define	MAXBP	12
#define	FOR_ALL_OF_bp_table for(ptr=bp_table; ptr < &(bp_table[MAXBP]); ptr++) 

#define	MAXIF	4

char *bp_has_cmd;
char *bp_go_cond;

struct	bptable	{
	uchar	*bp_addr;
	int	bp_count;
	int	bp_flag;
	uchar	bp_save;
	char	bp_cmd[128];
	char	bp_expr[128];
};

static	struct	bptable	bp_table[MAXBP];

struct	iftable {
	uint	if_addr;
	uchar	if_rel;
	uchar	if_len;
	uchar	if_cnt0;
	uchar	if_cnt1;
	uint	if_data;
	uint	if_mask;
};

static	struct	iftable if_table[MAXIF];

brcmd()
{
	char *cmd, *expr;
	extern int side_effects;

	if (getaddr(0)) {
		if (experr)
			return(-1);
		cmd = NULL;
		expr = NULL;
		skipsp();
		if (*cmdptr) {
			if (*cmdptr == 'i' && *(cmdptr+1) == 'f') {
				cmdptr += 2;
				skipsp();
				if (*cmdptr == '[' || *cmdptr == '\0') {
					printf("missing expression\n");
					return(-1);
				}
				expr = cmdptr;
				side_effects = 0;
				evalexp(-4);
				if (experr)
					return(-1);
			}
			skipsp();
			if (*cmdptr == '[') {
				*cmdptr++ = '\0';
				skipsp();
				cmd = cmdptr;
				while (*cmdptr) {
					if (*cmdptr == ']') {
						*cmdptr++ = '\0';
						break;
					}
					cmdptr++;
				}
				if (*cmd == '\0') {
					printf("missing command\n");
					return(-1);
				}
			}
			skipsp();
			if (*cmdptr) {
				printf("extra characters at end\n");
				return(-1);
			}
		}
		if (setbp(addr, remref, 1, cmd, expr) == -1)
			return(-1);
	} else {
		printbp();
	}
	return(0);
}

bccmd()
{
	if (getaddr(1)) {
		if (experr)
			return(-1);
		clrbp(addr);
	} else  {
		clrallbp();
	}
	return(0);
}

rbcmd()
{
	int	*p;

	p = (int *) dbr0ptr[EBP];
	setbp(p[1], 0, 0, 0, 0);
}

setbp(address, count, flag, cmd, expr)
	uchar	*address;
	int	count;
	int	flag;
	char	*cmd;
	char	*expr;
{
	register struct bptable *ptr;

	FOR_ALL_OF_bp_table {
		if(ptr->bp_addr == address) {
			if (flag == 0)
				flag = ptr->bp_flag;
			goto found;
		}
	}
	FOR_ALL_OF_bp_table {
		if(ptr->bp_addr == 0) 
			goto found;
	}
	if (flag)
		printf("BreakPoint table full\n");
	return(-1);
found:
	ptr->bp_addr = address;
	ptr->bp_count = count;
	ptr->bp_flag = flag;
	if (cmd == NULL)
		ptr->bp_cmd[0] = 0;
	else
		strcpy(ptr->bp_cmd,cmd);
	if (expr == NULL)
		ptr->bp_expr[0] = 0;
	else
		strcpy(ptr->bp_expr,expr);
	return(0);
}

clrbp(address)
uchar	*address;
{
	register struct bptable *ptr;

	FOR_ALL_OF_bp_table {
		if(ptr->bp_addr == address) {
			ptr->bp_addr = 0;
			return(0);
		}
	}
	return(-1);
}

static
clrallbp()
{
	register struct bptable *ptr;

	FOR_ALL_OF_bp_table {
		ptr->bp_addr = 0;
	}
}

brptin()
{
	register struct bptable *ptr;
	register int	eip;

	tocont();
	if(trcount) {
		trcount--;
		return;
	}
	dbflags &= ~DB_INSTEP;
	if (dbr0ptr[EFL] & PS_VM)		/* CS:IP is real mode addr */
		dbflags |= DB_REALMODE;
	eip = logtolin(dbr0ptr[EIP], dbr0ptr[CS]);
	FOR_ALL_OF_bp_table {
		if(ptr->bp_addr != 0) {
			if (ptr->bp_addr == (uchar *) eip) {
				if (ptr->bp_flag == 0)
					ptr->bp_addr = 0;
				else
					dbflags |= DB_INSTEP;
			} else {
				ptr->bp_save = ptr->bp_addr[0];
				ptr->bp_addr[0] = INT3;
			}
		}
	}
	if (dbflags & DB_INSTEP)
		((flags_t *)&dbr0ptr[EFL])->fl_tf = 1; /* turn on trace */
}

brptout()
{
	register struct bptable *ptr;
	int	ip, rval;

	rval = 1;

	if (dbwhy == 3) {
		if (dbr0ptr[EFL] & PS_VM)	/* CS:IP is real mode addr */
			dbflags |= DB_REALMODE;
		ip = logtolin(dbr0ptr[EIP], dbr0ptr[CS]) - 1;
	} else
		ip = 0;
	
	FOR_ALL_OF_bp_table {
		if (ptr->bp_addr != 0) {
			if(ptr->bp_addr[0] == INT3) {
				ptr->bp_addr[0] = ptr->bp_save;
			}
			if (ip == ptr->bp_addr) {
				dbr0ptr[EIP] -= 1;
				if (ptr->bp_flag == 0)
					ptr->bp_addr = 0;
				else
					rval = cond_bp(ptr);
			}
		}
	}
	return(rval);
}

cond_bp(ptr)
register struct bptable *ptr;
{
	register int rval;

	rval = 1;
	bp_has_cmd = (ptr->bp_cmd[0] ? ptr->bp_cmd : NULL);

	if (ptr->bp_expr[0]) {
		cmdptr = ptr->bp_expr;
		rval = evalexp(-1);
		if (experr) {
			rval = 1;
		} else if (rval) {
			if (ptr->bp_count > 0) {
				ptr->bp_count--;
				rval = 0;
			}
		}
		if (rval && !experr) {
			printf("Conditional breakpoint at %s: [%s] == 0x%x\n",
			spntadr(ptr->bp_addr),ptr->bp_expr,rval);
		}
	} else {
		/*
		** only has count
		*/
		if (ptr->bp_count > 0) {
			ptr->bp_count--;
			rval = 0;
		}
	}
	if (rval == 0)
		bp_has_cmd = NULL;
	return(rval);
}

static
printbp()
{
	register struct bptable *ptr;

	FOR_ALL_OF_bp_table {
		if(ptr->bp_addr != 0) {
			pntadr(ptr->bp_addr);
			if (ptr->bp_count > 0)
				printf(",%d", ptr->bp_count);
			if (ptr->bp_expr[0])
				printf(" if %s",ptr->bp_expr,ptr->bp_expr);
			if (ptr->bp_cmd[0])
				printf(" [%s]",ptr->bp_cmd,ptr->bp_cmd);
			printf("\n");
		}
	}
}

#define	IF_ACC	0
#define	IF_MOD	1
#define	IF_EQ	2
#define	IF_NE	3
#define	IF_GT	4
#define	IF_LT	5
#define	IF_GE	6
#define	IF_LE	7
#define IF_BIS	8
#define	IF_BIC	9

static
char	*if_rel[] = { "ACC", "MOD", "==", "!=", ">", "<", ">=", "<=", "BIS", "BIC" };

ifcmd()
{
	skipsp();
	if (*cmdptr == '?')
		help_if();
	else if (getaddr(0)) {
		if (experr)
			return(-1);
		if (setif(addr, remref) == -1)
			return(-1);
#ifdef	MP
		sendsoftall(DTBREAK);
#endif
	} else {
		printif();
	}
	return(0);
}

iccmd()
{
	if (getaddr(1)) {
		if (experr)
			return(-1);
		clrif(addr);
	} else  {
		clrallif();
	}
#ifdef	MP
	sendsoftall(DTBREAK);
#endif
	return(0);
}

static
printif()
{
	register struct	iftable 	*ifp;

	for (ifp = if_table; ifp < if_table + MAXIF; ++ifp) {
		if (ifp->if_len == 0) 
			continue;
		printf("IF %s,%d %s", spntadr(ifp->if_addr), ifp->if_len,
				      if_rel[ifp->if_rel]);
		if (ifp->if_rel > IF_MOD)
			printf(" 0x%x", ifp->if_data);
		if (ifp->if_cnt0)
			printf(" [%d]", ifp->if_cnt0);
		printf("\n");
	}
}

clrallif()
{
	register struct	iftable 	*ifp;

	for (ifp = if_table; ifp < if_table + MAXIF; ++ifp) 
		ifp->if_len = 0;
}

clrif(addr)
{
	register struct	iftable 	*ifp;

	for (ifp = if_table; ifp < if_table + MAXIF; ++ifp) {
		if (addr == ifp->if_addr)
			ifp->if_len = 0;
	}
}

static
setif(addr, len)
{
	int	rel, data, cnt;

	skipsp();
	if (*cmdptr == 0)  {
		help_if();
		return(-1);
	}

	switch(len) {

	case 0:
		len = 4;
	case 4:
		if (addr & 0x03) {
			printf("addr 0x%x not long word aligned\n",addr);
			return(-1);
		}
		break;
	case 2:
		if (addr & 0x01) {
			printf("addr 0x%x not short word aligned\n",addr);
			return(-1);
		}
	case 1:
		break;
	default:
		printf("illegal LEN option\n");
		return(-1);
	}

	rel = setrel();
	if (rel < 0)
		return(-1);

	if (rel > IF_MOD) {
		data = evalexp(0);
		if (experr)
			return(-1);
	} else {
		while(*cmdptr && *cmdptr != ',')
			++cmdptr;
		data = *(long *) addr;
	}

	if (len == 1)
		data &= 0xff;
	else
	if (len == 2)
		data &= 0xffff;

	skipsp();
	if(*cmdptr == ',') {
		cmdptr++;
		cnt = evalexp(1);
		if (experr)
			return(1);
	}
	else
		cnt = 0;
	databreak(addr, len, rel, data, cnt);
	return(0);
}

setdatabreak(addr, len, rel, data, cnt)
{
	clrif(addr);
	databreak(addr, len, rel, data, cnt);
	ifptout();
	ifptin();
#ifdef	MP
	sendsoftall(DTBREAK);
#endif
}

unsetdatabreak(addr)
{
	if (addr != -1)
		clrif(addr);
	else
		clrallif();
	ifptout();
	ifptin();
#ifdef	MP
	sendsoftall(DTBREAK);
#endif
}
	
databreak(addr, len, rel, data, cnt)
{
	register struct	iftable *ifp;
	static	 int	next;

	for (ifp = if_table; ifp < if_table + MAXIF; ++ifp) {
		if (ifp->if_len == 0)
			goto insert;
	}
	
	ifp = &if_table[next];
	if (++next >= MAXIF)
		next = 0;
insert:
	ifp->if_addr = addr;
	ifp->if_len  = len;
	ifp->if_rel  = rel;
	ifp->if_data = data;
	ifp->if_cnt0 = cnt;
	ifp->if_cnt1 = cnt;
}

static
setrel()
{
	register char *cp;

	skipsp();
	if (*cmdptr == 0) {
		help_if();
		return(-1);
	}
	cp = cmdptr;
	cmdptr += 3;
	if (strncmp(cp, "acc", 3) == 0) 
		return(IF_ACC);
	if (strncmp(cp, "ACC", 3) == 0) 
		return(IF_ACC);
	if (strncmp(cp, "mod", 3) == 0)
		return(IF_MOD);
	if (strncmp(cp, "MOD", 3) == 0)
		return(IF_MOD);
	if (strncmp(cp, "bis", 3) == 0)
		return(IF_BIS);
	if (strncmp(cp, "BIS", 3) == 0)
		return(IF_BIS);
	if (strncmp(cp, "bic", 3) == 0)
		return(IF_BIC);
	if (strncmp(cp, "BIC", 3) == 0)
		return(IF_BIC);
	cmdptr--;
	if (strncmp(cp, "==", 2) == 0)
		return(IF_EQ);
	if (strncmp(cp, "!=", 2) == 0)
		return(IF_NE);
	if (strncmp(cp, ">=", 2) == 0)
		return(IF_GE);
	if (strncmp(cp, "<=", 2) == 0)
		return(IF_LE);
	cmdptr--;
	if (*cp == '>')
		return(IF_GT);
	if (*cp == '<')
		return(IF_LT);
	cmdptr--;
	help_if();
	return(-1);
}

static
help_if()
{
	printf("DataBreak 'IF' command comes in 3 flavors\n\n");
	printf("IF address[,length] relation value\n");
	printf("   address: a static kernel text/data address\n");
	printf("   length: size of data to monitor(default=4)\n");
	printf("   relation: comparsion operator such as == != > >= etc\n");
	printf("   value: value to be compared\n\n");
	printf("IF address[,length] accessed\n\n");
	printf("IF address[,length] modified\n");
	printf("IF address[,length] bis bit_mask\n");
	printf("IF address[,length] bic bit_mask\n");
	return(-1);
}

#ifdef	__KERNEL__
#define	KERNEL_OFFSET	0xC0000000
#else
#define	KERNEL_OFFSET	0
#endif

ifptin()
{
	register struct	iftable 	*ifp;
	register int			dr7, i;
		 int	rel;

	dr7 = 0;
	for (i = 0, ifp = if_table; ifp < if_table + MAXIF; ++ifp, ++i) {
		if (ifp->if_len == 0)
			continue;
		if (ifp->if_rel == IF_ACC)
			rel = DR_RW_READ;
		else
			rel = DR_RW_WRITE;
		rel |= (ifp->if_len - 1) << 2;
		dr7 |= 3 << (i * 2);	/* set both global & local enable bit */
		dr7 |= rel << ((i * 4) + 16);
		if (i == 0)
			_wdr0(ifp->if_addr + KERNEL_OFFSET);
		else
		if (i == 1)
			_wdr1(ifp->if_addr + KERNEL_OFFSET);
		else
		if (i == 2)
			_wdr2(ifp->if_addr + KERNEL_OFFSET);
		else
		if (i == 3)
			_wdr3(ifp->if_addr + KERNEL_OFFSET);
	}
	if (dr7) {
		dr7 |= DR_LOCAL_SLOWDOWN | DR_GLOBAL_SLOWDOWN;
		_wdr7(dr7);
	}
}

ifptout()
{
	_wdr7(0);
}

ifreload()
{
	ifptout();	/* clear current databreak points */
	ifptin();	/* reload new databreak */
}

ifbrpt(dr6, r0ptr)
	register uint	dr6;
{
	register struct	iftable 	*ifp;
	register uint	i;
		 int	flag = 0;

	dr6 &= 0x0F;
	ifptout();
	while(dr6) {
		i = scanbit(dr6);
		dr6 -= (1 << i);
		ifp = if_table + i;
		if (ifp->if_len == 0)
			continue;
		
		i = *((uint *) ifp->if_addr);
		if (ifp->if_len == 1)
			i &= 0xff;
		else
		if (ifp->if_len == 2)
			i &= 0xffff;

		switch(ifp->if_rel) {

		case IF_ACC:
			flag += ifcnt(ifp, r0ptr);
			printf("Accessed from 0x%x to 0x%x\n", ifp->if_data, i);
			ifp->if_data = i;
			continue;
		case IF_MOD:
			flag += ifcnt(ifp, r0ptr);
			printf("Modified from 0x%x to 0x%x\n", ifp->if_data, i);
			ifp->if_data = i;
			continue;
		case IF_EQ:
			if (i == ifp->if_data)
				break;
			continue;
		case IF_NE:
			if (i != ifp->if_data)
				break;
			continue;
		case IF_GE:
			if (i >= ifp->if_data)
				break;
			continue;
		case IF_LE:
			if (i <= ifp->if_data)
				break;
			continue;
		case IF_GT:
			if (i > ifp->if_data)
				break;
			continue;
		case IF_LT:
			if (i < ifp->if_data)
				break;
			continue;
		case IF_BIS:
			if (i & ifp->if_data)
				break;	
			continue;
		case IF_BIC:
			if ((i & ifp->if_data) == 0)
				break;
			continue;
		}
		flag += ifcnt(ifp, r0ptr);
		printf("0x%x %s 0x%x\n", i, if_rel[ifp->if_rel], ifp->if_data);
	}
	if (flag == 0)
		ifptin();
	return(flag);
}

ifcnt(ifp, r0ptr)
	register struct	iftable 	*ifp;
	register int	*r0ptr;
{
	printf("DataBreak %s: ", spntadr(ifp->if_addr));
	if (ifp->if_cnt0) {
		printf("at %s: ", spntadr(r0ptr[EIP]));
		ifp->if_cnt1--;
		if (ifp->if_cnt1) 
			return(0);
		ifp->if_cnt1 = ifp->if_cnt0;
	}
	return(1);
}

