#include "debugger.h"


/*
** translate logical address to linear
** address if in user address space
*/
logtolin(offset,selector)
long offset;
int selector;
{
#if 0
	long base, linear;
	struct dscr *dscrp;

	if (offset >= KVBASE)
		return(offset);
	selector &= 0xffff;		/* selector is 16 bit long */
	if (dbflags & DB_REALMODE) {
		dbflags &= ~DB_REALMODE;
		return((selector << 4) + offset);
	}
	if (selector & 0x04)
		dscrp = (struct dscr *) u.u_procp->p_ldt + (selector >> 3);
	else
		dscrp = (struct dscr *) l.l_gdt + (selector >> 3);
	base = dscrp->a_base0015 & 0x0000ffff;
	base |= (dscrp->a_base1623 << 16) & 0x00ff0000;
	base |= (dscrp->a_base2431 << 24) & 0xff000000;
	linear = base + offset;
	return(linear);
#else
	return(offset);
#endif
}

/*
** tokens
*/
#define	EOL		0x00
#define	UNKNOWN		0x01
#define	ASSIGN		0x02
#define	INCR		0x03
#define	DECR		0x04

#define	LOROP		0x10
#define	LANDOP		0x11

#define	EQOP		0x20
#define	NEOP		0x21
#define	GTOP		0x22
#define	LTOP		0x23
#define	GEOP		0x24
#define	LEOP		0x25

/*
** the following two sets of operators
** must remain symmetric
**
** ie. (OROP-OREQ) == .. == (MULOP-MULEQ) == .. == (RSHOP-RSHEQ)
*/
#define	OREQ		0x30
#define	XOREQ		0x31
#define	ANDEQ		0x32
#define	ADDEQ		0x33
#define	SUBEQ		0x34
#define	MULEQ		0x35
#define	DIVEQ		0x36
#define	MODEQ		0x37
#define	LSHEQ		0x38
#define	RSHEQ		0x39

#define	OROP		0x40
#define	XOROP		0x41
#define	ANDOP		0x42
#define	ADDOP		0x43
#define	SUBOP		0x44
#define	MULOP		0x45
#define	DIVOP		0x46
#define	MODOP		0x47
#define	LSHOP		0x48
#define	RSHOP		0x49

#define	LPAREN		0x50
#define	RPAREN		0x51
#define	COMPOP		0x52
#define	NOTOP		0x53
#define	COMMA		0x54
#define	COLONOP		0x55
#define	PTROP		0x56

/*
** token classes
*/

#define	CONSTANT	0x60
#define	SYMBOL		0x61
#define	REGISTER	0x62
#define	VARIABLE	0x63
#define	MAXTOKENS	0x64

#define ISRELOP		0x01
#define	ISASSIGNOP	0x02

#define	isrelop(i)	(__optype[(i)] & ISRELOP)
#define	isassignop(i)	(__optype[(i)] & ISASSIGNOP)

unsigned char __optype[MAXTOKENS];

struct token {
	char which;
	unsigned char size;	/* size of REGISTER only */
	unsigned int value;
	unsigned int *addr;
	char *pos;
} token;

int		experr;
int		relops;
int		side_effects = 1;

#define	MAXNAMLEN	12
static  char	srch_name[MAXNAMLEN];

struct sym {
	char		name[MAXNAMLEN];
	unsigned int	value;
};

struct sym *srchsym();

#define	DBGSCNT	2500

int		dbgslow = 0;
int		dbgssize = DBGSCNT;
int		dbgscnt = DBGSCNT - 1;
struct sym	dbgstbl[DBGSCNT] = { "?", -1 };

#define	MAXALIASES	64
#define	MAXASPACE	4096
int naliases;
struct sym atable[MAXALIASES];
char abuffer[MAXASPACE];
char *aliasptr = abuffer;

#define MAXVARS	64
int nvars;
struct sym vtable[MAXVARS];

#define NREGS (sizeof(rtable)/sizeof(struct sym))
struct sym rtable[] = {
	{ "eax", EAX },
	{ "ebx", EBX },
	{ "ecx", ECX },
	{ "edx", EDX },
	{ "esi", ESI },
	{ "edi", EDI },
	{ "ebp", EBP },
	{ "esp", ESP },
	{ "eip", EIP },
#if 0
	{ "cs",  CS  },
	{ "ds",  DS  },
	{ "es",  ES  },
	{ "ss",  SS  },
#endif
	{ "EAX", EAX },
	{ "EBX", EBX },
	{ "ECX", ECX },
	{ "EDX", EDX },
	{ "ESI", ESI },
	{ "EDI", EDI },
	{ "EBP", EBP },
	{ "ESP", ESP },
	{ "EIP", EIP },
#if 0
	{ "CS",  CS  },
	{ "DS",  DS  },
	{ "ES",  ES  },
	{ "SS",  SS  },
#endif
	{ 0,     0 }, 
};

expr_init()
{
	dbgsyminit();
	dot = NULL;
	laddr.a_char = NULL;
	__optype[EQOP] = ISRELOP;
	__optype[NEOP] = ISRELOP;
	__optype[GTOP] = ISRELOP;
	__optype[LTOP] = ISRELOP;
	__optype[GEOP] = ISRELOP;
	__optype[LEOP] = ISRELOP;
	__optype[ASSIGN] = ISASSIGNOP;
	__optype[INCR] = ISASSIGNOP;
	__optype[DECR] = ISASSIGNOP;
	__optype[OREQ] = ISASSIGNOP;
	__optype[XOREQ] = ISASSIGNOP;
	__optype[ANDEQ] = ISASSIGNOP;
	__optype[ADDEQ] = ISASSIGNOP;
	__optype[SUBEQ] = ISASSIGNOP;
	__optype[MULEQ] = ISASSIGNOP;
	__optype[DIVEQ] = ISASSIGNOP;
	__optype[MODEQ] = ISASSIGNOP;
	__optype[LSHEQ] = ISASSIGNOP;
	__optype[RSHEQ] = ISASSIGNOP;
}


/*
** evalexp(terminal)
** legal values of terminal
**
** 0 - must have valid expression but don't require it to
**	be at end of line
** 1 - must have valid expression at end of line
** 2 - ok not to have an expression (set experr) but if there is
**	one it must be at end of line
** 3 - ok if no expression (eol) but if there is it must be at end
**	of line
** 4 - just like 0 (so evalexp can be passed -4)
**
** also if the terminal value if < 0 it has the same meanings as the
** above except it allows relational operators
*/
evalexp(terminal)
int terminal;
{
	int val;
	char *saddr;

	relops = 0;
	if (terminal < 0) {
		relops = 1;
		terminal = -terminal;
	}
	if (terminal == 4)
		terminal = 0;
	skipsp();
	saddr = cmdptr;
	experr = 0;
	if (terminal == 3 && *cmdptr == '\0') {
		side_effects = 1;
		return(0);
	}
	advance();
	val = expr();
	if (experr && terminal != 2) {
		printf("%s\n",saddr);
		for (val = token.pos - saddr; val > 0; val--)
			printf(" ");
		printf("^ bad expression\n");
		val = 0;
	}
	cmdptr = token.pos;
	if (!experr && terminal != 0 && *cmdptr != '\0') {
		printf("%s\n",saddr);
		for (val = token.pos - saddr; val > 0; val--)
			printf(" ");
		printf("^ extra characters after expression\n");
		val = 0;
	}
	side_effects = 1;
	return (val);
}

static
expr()
{
	int exprval;

	if (relops)
		exprval = lorop();
	else
		exprval = orop();

	return(exprval);
}

static
lorop()
{
	int lorval, oldse;

	lorval = landop();
	while (token.which == LOROP) {
		advance();
		if (lorval) {
			oldse = side_effects;
			side_effects = 0;
			landop();
			side_effects = oldse;
		} else {
			lorval = (lorval || landop());
		}
	}
	return(lorval);
}

static
landop()
{
	int landval, oldse;

	landval = relop();
	while (token.which == LANDOP) {
		advance();
		if (landval == 0) {
			oldse = side_effects;
			side_effects = 0;
			relop();
			side_effects = oldse;
		} else {
			landval = (landval && relop());
		}
	}
	return(landval);
}

static
relop()
{
	int relval;
	register char op;

	relval = orop();
	op = token.which;
	while (isrelop(op)) {
		advance();
		switch (op) {
		case EQOP: relval = (relval == orop()); break;
		case NEOP: relval = (relval != orop()); break;
		case GTOP: relval = (relval > orop()); break;
		case LTOP: relval = (relval < orop()); break;
		case GEOP: relval = (relval >= orop()); break;
		case LEOP: relval = (relval <= orop()); break;
		}
		op = token.which;
	}
	return(relval);
}

static
orop()
{
	int orval;

	orval = xorop();
	while (token.which == OROP) {
		advance();
		orval |= xorop();
	}
	return(orval);
}

static
xorop()
{
	int xorval;

	xorval = andop();
	while (token.which == XOROP) {
		advance();
		xorval ^= andop();
	}
	return(xorval);
}

static
andop()
{
	int andval;

	andval = shftop();
	while (token.which == ANDOP) {
		advance();
		andval &= shftop();
	}
	return(andval);
}

static
shftop()
{
	int shftval;
	register char	op;

	shftval = addop();
	op = token.which;
	while (op == LSHOP || op == RSHOP) {
		advance();
		if (op == LSHOP)
			shftval <<= addop();
		else
			shftval >>= addop();
		op = token.which;
	}
	return(shftval);
}


static
addop()
{
	int addval;
	register char	op;

	addval = mulop();
	op = token.which;
	while (op == ADDOP || op == SUBOP) {
		advance();
		if (op == ADDOP)
			addval += mulop();
		else
			addval -= mulop();
		op = token.which;
	}
	return(addval);
}

static
mulop()
{
	int mulval;
	register char	op;

	mulval = ptrop();
	op = token.which;
	while (op == MULOP || op == DIVOP || op == MODOP) {
		advance();
		if (op == MULOP)
			mulval *= ptrop();
		else if (op == DIVOP)
			mulval /= ptrop();
		else
			mulval %= ptrop();
		op = token.which;
	}
	return(mulval);
}

static
ptrop()
{
	int ptrval, ptrval2;
	register char	op;

	ptrval = colonop();
	op = token.which;
	while (op == PTROP) {
		advance();
		ptrval2 = colonop();
		ptrval = *(int *) (ptrval + ptrval2);
		op = token.which;
	}
	return(ptrval);
}

static
colonop()
{
	int colonval;

	colonval = unaryop();
	while (token.which == COLONOP) {
		advance();
		if (token.which == COLONOP) {
			/*
			** double colon means the segment selector
			** is really an index into the descriptor table
			*/
			advance();

			/*
			** logtolin converts the selector to an index
			*/
			colonval <<= 3;
		}
		/*
		** translate logical address to linear
		** address in user address space
		*/
		colonval = logtolin(unaryop(), colonval);
	}
	return(colonval);
}

static
unaryop()
{
	int unaryval;
	register char	op;

	op = token.which;
	if (op == NOTOP || op == COMPOP || op == MULOP || op == SUBOP) {
		advance();
		switch (op) {
		case NOTOP:
			unaryval = !unaryop();
			break;
		case COMPOP:
			unaryval =  ~unaryop();
			break;
		case MULOP:
			unaryval =  *(int *)unaryop();
			break;
		case SUBOP:
			unaryval =  -unaryop();
			break;
		}
		return (unaryval);
	}
	unaryval = term();
	return(unaryval);
}

static
term()
{
	unsigned int termval, newval;
	int termop, wasvlhs;
	struct token ptk;

	if (experr)
		return(0);
	if (token.which == LPAREN) {
		advance();
		termval = expr();
		if (token.which != RPAREN)
			experr = 1;
		else
			advance();
		return (termval);
	}
	termop = token.which;
	if (termop == INCR || termop == DECR) {
		advance();
		if (experr)
			return(0);
		if (token.which != REGISTER && token.which != VARIABLE) {
			experr = 1;
			return(0);
		}
		switch (token.size) {
		case 1:
			termval = *(unsigned char *)token.addr;
			break;
		case 2:
			termval = *(unsigned short *)token.addr;
			break;
		default:
			termval = *token.addr;
		}
		if (termop == INCR)
			termval++;
		else
			termval--;
		if (side_effects) {
			switch (token.size) {
			case 1:
				*(unsigned char *)token.addr = termval;
				break;
			case 2:
				*(unsigned short *)token.addr = termval;
				break;
			default:
				*token.addr = termval;
			}
		}
		advance();
		return(termval);
	}
	if (termop != CONSTANT && termop != SYMBOL &&
				termop != REGISTER && termop != VARIABLE) {
		experr = 1;
		return(0);
	}
	termval = token.value;
	ptk = token;
	advance();
	termop = token.which;
	if (isassignop(termop)) {
		if (ptk.which==REGISTER || ptk.which==VARIABLE) {
			switch (ptk.size) {
			case 1:
				termval = *(unsigned char *)ptk.addr;
				break;
			case 2:
				termval = *(unsigned short *)ptk.addr;
				break;
			default:
				termval = *ptk.addr;
			}
			newval = termval;
			advance();
			if (experr)
				return(0);
			switch (termop) {
			case INCR:
				newval++;
				break;
			case DECR:
				newval--;
				break;
			default:
				termval = expr();
				if (experr)
					return(0);
				switch (termop) {
				case ASSIGN:	newval = termval; break;
				case OREQ:	newval |= termval; break;
				case XOREQ:	newval ^= termval; break;
				case ANDEQ:	newval &= termval; break;
				case ADDEQ:	newval += termval; break;
				case SUBEQ:	newval -= termval; break;
				case MULEQ:	newval *= termval; break;
				case DIVEQ:	newval /= termval; break;
				case MODEQ:	newval %= termval; break;
				case LSHEQ:	newval <<= termval; break;
				case RSHEQ:	newval >>= termval; break;
				}
				termval = newval;
			}
			if (side_effects) {
				switch (ptk.size) {
				case 1:
					*(unsigned char *)ptk.addr = newval;
					break;
				case 2:
					*(unsigned short *)ptk.addr = newval;
					break;
				default:
					*ptk.addr = newval;
				}
			}
		} else {
			experr = 1;
			termval = 0;
		}
	}
	return(termval);
}

static
advance()
{
	register char	c, c2;
	unsigned int valu;
	int	i, base;
	struct sym *sp;
	
	skipsp();
	token.pos = cmdptr;
	c = *cmdptr++;
	if (c == '\0') {
		token.which = EOL;
		return;
	}
	if (c == '$') {
		token.which = CONSTANT;
		token.value = (unsigned int) laddr.a_int;
		return;
	}
	if (c == '@') {
		token.which = CONSTANT;
		token.value = (unsigned int) dot;
		return;
	}
	if (isdigit(c,0)) {
		base = 16;
		if (c == '0') {
			/*
			** check for number base
			*/
			c = *cmdptr++;
			switch (c) {
			case 'o':
			case 'O':
				base = 8; 
				break;
			case 't':
			case 'T':
				base = 10; 
				break;
			case 'x':
			case 'X':
				base = 16; 
				break;
			default:
				/*
				** back up to the '0'
				*/
				cmdptr -= 2;
			}
			c = *cmdptr++;
			if (!isdigit(c,(base==16)))
				goto err;
		}
number:
		valu = 0;
		for (;;) {
			valu = (valu * base) + digval(c);
			c = *cmdptr++;
			if (!isdigit(c, (base==16)))
				break;
		}
		cmdptr--;
		/*
		** "abcxyz" should be detected as
		** an error here
		*/
		if (isalpha(c))
			goto err;
		token.which = CONSTANT;
		token.value = valu;
		return;
	}
	if (c == '\'') {
		valu = 0;
		while ((c = *cmdptr++) != '\'') {
			if (c == 0)
				break;
			valu <<= 8;
			valu |= (unsigned char) c;
		}
		token.which = CONSTANT;
		token.value = valu;
		return;
	}
	if (isalpha(c)) {
		getname(token.pos);
		if (sp = srchsym(srch_name,vtable,nvars)) {
			if (isfunc(sp,&valu)) {
				if (experr)
					goto err;
				token.which = CONSTANT;
				token.value = valu;
			} else {
				token.which = VARIABLE;
				token.size = 4;
				token.value = sp->value;
				token.addr = &sp->value;
			}
			return;
		}
		if (sp = srchsym(srch_name,rtable,NREGS)) {
			token.which = REGISTER;
			token.addr = (unsigned int *) &dbr0ptr[sp->value];
			/*
			** all two character register names
			** are 16 bit segment registers
			*/
			if (srch_name[2] == '\0') {
				token.size = 2;
				token.value = *(unsigned short *) token.addr;
			} else {
				token.size = 4;
				token.value = *(unsigned int *) token.addr;
			}
			return;
		}
		if (dbgstbl[0].value != -1 &&
				(sp = srchsym(srch_name,dbgstbl,dbgscnt))) {
			if (isfunc(sp,&valu)) {
				if (experr)
					goto err;
				token.which = CONSTANT;
				token.value = valu;
			} else {
				token.which = SYMBOL;
				token.value = sp->value;
				token.addr = &sp->value;
			}
			return;
		}
		/*
		** token not accepted as symbol so must
		** restore cmdptr to start of current token
		*/
		cmdptr = token.pos;
		c = *cmdptr++;
	}
	if (isdigit(c,1)) {
		base = 16;
		goto number;
	}
	/*
	** error if c is [a-zA-Z] and not symbol or number
	*/
	if (isalpha(c))
		goto err;
	/*
	** operators
	*/
	c2 = *cmdptr;
	switch (c) {
	case '|':
		if (c2 == '|') {
			token.which = LOROP;
			cmdptr++;
		} else {
			token.which = OROP;
		}
		break;
	case '&':
		if (c2 == '&') {
			token.which = LANDOP;
			cmdptr++;
		} else {
			token.which = ANDOP;
		}
		break;
	case '=':
		token.which = ASSIGN;
		break;
	case '!':
		if (c2 == '=') {
			token.which = NEOP;
			cmdptr++;
		} else {
			token.which = NOTOP;
		}
		break;
	case '>':
		if (c2 == '=') {
			token.which = GEOP;
			cmdptr++;
		} else if (c2 == '>') {
			token.which = RSHOP;
			cmdptr++;
		} else {
			token.which = GTOP;
		}
		break;
	case '<':
		if (c2 == '=') {
			token.which = LEOP;
			cmdptr++;
		} else if (c2 == '<') {
			token.which = LSHOP;
			cmdptr++;
		} else {
			token.which = LTOP;
		}
		break;
	case '^':
		token.which = XOROP;
		break;
	case '+':
		if (c2 == '+') {
			token.which = INCR;
			cmdptr++;
		} else {
			token.which = ADDOP;
		}
		break;
	case '-':
		if (c2 == '-') {
			token.which = DECR;
			cmdptr++;
		} else if (c2 == '>') {
			token.which = PTROP;
			cmdptr++;
		} else {
			token.which = SUBOP;
		}
		break;
	case '*':
		token.which = MULOP;
		break;
	case '/':
		token.which = DIVOP;
		break;
	case '%':
		token.which = MODOP;
		break;
	case '(':
		token.which = LPAREN;
		break;
	case ')':
		token.which = RPAREN;
		break;
	case '~':
		token.which = COMPOP;
		break;
	case ',':
		token.which = COMMA;
		break;
	case ':':
		token.which = COLONOP;
		break;
	default:
		token.which = UNKNOWN;
		break;
	}
	c2 = *cmdptr;
	if ((c2 = *cmdptr) == '=') {
		switch (token.which) {
		case ASSIGN:
			token.which = EQOP;
			cmdptr++;
			break;
		case OROP:
		case XOROP:
		case ANDOP:
		case ADDOP:
		case SUBOP:
		case MULOP:
		case DIVOP:
		case MODOP:
		case LSHOP:
		case RSHOP:
			token.which = token.which - OROP + OREQ;
			cmdptr++;
			break;
		}
	}
	return;

err:
	token.which = UNKNOWN;
	return;
}

static
isdigit(c,hexok)
char c;
{
	if (c >= '0' && c <= '9')
		return(1);
	if (hexok && c >= 'a' && c <= 'f')
		return(1);
	if (hexok && c >= 'A' && c <= 'F')
		return(1);
	return(0);
}

static
digval(c)
char c;
{
	if (c >= '0' && c <= '9')
		return(c - '0');
	if (c >= 'a' && c <= 'f')
		return(c - 'a' + 10);
	if (c >= 'A' && c <= 'F')
		return(c - 'A' + 10);
}

static
getname(s)
register char *s;
{
	register char *t;
	register int i;

	/*
	** zero out srch_name
	** assumes MAXNAMLEN == 12
	*/
#if MAXNAMLEN != 12
	fix the code below
#endif
	*(long *)srch_name = 0L;
	*(long *)(srch_name + 4) = 0L;
	*(long *)(srch_name + 8) = 0L;

	t = srch_name;
	i = 0;
	for (;;) {
		if (!isalpha(*s)) 
			break;
		if (i < MAXNAMLEN)
			*t++ = *s;
		s++;
		i++;
	}
	cmdptr = s;
}

static
isalpha(c)
register char c;
{
	if (c >= '0' && c <= '9')
		return(1);
	if (c >= 'a' && c <= 'z')
		return(1);
	if (c >= 'A' && c <= 'Z')
		return(1);
	if (c == '_')
		return(1);
	return(0);
}


#define MAXARGS	10

static
isfunc(sp,value)
struct sym *sp;
unsigned int *value;
{
	int numargs;
	unsigned int	(*f)(), a[MAXARGS], val;

	skipsp();
	if (*cmdptr != '(')
		return(0);

	f = (unsigned int (*)())sp->value;
	numargs = 0;
	advance();	/* get '(' */
	advance();	/* get first arg or ')' */
	if (token.which != RPAREN) {
		for (;;) {
			val = expr();
			if (experr)
				return(1);
			if (numargs >= MAXARGS) {
				printf("too many arguments\n");
				experr = 1;
				return;
			}
			a[numargs++] = val;
			if (token.which != COMMA)
				break;
			advance();
		}
		if (token.which != RPAREN) {
			experr = 1;
			return(1);
		}
	}
	switch(numargs) {
	case 0:
		val = (*f)();
		break;
	case 1:
		val = (*f)(a[0]);
		break;
	case 2:
		val = (*f)(a[0],a[1]);
		break;
	case 3:
		val = (*f)(a[0],a[1],a[2]);
		break;
	case 4:
		val = (*f)(a[0],a[1],a[2],a[3]);
		break;
	case 5:
		val = (*f)(a[0],a[1],a[2],a[3],a[4]);
		break;
	case 6:
		val = (*f)(a[0],a[1],a[2],a[3],a[4],a[5]);
		break;
	case 7:
		val = (*f)(a[0],a[1],a[2],a[3],a[4],a[5],a[6]);
		break;
	case 8:
		val = (*f)(a[0],a[1],a[2],a[3],a[4],a[5],a[6],a[7]);
		break;
	case 9:
		val = (*f)(a[0],a[1],a[2],a[3],a[4],a[5],a[6],a[7],a[8]);
		break;
	case 10:
		val = (*f)(a[0],a[1],a[2],a[3],a[4],a[5],a[6],a[7],a[8],a[9]);
		break;
	default:
		printf("too many arguments\n");
		experr = 1;
		return(1);
	}
	*value = val;
	return(1);
}



/*
** aliases
*/

#define MAXARG 10

doalias()
{
	static char lname[MAXNAMLEN+1], work[MAXLINE];
	struct sym *sp;
	int i, narg, marg;
	char	*args[MAXARG], c;
	register char *cp;

	cp = cmdptr;
	cmdptr = cmdline;
	skipsp();
	if (!isalpha(*cmdptr)) {
		cmdptr = cp;
		return(0);
	}
	getname(cmdptr);
	sp = srchsym(srch_name,atable,naliases);
	if (sp == NULL) {
		cmdptr = cp;
		return(0);
	}
	bcopy(cmdline,work,MAXLINE);
	cmdptr = work;
	narg = 0;
	for(;;) {
		skipsp();
		if (*cmdptr == '\0')
			break;
		if (narg >= MAXARG) {
			printf("too many args\n");
			return(0);
		}
		args[narg++] = cmdptr;
		skipnsp();
		if (*cmdptr == '\0')
			break;
		*cmdptr++ = '\0';
	}
	cmdptr = cmdline;
	marg = 0;
	for (cp=(char *)sp->value; *cp; ) {
		if (cmdptr >= &cmdline[MAXLINE-1])
			goto toolong;
		if ((c = *cp++) != '$') {
			*cmdptr++ = c;
			continue;
		}
		c = *cp++;
		if (c == '*') {
			for (c=1; c<narg; c++)
				if (!addarg(args[c]))
					goto toolong;
			marg = narg-1;
			continue;
		}
		if (c < '0' || c > '9') {
			*cmdptr++ = c;
			continue;
		}
		c -= '0';
		if (c >= narg)
			goto neargs;
		if (c > marg)
			marg = c;
		if (!addarg(args[c]))
			goto toolong;
	}
	while (marg < narg-1) {
		*cmdptr++ = ' ';
		if (!addarg(args[++marg]))
			goto toolong;
	}
	*cmdptr = '\0';
	cmdptr = cmdline;
	return(1);

toolong:
	printf("alias too long\n");
	return(0);

neargs:
	printf("not enough args\n");
	return(0);
}

addarg(dp)
register char *dp;
{
	if (cmdptr+strlen(dp) >= &cmdline[MAXLINE-1])
		return(0);
	while (*dp)
		*cmdptr++ = *dp++;
	return(1);
}

alcmd()
{
	register struct sym *sp;
	register int i;
	static char lname[MAXNAMLEN+1];

	skipsp();
	if (*cmdptr == '\0') {
		/*
		** list aliases
		*/
		for (i=0; i<naliases; i++) {
			bcopy(atable[i].name,lname,MAXNAMLEN);
			lname[MAXNAMLEN] = '\0';
			printf("\t%s = %s\n",lname,atable[i].value);
		}
		return(0);
	}
	if (!isalpha(*cmdptr)) {
		printf("bad alias\n");
		return(-1);
	}
	getname(cmdptr);
	sp = srchsym(srch_name,atable,naliases);
	skipsp();
	if (*cmdptr == '\0') {
		if (sp == NULL) {
			printf("not aliased\n");
			return(-1);
		}
		bcopy(sp->name,lname,MAXNAMLEN);
		lname[MAXNAMLEN] = '\0';
		printf("\t%s = %s\n",lname,sp->value);
		return(0);
	}
	if (sp) {
		if ((i=freesym(srch_name,atable,naliases)) == 0) {
			printf("failed to free alias\n");
			return(-1);
		} else {
			aliasptr -= i;
			naliases--;
		}
	}
	if (naliases >= MAXALIASES) {
		printf("no more alias space\n");
		return(-1);
	}	
	i = strlen(cmdptr) + 1;
	if (i > (MAXASPACE - (aliasptr - abuffer))) {
		printf("no more alias space\n");
		return(-1);
	}
	sp = &atable[naliases++];
	bcopy(srch_name,sp->name,MAXNAMLEN);
	bcopy(cmdptr,aliasptr,i);
	sp->value = (long)aliasptr;
	aliasptr += i;
	return(0);
}

uacmd()
{
	register struct sym *sp;
	register int i;

	skipsp();
	if (*cmdptr == '\0') {
		/*
		**  unalias all aliases
		*/
		naliases = 0;
		aliasptr = abuffer;
		return(0);
	}
	if (!isalpha(*cmdptr)) {
		printf("bad alias\n");
		return(-1);
	}
	getname(cmdptr);
	sp = srchsym(srch_name,atable,naliases);
	skipsp();
	if (sp == NULL) {
		printf("not aliased\n");
		return(-1);
	} else {
		if ((i=freesym(srch_name,atable,naliases)) == 0) {
			printf("failed to free alias\n");
			return(-1);
		} else {
			aliasptr -= i;
			naliases--;
		}
	}

	return(0);
}


/*
** assign value to variable
*/

secmd()
{
	register struct sym *sp;
	register int i;
	static char lname[MAXNAMLEN+1];

	skipsp();
	if (*cmdptr == '\0') {
		/*
		** list variables
		*/
		for (i=0; i<nvars; i++) {
			bcopy(vtable[i].name,lname,MAXNAMLEN);
			lname[MAXNAMLEN] = '\0';
			printf("\t%s = 0x%x (%d.)\n",lname,
				vtable[i].value,vtable[i].value);
		}
		return(0);
	}
	if (!isalpha(*cmdptr)) {
		printf("bad variable name\n");
		return(-1);
	}
	getname(cmdptr);
	sp = srchsym(srch_name,vtable,nvars);
	skipsp();
	if (*cmdptr != '=') {
		printf("missing =\n");
		return(-1);
	}
	cmdptr++;
	skipsp();
	if (*cmdptr == '\0') {
		/*
		** delete variable
		*/
		if (sp == NULL) {
			printf("variable not defined\n");
			return(-1);
		}
		delete_varp(sp);
		return(0);
	}
	bcopy(srch_name,lname,MAXNAMLEN);
	i = evalexp(1);
	if (experr)
		return(-1);
	if (sp) {
		sp->value = i;
	} else {
		if (insert_var(lname,i)) {
			printf("no more variable space\n");
			return(-1);
		}
	}
	return(0);
}

/*
** Support for external insert/delete of variables.
** Used mainly by loadable drivers.
*/
insert_var(name,value)
char *name;
long value;
{
	struct sym *sp;

	if (nvars >= MAXVARS)
		return(-1);

	sp = &vtable[nvars++];
	bcopy(name,sp->name,MAXNAMLEN);
	sp->value = value;
	return(0);
}

delete_var(name)
register char *name;
{
	register int i;
	register struct sym *vp;

	for (i=0, vp=vtable; i<nvars; i++, vp++) {
		/*
		** the following assumes MAXNAMLEN is 12
		*/
#if MAXNAMLEN != 12
	fix the code below
#endif
		if (*(long *)name != *(long *)vp->name)
			continue;
		if (*(long *)(name+4) != *(long *)(vp->name+4))
			continue;
		if (*(long *)(name+8) != *(long *)(vp->name+8))
			continue;

		delete_varp(vp);
		return(0);
	}
	return(-1);
}

delete_varp(sp)
struct sym *sp;
{
	register int i;

	if (sp == NULL)
		return;

	for (i=nvars-(sp-vtable); --i > 0; ) {
		*sp = *(sp+1);
		sp++;
	}
	nvars--;
}




static struct sym *
srchsym(name,tab,count)
register char *name;
register struct sym *tab;
register int count;
{

	for (; count; count--, tab++) {
		/*
		** the following assumes MAXNAMLEN is 12
		*/
#if MAXNAMLEN != 12
	fix the code below
#endif
		if (*(long *)name != *(long *)tab->name)
			continue;
		if (*(long *)(name+4) != *(long *)(tab->name+4))
			continue;
		if (*(long *)(name+8) != *(long *)(tab->name+8))
			continue;
		return(tab);
	}
	return(NULL);
}

int
freesym(name,tab,count)
register char *name;
register struct sym *tab;
register int count;
{
	int		symsize;
	struct sym	*t;
	char		*p;
	int		nbytes;

	for (; count; count--, tab++) {
		/*
		** the following assumes MAXNAMLEN is 12
		*/
#if MAXNAMLEN != 12
	fix the code below
#endif
		if (*(long *)name != *(long *)tab->name)
			continue;
		if (*(long *)(name+4) != *(long *)(tab->name+4))
			continue;
		if (*(long *)(name+8) != *(long *)(tab->name+8))
			continue;

		symsize = strlen((char *) tab->value) + 1;
		p = (char *) tab->value;
		t = tab++;
		for (count--; count; count--,tab++) {
			t->value = (unsigned int) p;
			bcopy(tab->name,t->name,MAXNAMLEN);
			nbytes = strlen((char *) tab->value) + 1;
			bcopy((char *) tab->value,p,nbytes);
			p += nbytes;
			t = tab;
		}
		return(symsize);		/* amount of space freed */
	}
	return(0);				/* no space freed */
}

struct sym *
adrsrch( adr )
	uint	adr;
{
	register struct sym *p;
	register int	l, m, h;

	l = dbgslow;
	h = dbgscnt;
	for ( ; h-l > 1; ) {
		m = (h - l) / 2 + l;
		p = dbgstbl + m;
		if ( p->value < adr ) 
			l = m;
		else
		if ( p->value > adr )
			h = m;
		else
			return( p );
	}
	return( dbgstbl + l );
}

dbgsyminit()
{
	register struct sym *p;

	for (p = dbgstbl; dbgslow < dbgscnt; ++p, ++dbgslow) 
		if (p->value >= 0x10000)
			break;
}

char *
spntadr( adr )
	unsigned	adr;
{
	static	char	buf[32];
	static	char	name[MAXNAMLEN+1];
	struct	sym	*p;

	if ( adr >= dbgstbl[0].value && kvirtaddr(adr) ) {
		p = adrsrch( adr );	/* binary search the addr   */
		adr -= p->value;	/* offset from the symbol   */
		bcopy(p->name, name, MAXNAMLEN);
		name[MAXNAMLEN] = 0;
		if ( adr ) 	
			sprintf( buf, "%s+%x", name, adr );
		else
			strcpy( buf, name );
	}
	else
		sprintf( buf, "0x%X", adr );
	return(buf);
}

pntadr( adr )
{
	char	*p;

	p = spntadr(adr);
	printf("%s", p);
}
