
/*
 * Location of the users' stored registers relative to EAX.
 * Usage is u.u_ar0[XX].
 *
 * NOTE: ERR is the error code.
 *	 ES, FS & GS not actually residing on the stack, they
 *	 are visible only thru ptrace().
 */
#if 0
#define GS	16
#define FS	15

#define SS	14
#define UESP	13
#define EFL	12
#define CS	11
#define EIP	10
#define ERR	9
#define TRAPNO	8
#define EAX	7
#define ECX	6
#define EDX	5
#define EBX	4
#define ESP	3
#define EBP	2
#define ESI	1
#define EDI	0
#else
#define EBX 0
#define ECX 1
#define EDX 2
#define ESI 3
#define EDI 4
#define EBP 5
#define EAX 6
#define DS 7
#define ES 8
#define FS 9
#define GS 10
#define TRAPNO 11
#define EIP 12
#define CS  13
#define EFL 14
#define ESP 15
#define UESP 15
#define SS   16
#endif
/*
 * All kernel entries are thru a trap/interrupt stack frame.
 * This is SW extended to save/restore registers, etc.
 *
 * The error code pushed on some traps is poped into l.trap_err_code
 * to make stack frame consistent.
 *
 * A push-all instruction is used to push all user general registers,
 * thus syscall() and trap() handler can have arbitrary call frames.
 *
 * Currently don't save/restore user DS,ES on entry/exit; rather assume
 * user code never changes these, and set to constants on kernel exit.
 * FS, GS are totally ignored.
 *
 * Once in syscall/trap handler, stack looks like:
 *
 *	old SS, padded		only if inter-segment (user->kernel)
 *	old SP			only if inter-segment (user->kernel)
 *	flags
 *	CS, padded		sense user-mode entry from RPL field
 *	EIP			return context
 *	EAX			scratch registers
 *	ECX			ditto
 *	EDX			more such
 *	EBX			register variable
 *	(unused)		temp SP (from push-all instruction)
 *	EBP			of interrupted frame
 *	ESI			register variable
 *	EDI			register variable
 *	trap-type		trap only
 *	EIP			back to locore
 *	XXX			syscall or trap call frame
 */

/*
 * Interrupts save scratch registers in a consistent order with a push-all;
 * save entry SPL below this.  Thus, after interrupt entry, stack looks like:
 *
 *	old SS, padded		only if inter-segment (user->kernel)
 *	old SP			only if inter-segment (user->kernel)
 *	flags
 *	CS, padded		sense user-mode entry from RPL field
 *	IP			return context
 *	EAX			scratch registers
 *	ECX			ditto
 *	EDX			more such
 *	old SPL			entry SLIC local mask  (esp points here)
 */

