#include "serial.h"

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/time.h>
#include <linux/delay.h>

#include <asm/segment.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/delay.h>


void serial_init(void)
{
  	hw_serial *p = (hw_serial *) hw_SERIAL;

	OUTB(p, IER, 0); /* disable interrupts */
	OUTB(p, FCR, 0); /* disable FIFO */

	OUTB(p, LCR, hw_serial_LCR_DLAB); /* Divisor Latch Access */        
	OUTB(p, DLL, hw_serial_DLR_9600 & 0xFF); /* set baud rate */
	OUTB(p, DLM, hw_serial_DLR_9600 >> 8);

	/* disable Divisor Latch Access,9600 Baud 8 1 stop no parity */
	OUTB(p, LCR, hw_serial_LCR_WLS_mask);

	/* Enable both Rx and Tx FIFOs and flush them as well */
	OUTB(p, FCR, FCRvalue);

	/* enable Modem Control Register */
	OUTB(p, MCR, MCRvalue);
	INB(p, LSR);
	INB(p, RBR);
}

int port_getchar()
{
  	hw_serial *p = (hw_serial *) hw_SERIAL;
	int	c;

	if (INB(p, LSR) & hw_serial_LSR_RDR)
		c = INB(p, RBR);
	else
		c = -1;

	return(c);
}

void port_putchar(unsigned char c)
{
  	hw_serial *p = (hw_serial *) hw_SERIAL;

	if (c == '\n')
		port_putchar((unsigned char) '\r');
	while ((INB(p, LSR) & hw_serial_LSR_THRE) == 0);
	OUTB(p, THR, c); 
}

void port_printf(unsigned char *fmt, int a, int b, int c, int d, int e)
{
	char	*p, msg[128];

	sprintf(msg, fmt, a, b, c, d, e);
	p = msg;
	while(*p)
		port_putchar(*p++);
}
