
typedef volatile char	hw_serial;

#define	RBR	0
#define	IER	1
#define	IIR	2
#define	LCR	3
#define	MCR	4
#define	LSR	5
#define	MSR	6
#define	SCR	7

#define	hw_SERIAL	0x3f8
#define	hw_SERIALb	0x2f8

#define	OUTB(hwptr, off, val)	outb(val, (int) (hwptr + off))
#define	INB(hwptr, off)		inb((int) (hwptr + off))

#define	THR	RBR
#define FCR	IIR
#define	DLL	RBR
#define DLM	IER

#define hw_serial_LCR_WLS_mask	(0x03)	/* Word length mask	*/
#define hw_serial_LCR_WL_5bits	(0x00)
#define hw_serial_LCR_WL_6bits	(0x01)
#define hw_serial_LCR_WL_7bits	(0x02)
#define hw_serial_LCR_WL_8bits	(0x03)
#define hw_serial_LCR_SBSEL	(1 << 2)
#define hw_serial_LCR_PAREN	(1 << 3)
#define hw_serial_LCR_EVPAR	(1 << 4)
#define hw_serial_LCR_STICK	(1 << 5)
#define hw_serial_LCR_BREAK	(1 << 6)
#define hw_serial_LCR_DLAB	(1 << 7)

#define	hw_serial_LCR_default	(hw_serial_LCR_WL_8bits)

/* For 1.8432MHz crystal */
#define hw_serial_DLR_300	(384)
#define hw_serial_DLR_1200	(96)
#define hw_serial_DLR_2400	(48)
#define hw_serial_DLR_4800	(24)
#define hw_serial_DLR_9600	(12)
#define hw_serial_DLR_19200	(6)
#define hw_serial_DLR_38400	(3)
#define hw_serial_DLR_56000	(2)	/* Error = 2.85% */

/* IER : Interrupt Enable Register */
#define hw_serial_IER_ERBFI	(1 << 0) /* 1 = enable Rx data available interrupt */
#define hw_serial_IER_ETBEI	(1 << 1) /* 1 = enable Tx holding reg. empty interrupt */
#define hw_serial_IER_ELSI	(1 << 2) /* 1 = enable Rx line status interrupt */
#define hw_serial_IER_EDSSI	(1 << 3) /* 1 = enable modem status interrupt */

/* IIR : Interrupt Identification Register */
#define hw_serial_IIR_IPENDN	(1 << 0)   /* 0 = interrupt pending; 1 = none */
#define hw_serial_IIR_IID_mask	(0x3 << 1) /* pending interrupts */
#define hw_serial_IIR_IID_MS	(0x0 << 1) /* Modem Status */
#define hw_serial_IIR_IID_THRE	(0x1 << 1) /* Tx Holding Reg. Empty */
#define hw_serial_IIR_IID_RDA	(0x2 << 1) /* Rx Data Available */
#define hw_serial_IIR_IID_RLS	(0x3 << 1) /* Rx Line Status */

/* NOTE: The following is the status when the "hw_serial_IIR_FIFEN" bit is
 * also set.
 */
#define hw_serial_IIR_IID_FS	(0x2 << 1) /* FIFO Status; FIFO thresh; wait timeout */
#define hw_serial_IIR_FIFEN	(1 << 3)   /* 0 = 16C450 mode; 1 = 16C550 mode */
#define hw_serial_IIR_FE_mask	(0x3 << 6) /* FIFO enable information */

/* LSR : Line Status Register */
#define hw_serial_LSR_RDR	(1 << 0) /* Rx Data Ready */
#define hw_serial_LSR_EROVR	(1 << 1) /* Error Rx OVerRun */
#define hw_serial_LSR_ERPAR	(1 << 2) /* Error Rx PARity */
#define hw_serial_LSR_ERFRM	(1 << 3) /* Error Rx FRaMe */
#define hw_serial_LSR_ERBRK	(1 << 4) /* Error Rx BReaK */
#define hw_serial_LSR_THRE	(1 << 5) /* Tx Holding Register Empty */
#define hw_serial_LSR_TEMT	(1 << 6) /* Tx EMpTy */
#define hw_serial_LSR_ERFIFO	(1 << 7) /* Error Rx FIFO full */

/* FCR : FIFO Control Register */
#define hw_serial_FCR_FIFOEN	(1 << 0)   /* 1 = enable Rx and Tx FIFOs */
#define hw_serial_FCR_RFCLR	(1 << 1)   /* 1 = clear Rx FIFO */
#define hw_serial_FCR_TFCLR	(1 << 2)   /* 1 = clear Tx FIFO */
#define hw_serial_FCR_RMODE	(1 << 3)   /* changes RXRDY and TXRDY modes */
#define hw_serial_FCR_FTL_mask	(0x3 << 6) /* Rx FIFO trigger level */
#define hw_serial_FCR_FTL_1	(0x0)	   /* 1 byte */
#define hw_serial_FCR_FTL_4	(0x1)	   /* 4 bytes */
#define hw_serial_FCR_FTL_8	(0x2)	   /* 8 bytes */
#define hw_serial_FCR_FTL_14	(0x3)      /* 14 bytes */

/* MCR : Modem Control Register */
#define hw_serial_MCR_DTR	(1 << 0) /* 1 = DTR forced low */
#define hw_serial_MCR_RTS	(1 << 1) /* 1 = RTS forced low */
#define hw_serial_MCR_OUT2	(1 << 3) /* 0 = OUT2 high; 1 = OUT2 low */
#define hw_serial_MCR_LBACK	(1 << 4) /*  LoopBACK enable (testing) */

/* MSR : Modem Status Register */
#define hw_serial_MSR_DCTS	(1 << 0) /* Delta CTS */
#define hw_serial_MSR_DDSR	(1 << 1) /* Delta DSR */ 
#define hw_serial_MSR_TERI	(1 << 2) /* Trailing Edge Ring Indicator */
#define hw_serial_MSR_CTS	(1 << 4) /* Clear To Send */
#define hw_serial_MSR_DSR	(1 << 5) /* Data Set Ready */
#define hw_serial_MSR_RI	(1 << 6) /* Ring Indicator */

/*---------------------------
 *Line Status Register (LSR)
 *---------------------------
 */
#define FifoError               0x80
#define TxEmpty                 0x40
#define TxHoldEmpty             0x20
#define RxBreakDetectError      0x10
#define RxFramingError          0x08
#define RxParityError           0x04
#define RxOverrunError          0x02
#define RxDataReady             0x01

/*
 * FIFO Control Register (FCR)
 */
#define TxFIFOclear     0x4
#define RxFIFOclear     0x2
#define TxRxFIFOen      0x1
#define FIFOlevel       (0x2 << 6)	/* 8 bytes deep */
#define	FCRvalue	(TxRxFIFOen | TxFIFOclear | RxFIFOclear | FIFOlevel)

#define DLAccess        0x80    /* 0=Access Rx buffer,1=Div. latch access*/

/*-----------------------------
 *Modem Control Register (MCR)
 *-----------------------------
 */
#define LoopBackEnable          0x10
#define MCRInterruptEnable      0x8
#define RequestToSend           0x2
#define DataTransmReady         0x1

#define MCRvalue     (RequestToSend | DataTransmReady | MCRInterruptEnable)


/*--------------------------------
 *Interrupt Enable Register (IER)
 *--------------------------------
 */
#define EnModemStatusInt        0x8     /* 1=modem status interrupt*/
#define EnRxLineStatusInt       0x4     /* 1=Rx line status interrupt*/
#define EnTxRegEmptyInt         0x2     /* 1=Tx holding reg empty interrupt*/
#define EnRxDataAvailInt        0x1     /* 1=Rx data available interrupt*/

/*
 *- enable INT for Rx data available
 *- enable INT for Rx line status
 */
#define       IERvalue   EnRxDataAvailInt+EnRxLineStatusInt+EnTxRegEmptyInt

