
#include "pic.h"       /* PIC constants */
#include <asm/io.h>

#define	OUTB(port, val)	outb(val, port)

/* Programmable Interrupt Controllers */

/* Initialized data for Programmable Interupt Controllers */

unsigned short cmdport[NPIC]    /* command port addrs for pics */
	= { MCMD_PORT, SCMD_PORT };

unsigned short imrport[NPIC]    /* intr mask port addrs for pics */
	= { MIMR_PORT, SIMR_PORT };

unsigned char masterpic[NPIC]   /* index of this pic's master (for 82380) */
	= { 0, 0 };

unsigned char masterline[NPIC]  /* line on master this slave connected to */
	= { 0, MASTERLINE };

unsigned char intbase[NPIC]     /* intr number of this pic's IR0 */
	= { 0, SLAVEBASE };

unsigned char picbuffered = PICBUFFERED;        /* PICs in buffered mode */

int npic = NPIC;                /* number of pics configured */

/*
 * picinit() initializes the PICs to ipl 0 and leaves the IF flag off,
 * putting the system at interrupt priority level IPLHI,
 * the level set by splhi and spl7.
 *
 * Interrupts are enabled by spl'ing to a lower ipl.
 *
 * picinit() is called first thing in main().
 *
 * Failed logic check actions in picinit() require that the console driver
 * be initialized before the pics.  cmn_err() should not
 * be called because the sysdump() routine uses interrupts.
 */

pic_init()
{
	int mask, intno, bit, slaves, cmd, imr, pic, level;

	/*
	 * Set up iplmask[][] from ivect[] and intpri[].
	 * iplmask[0][] is ORed into other levels
	 * so that unused interrupts are disabled.
	 * If an interupt number is not configured, set its bit in
	 * iplmask[0][]; otherwise set its bit in the masks at its
	 * priority level and higher.
	 *
	 * NOTE: master pic must always be pic zero.
	 *	 pic base vector is always PIC_VECTBASE + picno * 8.
	 */

	/* identify lines on master to which slaves are connected */
	slaves = 0;
	for (pic = 1; pic < npic; pic++)        /* for each slave */
		slaves |= 1 << masterline[pic];

	/*
	 * initialize master first
	 */
	
	/* ICW1: Edge-triggered, Cascaded, need ICW4 */
	OUTB(cmdport[0], PIC_ICW1BASE|PIC_NEEDICW4);
	
	/* ICW2: start master vectors at PIC_VECTBASE */
	OUTB(imrport[0], PIC_VECTBASE);
	
	/* ICW3: define which lines are connected to slaves */
	OUTB(imrport[0], slaves);

	/* ICW4: buffered master (?), norm eoi, mcs 86 */
	OUTB(imrport[0],
	     picbuffered ? PIC_MASTERBUF|PIC_86MODE : PIC_86MODE );

	/* OCW1: Start the master with all interrupts off */
	OUTB(imrport[0], 0xFF);
	
	/* OCW3: set master into "read isr mode" */
	OUTB(cmdport[0], PIC_READISR);
	
	/*
	 * now initialize slaves
	 */

	for (pic = 1; pic < npic; pic++) {
		cmd = cmdport[pic];
		imr = imrport[pic];

		/* ICW1: Edge-triggered, Cascaded, need ICW4 */
		OUTB(cmd, PIC_ICW1BASE|PIC_NEEDICW4);

		/* ICW2: set base of vectors */
		OUTB(imr, PIC_VECTBASE + pic * 8);

		/* ICW3: specify ID for this slave */
		OUTB(imr, masterline[pic]);

		/* ICW4: buffered slave (?), norm eoi, mcs 86 */
		OUTB(imr,
		     picbuffered ? PIC_SLAVEBUF|PIC_86MODE : PIC_86MODE );

		/* OCW3: set pic into "read isr mode" */
		OUTB(cmd, PIC_READISR);
	}
}

