#ifdef	__KERNEL__
#include <linux/config.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/ioctl.h>
#include <linux/module.h>
#include <linux/stat.h>
#include <linux/flash_fs.h>
#include <linux/flash.h>
#include <linux/delay.h>
#include "fcard.h"

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>

/*	Flash Card Major Device is set to 63 */
#define MAJOR_NR 	FCARD_MAJOR

#include <linux/debug.h>
#include <linux/blk.h>

#else

#include "system.h"
#include "errno.h"
#include "flash_fs.h"
#include "flash.h"
#include "fcard.h"

#endif
//#ifdef  WYSE_MIDWAY
#define MIDWAY_INDEX_REG_PORT   0x03e0   /*Index register port address*/
#define MIDWAY_DATA_REG_PORT 	 0x03e1   /*Data register port address*/

#define	MIDWAY_FC_WINDOW_SIZE	 0x04000   // only 16K can be used from 0xD4000-0xD7FFF
#define	MIDWAY_FC_WINDOW_START  0xD4000   // must start here to 0xD7FFF
//#else
#define MARS_INDEX_REG_PORT  0x03e0   /*Index register port address*/
#define MARS_DATA_REG_PORT 	 0x03e1   /*Data register port address*/

#define	MARS_FC_WINDOW_SIZE	 0x0200000 //2048K=2M
#define	MARS_FC_WINDOW_START 0x3900000 //57M

#define LEO_INDEX_REG_PORT 	0xF900   /*Index register port address*/
#define LEO_DATA_REG_PORT 	0xF901   /*Data register port address*/

#define	LEO_FC_WINDOW_SIZE	0x00400000 //4096K=4M
#define	LEO_FC_WINDOW_START	0xC0000000 //3072M=3G

#define INDEX_REG_PORT	(WYSE_MARS ? MARS_INDEX_REG_PORT : (WYSE_MIDWAY ? MIDWAY_INDEX_REG_PORT : LEO_INDEX_REG_PORT))
#define DATA_REG_PORT	(WYSE_MARS ? MARS_DATA_REG_PORT : (WYSE_MIDWAY ? MIDWAY_DATA_REG_PORT : LEO_DATA_REG_PORT))
#define FC_WINDOW_SIZE	(WYSE_MARS ? MARS_FC_WINDOW_SIZE : (WYSE_MIDWAY ? MIDWAY_FC_WINDOW_SIZE : LEO_FC_WINDOW_SIZE))
#define FC_WINDOW_START (WYSE_MARS ? MARS_FC_WINDOW_START : (WYSE_MIDWAY ? MIDWAY_FC_WINDOW_START : LEO_FC_WINDOW_START))
//#endif  /* WYSE_MARS or WYSE_LEO */

typedef volatile unsigned short FC_DATA;
#define	fc_setdata(off, value)	*((FC_DATA *) (off + fcard_vaddr)) = value
#define	fc_getdata(off)		*((FC_DATA *) (off + fcard_vaddr))

#define	FC_ROOTBLOCKNO	0

/* RYAN:
   Block size is decided by the PCMCIA card(hardware) we are using
   The card is AMD serial D compatible chip
   For example:It has 2 2Mbyte chips as a pair in 4MByte PC card 
   when do erase sector, it will erase 64K per chip and 128k for
   the pair of chips.
   So in MIDWAY, the WINDOW_SIZE(16K) is smaller than BLOCKSIZE(128k)
   The code need to be modified to support this case.
 */ 
#define FC_BLOCKSIZE	0x0020000	//128K for MARS,LEO and MIDWAY
#define	FC_PAGESIZE	    0x0000200	
#define	FC_BLKSHIFT	    8
#define	FC_BLKMASK	    0xFF

int	        fcard_vaddr;
u_short	    fcard_cksum;

STATIC	int	fc_window_offset;
STATIC	int	fc_card_page = -1;

#define	FC_PGOFST(pg)		((pg >> FC_BLKSHIFT) * FC_BLOCKSIZE + \
			         (pg & FC_BLKMASK) * (FC_PAGESIZE + 16))

#define	FC_PGOFST_B(pg)		FC_PGOFST(pg) + FC_PAGESIZE
				 
#ifdef WYSE_KDB
int fcard_ryan_test() 
{
	int i,j,k,blockno, windowno;
	unsigned char cmal, cmah;
	FC_DATA *data;
	int card_address = 0;
	int tmp_card_addr;
    unsigned short buf[256];
	unsigned short olddata, newdata;
    unsigned char  setdata;
	olddata = 0x2d2d;
	newdata = 0x4747;
	setdata = (newdata>>8) & 0xff;

	printf("sizeof buf %d\n", sizeof(buf));
	memset((char *) buf, setdata, sizeof(buf)); //fill newdata
	//asm("int3");
	
	//for 4M 32 128K blocks
	blockno = (4*1024*1024) / FC_BLOCKSIZE;
	windowno= FC_BLOCKSIZE / FC_WINDOW_SIZE;
	for (i=0; i<blockno; i++) {
	
		//1. read data on the card first
		tmp_card_addr = card_address;
		for (k=0;k<windowno;k++) {
			data = (FC_DATA*) fcard_vaddr;// Init dump data pointer
			fc_set_page(tmp_card_addr, 2); //set card offset
			for (j = 0; j < FC_WINDOW_SIZE; j += sizeof(*data), ++data) { 
				//if (*data == olddata)
				//	continue;
				if (j%2000 == 0)  {
					//printk("RYAN:old data:address 0x%x j 0x%x, bad data 0x%x\n",
					//	data, j, *data);
					hexdump(data, 0x10, 0x0000); 
				}
			}  //j
			tmp_card_addr +=FC_WINDOW_SIZE; //shift to next 16k card address
		} //k
        //asm("int3");
		/*
		// 2. erase the sector
		fc_set_page(card_address, 2);
		cmal = fc_getreg(CMA0l);
		cmah = fc_getreg(CMA0h);
		printf("RYAN:i %d CMA=0x%x 0x%x\n", i, cmah, cmal);

		// output command sequence to erase the block
		fc_cmd_seq(FC_RESET);
		fc_window_offset = 0x00000;
		//fc_cmd_seq(FC_SEGMENT_ERASE);
		fc_cmd_seq(FC_SECTOR_ERASE);
		udelay(10000);
		udelay(10000);
		udelay(10000);
		udelay(10000);
		fc_cmd_seq(FC_RESET);

		if (fc_test(0xFFFF)<0 ) {
			//return(-ENXIO);
			printf("RYAN: erase time out at %d block\n", i);
			while(1) {
				if (fc_getdata(fc_window_offset) == 0xffff)
					break;
			}
		}
		//3. read the block data back and dump it to see if it is all FFFF
		//printf("RYAN:begin dump from address 0xd4000 i %d\n",i);
		// read back 128k to see how it is being earsed
		tmp_card_addr = card_address;
		for (k=0;k<windowno;k++) {
			data = (FC_DATA*) fcard_vaddr;
			fc_set_page(tmp_card_addr, 2);
			//hexdump(data, 0x20, 0x0000);
			for (j = 0; j < FC_WINDOW_SIZE; j += sizeof(*data), ++data) { 
				if (*data == 0xFFFF)
					continue;
				if (j%2000 == 0)
					printk("RYAN:read erase data back: address 0x%x j 0x%x, bad data 0x%x\n",
					data, j, *data);
			}  //j
        	tmp_card_addr +=FC_WINDOW_SIZE; //shift to next 16k card address
		} //k
        //asm("int3");

		// 4. write block data
		tmp_card_addr = card_address;
		for (k=0;k<windowno;k++) {
			fc_set_page(tmp_card_addr, 2);
			fc_window_offset = 0;
			for (j=0;j<32;j++) {     //32 times 512 = 16K window 
				fc_output(buf, 512);  //count=512 in bytes
			}
			//hexdump(data, 0x20, 0x0  000);
		   	tmp_card_addr +=FC_WINDOW_SIZE; //shift to next 16k card address
		} //k

		// 5. read the write block data back
		tmp_card_addr = card_address;
		for (k=0;k<windowno;k++) {
			data = (FC_DATA*) fcard_vaddr;// Init dump data pointer
			fc_set_page(tmp_card_addr, 2); //set card offset
			for (j = 0; j < FC_WINDOW_SIZE; j += sizeof(*data), ++data) { 
				if (*data == newdata)
					continue;
				if (j%2000 == 0) {
					printk("RYAN:read write data back:address 0x%x j 0x%x, bad data 0x%x\n",
					data, j, *data);
				   hexdump(data, 0x10, 0x0000); 
				}
			}  //j
        	tmp_card_addr +=FC_WINDOW_SIZE; //shift to next 16k card address
		} //k
		*/
		asm("int3");
		
		card_address = tmp_card_addr; //more to next 128K blcok

    } //i
	asm("int3");
	
}
#endif

/*
 *	Static internal routines
*/
STATIC void 
fc_setreg(unsigned char regadd, unsigned char regdata)
{
	OUTB(INDEX_REG_PORT, regadd);
	OUTB(DATA_REG_PORT, regdata);
}

STATIC int 
fc_getreg(unsigned char regadd)
{
   	OUTB(INDEX_REG_PORT, regadd);
   	return(INB(DATA_REG_PORT));
}

void 
fc_cmd_seq(int operation)              
{
	switch (operation) {
	case FC_RESET:                             
		fc_setdata(0xaaa, 0xaaaa);
		fc_setdata(0x554, 0x5555);
		fc_setdata(0xaaa, 0xf0f0);
		break;
	case FC_AUTOSELECT:                                
		fc_setdata(0xaaa, 0xaaaa);
		fc_setdata(0x554, 0x5555);
		fc_setdata(0xaaa, 0x9090);
		break;
	case FC_WRITE:                             
		fc_setdata(0xaaa, 0xaaaa);
		fc_setdata(0x554, 0x5555);
		fc_setdata(0xaaa, 0xa0a0);
		break;
	case FC_SEGMENT_ERASE:
		fc_setdata(0xaaa, 0xaaaa);
		fc_setdata(0x554, 0x5555);
		fc_setdata(0xaaa, 0x8080);
		fc_setdata(0xaaa, 0xaaaa);
		fc_setdata(0x554, 0x5555);
		fc_setdata(0xaaa, 0x1010);
		break;
	case FC_SECTOR_ERASE:
		fc_setdata(0xaaa, 0xaaaa);
		fc_setdata(0x554, 0x5555);
		fc_setdata(0xaaa, 0x8080);
		fc_setdata(0xaaa, 0xaaaa);
		fc_setdata(0x554, 0x5555);
		fc_setdata(fc_window_offset, 0x3030);
		break;
	case FC_SECTOR_ERASE_SUSPEND:
		fc_setdata(fc_window_offset, 0xb0b0);
		break;          
	case FC_SECTOR_ERASE_RESUME:
		fc_setdata(fc_window_offset, 0x3030);
		break;
	}                                               
}

STATIC int 
fc_test(u_short write_data)
{
	int	i;

	for (i = 0; ; ++i) {
		if (i >= 0x600000) {
			fc_cmd_seq(FC_RESET);
			ASSERT(0);
			printf("fc_test: write 0x%x, read 0x%x\n",
				write_data, fc_getdata(fc_window_offset));
			break;
		}
		/* read data */
		if (fc_getdata(fc_window_offset) == write_data)
			return(0);
	}

	return(-ENXIO);
}

STATIC int 
fc_set_page(int offset, int page_cnfg)
{
	unsigned char card_addr_regl, card_addr_regh;
   	int 	card_addr_reg;

	card_addr_reg = (offset - FC_WINDOW_START) >> 12; 
	card_addr_regh = (card_addr_reg >> 8) & 0x3F;	/* top bits off */
	/* is it attribute memory? */
	if (0x1 & page_cnfg) 
		card_addr_regh |= 0x40; /* Yes */

	/* is it write protect? */
     	if (0x4 & page_cnfg) 
                card_addr_regh |= 0x80; /* Yes */

	card_addr_regl = (unsigned char) card_addr_reg;

	fc_setreg(CMA0l, card_addr_regl);
	fc_setreg(CMA0h, card_addr_regh);

	return(0);
}

#ifdef	LATER 
/* 	On MARS, window is 0x3900000 - 3AFF000 (size = 0x200000) */
int 
set_window(int window_base, int window_cnfg)
{
	unsigned char regdata;
	unsigne d char window_size = 0xf; /* (window_size+1)*4096 bytes */
	
	/* disable window   */
	fc_setreg(MAP_EN, 0x20);

	fc_setreg(SMA0sl, window_base);

	regdata = window_base >> 8;
	if (window_cnfg & 0x0002)	/* 16 bit access? */
		regdata |= 0x80;
	else
		regdata &= 0x7f;
	fc_setreg(SMA0sh, regdata); 

	/* system memory address lower stop byte */
	fc_setreg(SMA0el, window_base + window_size);

	/* set timing requirements as well as system memory address 
	   upper stop byte, d7=wait state bit 1
	   d6=wait state bit 0 */

	regdata = (window_base >> 8) | (0xc0 & (window_cnfg << 8));
	fc_setreg(SMA0eh, regdata);

	fc_setreg(MAP_EN, 0x21);
	
	return(0);
}
#endif

int 
fc_window(int offset, int memctrl)
{
	int	page;

	page = offset / FC_WINDOW_SIZE;
	if (page != fc_card_page) {
		fc_card_page = page;
		fc_set_page(page * FC_WINDOW_SIZE, memctrl);
	}
	fc_window_offset = offset - page * FC_WINDOW_SIZE;
	return(0);
}

STATIC void
fc_input(u_short *buf, int count)
{
	while(count > 0) { 
		*buf = fc_getdata(fc_window_offset);
		fcard_cksum += *buf++;
		fc_window_offset += sizeof(*buf);
		count -= sizeof(*buf);
	}
}

STATIC int 
fc_output(u_short *buf, int count)
{       
	u_short	s;

	while(count > 0) { 
		if (*buf != 0xFFFF) {
			s = fc_getdata(fc_window_offset);
			if (s != *buf) {
				fc_cmd_seq(FC_WRITE);
				fc_setdata(fc_window_offset, *buf);
				if (fc_test(*buf))
					return(-EIO);
			}
		}
		fcard_cksum += *buf++;
		fc_window_offset += sizeof(*buf);
		count -= sizeof(*buf);
	}
	return(0);
}

/* RYAN: 
 fcard_input: handles if the buffer size is out of the window boundary.
 We need it for MIDWAY platform
 */
STATIC int
fcard_input(int card_offset, u_short* bufptr, int buf_size) {
	int    access_size;

	while (buf_size > 0) {
		fc_window(card_offset, 2);	/* make sure it is mapped */
		access_size = FC_WINDOW_SIZE - fc_window_offset; /*remaning data in the window*/
		if (access_size > buf_size)
			access_size = buf_size;
		fc_input( bufptr, access_size); /* copy data in */

		card_offset += access_size;  //adjustments
		bufptr += access_size/2;
		buf_size -= access_size;
    }
	return (0);
}

/* RYAN:
 fcard_output: handles if the buffer size is out of the window boundary.
 We need it for MIDWAY platform
 */
STATIC int
fcard_output(int card_offset, u_short* bufptr, int buf_size) 
{
	int    access_size;

	while (buf_size > 0) {
		fc_window(card_offset, 2);	/* make sure it is mapped */
		access_size = FC_WINDOW_SIZE - fc_window_offset; /*remaning data in the window*/
		if (access_size > buf_size)
			access_size = buf_size;
		if (fc_output( bufptr, access_size)<0 )
			return (-EIO);
		card_offset += access_size;  //adjustments
		bufptr += access_size/2;
		buf_size -= access_size;
    }

	return (0);

}

STATIC int
fcard_readb(FLASHDEV *fd, int pageno, AREAB *b)
{
//#ifdef WYSE_MIDWAY
	int     card_offset, size;
    u_short *bufptr;

	ASSERT(pageno < fd->pages);

	/* init and calculate count */
	card_offset = FC_PGOFST_B(pageno);
	bufptr      = (u_short*) b;
	size        = sizeof(*b);
	
	fcard_input(card_offset, bufptr, size);

	return (0);
//#else 	/***LEO or MARS ***/
/*
	int	offset = FC_PGOFST_B(pageno);

	ASSERT(pageno < fd->pages);
	fc_window(offset, 2);	// make sure it is mapped 
	fc_input((u_short *) b, sizeof(*b));
	return(0);
#endif
*/
}

STATIC int
fcard_read(FLASHDEV *fd, int pageno, char *buf, AREAB *b)
{
//#ifdef WYSE_MIDWAY
	int     card_offset, size;
    u_short *bufptr;
	ASSERT(pageno < fd->pages);

	fcard_cksum = 0;   // init checksum

	/* init and calculate count for the buf */
	card_offset = FC_PGOFST(pageno);
	bufptr = (u_short*) buf;
	size = fd->pagesize;

	fcard_input(card_offset, bufptr, size);
	
	/* get rest of data to AREAB buffer */
	card_offset += size;
	bufptr      = (u_short*) b;
	size        = sizeof(*b);
	
	fcard_input(card_offset, bufptr, size);
	
	// compare checksum
	if (fcard_cksum != FLASH_CHECKSUM) {
		printf("fcard_read: cksum 0x%x on page 0x%x\n", 
			fcard_cksum, pageno);
		ASSERT(0);
		return(-EIO);
	}

	return (0);
//#else 	/***LEO or MARS ***/
/*
	int	offset = FC_PGOFST(pageno);

	ASSERT(pageno < fd->pages);
	fc_window(offset, 2);	// make sure it is mapped 
	fcard_cksum = 0;
	fc_input((u_short *) buf, fd->pagesize);
	fc_input((u_short *) b, sizeof(*b));
	if (fcard_cksum != FLASH_CHECKSUM) {
		printf("fcard_read: cksum 0x%x on page 0x%x\n", 
			fcard_cksum, pageno);
		ASSERT(0);
		return(-EIO);
	}
	return(0);
#endif
*/
}

STATIC int
fcard_writeb(FLASHDEV *fd, int pageno, AREAB *b)
{
//#ifdef WYSE_MIDWAY
	int     card_offset, size;
    u_short *bufptr;

	ASSERT(pageno < fd->pages);

	/* init and calculate count */
	card_offset = FC_PGOFST_B(pageno);
	bufptr      = (u_short*) b;
	size        = sizeof(*b);
	
	return (fcard_output(card_offset, bufptr, size));
//#else 	/***LEO or MARS ***/
/*
	int	offset = FC_PGOFST_B(pageno);

	ASSERT(pageno < fd->pages);
	fc_window(offset, 2);	// make sure it is mapped
	return(fc_output((u_short *) b, sizeof(*b)));
#endif
*/
}

int
fcard_write(FLASHDEV *fd, int pageno, char *buf, AREAB *b)
{
//#ifdef WYSE_MIDWAY
	int     card_offset, size, ret;
    u_short *bufptr;
	u_short tmpbuf[256];

	ASSERT(pageno < fd->pages);
	fcard_cksum = 0;   // init checksum

	/* init and calculate count for the buf */
	card_offset = FC_PGOFST(pageno);
	bufptr = (u_short*) buf;
	size = fd->pagesize;

	if (fcard_output(card_offset, bufptr, size)<0)
		return (-EIO);
	
	b->checksum = 0;
	fcard_cksum += flash_checksum((u_short *) b, (sizeof(*b) / 2));
	b->checksum = FLASH_CHECKSUM - fcard_cksum;
	
	/* get rest of data to AREAB buffer */
	card_offset += size;
	bufptr      = (u_short*) b;
	size        = sizeof(*b);
	
	ret = fcard_output(card_offset, bufptr, size);
#ifdef WYSE_KDB
	//fcard_input(card_offset, tmpbuf, 16);
	//hexdump((u_char*) tmpbuf, 16, 0);
#endif
	return ret;

//#else 	/***LEO or MARS ***/
/*
	int	offset = FC_PGOFST(pageno);

	ASSERT(pageno < fd->pages);
	fc_window(offset, 2);	// make sure it is mapped
	fcard_cksum = 0;
	if (fc_output((u_short *) buf, fd->pagesize) < 0)
		return(-EIO);
	b->checksum = 0;
	fcard_cksum += flash_checksum((u_short *) b, (sizeof(*b) / 2));
	b->checksum = FLASH_CHECKSUM - fcard_cksum;
	return(fc_output((u_short *) b, sizeof(*b)));
#endif
*/
}

int
fcard_erase(FLASHDEV *fd, int blockno, int verify)
{
//#ifdef WYSE_MIDWAY
	int     pageno, card_offset, buf_size;
	int     j, access_size;
    FC_DATA *data;

	pageno = blockno << FC_BLKSHIFT;
	card_offset = FC_PGOFST(pageno);
	ASSERT(pageno < fd->pages);
	ASSERT((card_offset % FC_BLOCKSIZE) == 0);
	
	/*1 erase the secotr */
	fc_cmd_seq(FC_RESET);
	fc_window(card_offset, 2);
	fc_cmd_seq(FC_SECTOR_ERASE);
#ifdef WYSE_KDB
	//printk("RYAN:earse fc_window_offset 0x%x should be 0\n", fc_window_offset);
#endif
	if (fc_test(0xFFFF) < 0)
		return(-ENXIO);
	fc_cmd_seq(FC_RESET);
	//2. read the block data back and dump it to see if it is all FFFF
	// read back 128k to see how it is being earsed
	buf_size = FC_BLOCKSIZE;
	while (buf_size > 0) {
		fc_window(card_offset, 2);	/* make sure it is mapped */
		access_size = FC_WINDOW_SIZE - fc_window_offset; /*remaning data in the window*/
		if (access_size > buf_size)
			access_size = buf_size;
		data = (FC_DATA*) (fcard_vaddr + fc_window_offset);
		for (j = 0; j < access_size; j += sizeof(*data), ++data) { 
			if (*data == 0xFFFF)
				continue;
			printk("fcard_erase:read erase data back: address 0x%x j 0x%x, bad data 0x%x\n",
				data, j, *data);
			ASSERT(0);
			return(-ENXIO);
		}  //j
        card_offset += access_size;  //adjustments
		buf_size -= access_size;
    }
    return (0);
//#else 	/***LEO or MARS ***/
/*
	int	pageno = blockno << FC_BLKSHIFT;
	int	offset = FC_PGOFST(pageno);
	int	i;
	FC_DATA *p;

	ASSERT(pageno < fd->pages);
	ASSERT((offset % FC_BLOCKSIZE) == 0);
	fc_cmd_seq(FC_RESET);
	fc_window(offset, 2);
	p = (FC_DATA *) (fc_window_offset + fcard_vaddr);
	fc_cmd_seq(FC_SECTOR_ERASE);
*/
	/* According to datasheet, timeout needed here is 100us */
/*
	if (fc_test(0xFFFF) < 0)
		return(-ENXIO);

	for (i = 0; i < FC_BLOCKSIZE; i += sizeof(*p), ++p) { 
		if (*p == 0xFFFF)
			continue;
		printk("fcard_erase:fc_window_offset 0x%x i 0x%x, block 0x%x, offset 0x%x, bad data 0x%x\n",
			fc_window_offset, i, blockno, offset, *p);
		ASSERT(0);
		return(-ENXIO);
	}
	
	return(0);
#endif
*/
}

void
fc_cis(void)
{
#ifdef	LATER
	u_short	buf[32];

	/* this is how to read cis, but dont know what struct is like */
	fc_set_page(0, 0x01);	/* set to attribute memory */
	fc_input(buf, sizeof(buf));
	fc_set_page(0, 0x02);	/* back to card memory */
#endif
}

int 
fc_cardinfo(void)
{
	u_short	mfg_code, dev_code;

	/* These additional resets are apparently necessary to
	   read the AMD flash card. */
    if (WYSE_MIDWAY)
		fc_cmd_seq(FC_RESET);  	/* resets before autoselect */
	else if (WYSE_LEO) {
		fc_setup();
		udelay(10000);
		udelay(10000);
		fc_cmd_seq(FC_RESET);  	/* resets before autoselect */
		udelay(10000);
		udelay(10000);
	} else {
		fc_cmd_seq(FC_RESET);  	/* resets before autoselect */
	}
	fc_cis();		    /* read card info struct */
	fc_window(0, 0x02);	/* set to card memory */
	fc_cmd_seq(FC_AUTOSELECT);
	mfg_code = fc_getdata(0);
	dev_code = fc_getdata(2);
#ifdef WYSE_KDB
	printf("RYAN:fc_cardinfo:mfg code 0x%x, dev code 0x%x\n", mfg_code, dev_code);
#endif
	fc_cmd_seq(FC_RESET); 
	if (mfg_code != 0x0101 ||
	    (dev_code != 0x3d3d && dev_code != 0xadad)) {
		printf("mfg code 0x%x, dev code 0x%x\n", mfg_code, dev_code);
		return(-ENXIO);
	}
	return(0);
}

/*
 *	Useful general callable routines from KDB or block device
*/

int
fcard_present(void)
{
	return((fc_getreg(1) & 0x0C) == 0x0C);
} 

int
fcard_ready(void)
{
	return(fc_getreg(1) & 0x20);
}

#define	FC_AUTOSIZE	32

STATIC int
fc_autosize(FLASHDEV *fd)
{
	static	char	areab[16 + 1] = "xxxxwyseWYSEyyyy";
	AREAB	tmp;
	short	*s1, *s2;
	int	i, j, blockno, pageno;

	s1 = (short *) &areab[0];
	s2 = (short *) &areab[14];

	/* info needed by internal primitives */
	/* detect upto 32mb card */
	i  = FC_AUTOSIZE * 1024 * 1024;
	fd->ppb = FC_BLOCKSIZE / (FC_PAGESIZE + sizeof(AREAB));
	fd->blocks = i / FC_BLOCKSIZE;
	fd->pages = i / FC_PAGESIZE;
	fd->blocksize = FC_BLOCKSIZE;
	fd->pagesize = FC_PAGESIZE;

	for (i = 4; i < FC_AUTOSIZE + 4; i += 4) {
		blockno = (i * 0x100000 / FC_BLOCKSIZE) - 1;
		pageno  = (blockno << FC_BLKSHIFT) + 1;
		printf("fc_autosize: testing %dMB bytes on blockno=0x%0x\n", i, blockno);
		if (fcard_erase(fd, blockno, 0) < 0) {
			printf("fc_autosize: breaking out on erase blockno=0x%x\n",blockno);
			break;
		}
		*s1 = blockno;
		*s2 = pageno;
		if (fcard_writeb(fd, pageno, (AREAB *) areab) < 0) {
			printf("fc_autosize: breaking out on writeb\n");
			break;
		}
		
		for (j = 4; j <= i; j += 4) {
			blockno = (j * 0x100000 / FC_BLOCKSIZE) - 1;
			pageno  = (blockno << FC_BLKSHIFT) + 1;
			*s1 = blockno;
			*s2 = pageno;
			memset((char *) &tmp, 0x5a, sizeof(tmp));
			fcard_readb(fd, pageno, &tmp);
			if (memcmp((char *) &tmp, areab, sizeof(tmp)) != 0) {
				printf("fc_autosize: break out on compare\n");
				break;
			}
		}
		if (j <= i) {
			PRINTF("fc_autosize: stop at %dMB\n", j = 4);
			break;
		}
	}
	i -= 4;
	printf("fc_autosize: %dMB bytes\n", i);
	fd->blocks = 0;		// prevent EBUSY from fcard_open
	return(i * 0x100000);
}

int 
fcard_open(struct inode * inode, struct file * filp)
{
	int	minor, autosizing, size = 0;
	int	blocks, pages;
	int	i, blockno, pageno;
	AREAB	areab;
	FLASHDEV *fd;
#ifdef WYSE_KDB
	unsigned char smsl, smsh, smel, smeh;
	printf("RYAN:fcard_open\n");
#endif

	ASSERT(inode);
	minor = FLASH_MINOR(inode->i_rdev);
	if (minor >= NFCARDS)
		return(-ENODEV);
	fd = FCARD_DEV0 + minor;
	if (fd->blocks != 0)
		return(-EBUSY);

	if (!fcard_present()) {
		printf("fcard_open(): no card detected\n");
		return(-ENXIO);
	}

	/* Test for Wyse Nand Card first */
	fd->dev	= inode->i_rdev;
	fd->port = fcard_vaddr + minor * 0x4000;
	if (fncard_open(fd) >= 0)
		return(0);

	fd->dev = 0;
	if (minor > 0)
		return(-ENODEV);
	if (fc_cardinfo() < 0) {
		printf("fcard_open(): unknown type of card\n");
		return(-ENXIO);
	}

	fd->read 	= fcard_read;	/* export 5 functions */
	fd->readb	= fcard_readb;
	fd->write	= fcard_write;
	fd->writeb	= fcard_writeb;
	fd->erase	= fcard_erase;  

	fd->rootblockno	= FC_ROOTBLOCKNO;
	fd->rootpageno	= FC_ROOTBLOCKNO * (FC_BLOCKSIZE / FC_PAGESIZE); 

	/* Auto sizing flash card */
	blockno = FC_ROOTBLOCKNO; 
	fd->pages = (blockno + 8) << FC_BLKSHIFT;
	for (autosizing = i = 0; i < 8; ++i, ++blockno) {
		pageno = blockno << FC_BLKSHIFT;
#ifdef WYSE_KDB
	printf("RYAN:fcard_open:blockno=0x%x, pageno=0x%x, fd->pages=%d\n",blockno, pageno, fd->pages);
#endif
		if (fcard_readb(fd, pageno, &areab) != 0)
			continue;
		if (areab.ino != FLASH_ROOTINO - 1 ||
		    areab.firstno != pageno ||
		    areab.sequence != FLASH_ROOTINO) {
			/* unformatted card found */
#ifdef WYSE_KDB
	printf("RYAN:fcard_open:unformatted card found\n");
#endif
			autosizing = 1;
			break;
		}
		size = areab.size * 0x100000;
		if (size < 0x400000 || size > 0x2000000) {
			/* bad sized card */
			//PRINTF("fcard_open: size 0x%x\n", size);
			autosizing = 1;
		}
		break;
	}
#ifdef WYSE_KDB
	//RYAN: read the start and end reg back
	smsl = fc_getreg(SMA0sl);
    smsh = fc_getreg(SMA0sh);
    smel = fc_getreg(SMA0el);
    smeh = fc_getreg(SMA0eh);
	printf("RYAN:fcard_open:sl=0x%x,sh=0x%x,el=0x%x,eh=0x%x\n", smsl,smsh,smel,smeh);
	//fcard_ryan_test();
#endif

	if (autosizing) {
		if ((inode->i_rdev & (FLASH_FORMAT | FLASH_AUTOFMT)) == 0)
			return -ENXIO;
		size = fc_autosize(fd);
		if (size <= 0)
			return(-ENXIO);
	}

	fd->dev	= inode->i_rdev;
	blocks = size / FC_BLOCKSIZE;
	pages = size / FC_PAGESIZE;
#ifdef WYSE_KDB
	printf("RYAN:fcard_open:blocks=%d, pages=%d\n",blocks, pages);
#endif
	return(flash_init(fd, blocks, pages, FC_BLOCKSIZE, FC_PAGESIZE)); 
}

#ifdef	__KERNEL__

extern	char	fd_tmpbuf[];	/* borrow from flash.c */

STATIC void 
fcard_close(struct inode *inode, struct file *file)
{
	sync_dev(inode->i_rdev);
	flash_close(FCARD_DEV0 + FLASH_MINOR(inode->i_rdev));
}

int 
fcard_blkread(struct inode *inode, struct file *filp, char *buf, int count)
{
	int	c;
	int	off;
	int	offset = filp->f_pos;
	char	*b0 = buf;
	FLASHDEV *fd = FCARD_DEV0 + FLASH_MINOR(inode->i_rdev);

	/* To read flash card, count must be in even number for short read */
	ASSERT((offset & 1) == 0);
	ASSERT((count & 1) == 0);
	if (offset + count > fd->rawsize) 
		count = fd->rawsize - offset;
	while (count > 0) {
		off = offset % FC_WINDOW_SIZE;
		c = FC_WINDOW_SIZE - off;	/* remaining data in a window */
		if (c > count) 
			c = count;
		if (c > FFS_SECTSIZE)
			c = FFS_SECTSIZE;
		fc_window(offset, 2);	/* make sure it is mapped */
		fc_input((u_short *) fd_tmpbuf, c);	/* copy data in */
		memcpy_tofs(buf, fd_tmpbuf, c);		/* copy to user space */
		count -= c;
		buf += c;
		offset += c;
	}
	filp->f_pos = offset;
	return(buf - b0);
}

int 
fcard_blkwrite(struct inode *inode, struct file *filp,
	const char *buf, int count)
{
	int	c;
	int	off;
	int	offset = filp->f_pos;
	char	*b0 = (char *) buf;
	FLASHDEV *fd = FCARD_DEV0 + FLASH_MINOR(inode->i_rdev);

	ASSERT((count & 1) == 0);
	ASSERT((offset & 1) == 0);
	if (offset + count > fd->rawsize) 
		count = fd->rawsize - offset;
	while (count > 0) {
		off = offset % FC_WINDOW_SIZE;
		c = FC_WINDOW_SIZE - off;	/* remaining data in a window */
		if (c > count) 
			c = count;
		if (c > FFS_SECTSIZE)
			c = FFS_SECTSIZE;
		fc_window(offset, 2);	/* make sure it is mapped */
		memcpy_fromfs(fd_tmpbuf, (char *) buf, c);
		if (fc_output((u_short *) fd_tmpbuf, c) < 0)
			break;
		count -= c;
		buf += c;
		offset += c;
	}
	filp->f_pos = offset;
	return(buf - b0);
}

STATIC int 
fcard_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int 	err;
	FLASHDEV *fd = FCARD_DEV0 + FLASH_MINOR(inode->i_rdev);
	
	if (!inode || !inode->i_rdev) 	
		return -EINVAL;

	switch (cmd) {
	case BLKFLSBUF:
		if (!suser()) 
			return -EACCES;
		invalidate_buffers(inode->i_rdev);
		break;
	case BLKGETSIZE:   /* Return device size */
		if (!arg)  
			return -EINVAL;
		err = verify_area(VERIFY_WRITE, (long *) arg, sizeof(long));
		if (err)
			return err;
		put_user(fd->rawsize / 512, (long *) arg);
		return 0;
	}

	return 0;
}

static struct file_operations fc_fops = {
	NULL,		/* lseek - default */
	fcard_blkread,	/* read - block dev read */
	fcard_blkwrite,	/* write - block dev write */
	NULL,		/* readdir - not here! */
	NULL,		/* select */
	fcard_ioctl, 	/* ioctl */
	NULL,		/* mmap */
	fcard_open,	/* open */
	fcard_close,	/* release */
	block_fsync	/* fsync */ 
};

/* This is the registration and initialization section of the card driver */
STATIC int 
fcard_linux_init()
{
	if (register_blkdev(MAJOR_NR, "fcard", &fc_fops)) {
		printf("fcard: Could not get major %d", MAJOR_NR);
		return -EIO;
	}
#ifdef WYSE_KDB
	printf("RYAN:fcard_linux_init register_blkdev successful\n");
    //asm("int3");
#endif
	printf("FlashCard driver initialized - 1 Card slot\n");

	return 0;
}
#endif

void
fc_setup(void)
{
	unsigned char x, *p;

    if (WYSE_MIDWAY) {
		fc_setreg(Glob_Ctrl, 0x0a);   /* 5 volt core*/		
		fc_setreg(POWER,     0xB0);   /* Auto Power Enable */
		fc_setreg(INT_CTRL,  0x00);   /* Reset & IRQ disabled */
		fc_setreg(INT_CTRL,  0x60);   /* Reset is set inactive */

		/* system memory address start at 0xD0000 */
		fc_setreg(SMA0sl,    0xd4);   //D4000
		fc_setreg(SMA0sh,    0x80); 
		/* system memory address stop at  0xD7FFF as 16K window*/
		fc_setreg(SMA0el,    0xd7);   //D7FFF (16K)
		fc_setreg(SMA0eh,    0x00);

		fc_setreg(DELTA_STAT,0x00);
		fc_setreg(INT_CNFG,  0x00);
		fc_setreg(MAP_EN,    0x01);   /* Enable MEM window 0 */
		fc_setreg(IO_CTRL,   0x00);   /* I/O timing & sizing */
    } else if (WYSE_LEO) {
		p = (unsigned char *) 0xD0800;
		p[0x06] = 0x01;		// enable memory windows
		p[0x03] = 0x40;		// 0x40 mem card, 60 = io card
		p[0x10] = 0x00;		
		p[0x40] = FC_WINDOW_START >> 24;
		p[0x11] = 0x80;		// 16-bit memory
		p[0x12] = 0xff;		// card memory window limit
		p[0x13] = 0x03;		// card memory window limit
		x = p[0x16];		// general purpose reg
		x |= 0x40;		// deassert reset
		x &= 0xdf;		// mem card installed
		p[0x16] = x;
		p[0x03] = 0x60;
		p[0x02] = 0x91;		// card power control
	} else {
		fc_setreg(POWER, 0xBA);		/* Auto Power Enable */
		fc_setreg(INT_CTRL, 0x40);	/* Reset & IRQ disabled */
		fc_setreg(DELTA_STAT, 0x00);
		fc_setreg(INT_CNFG, 0x00);
		fc_setreg(MAP_EN, 0x21);	/* Enable MEM window 0 */
		fc_setreg(IO_CTRL, 0x00);	/* I/O timing & sizing */
	}
	udelay(20000);
	fc_cmd_seq(FC_RESET);
	udelay(20000);
}

int 
fcard_init(void)
{
	unsigned char id;

#ifdef WYSE_KDB
	unsigned char smsl, smsh, smel, smeh;
	FC_DATA *data;
	printf("RYAN:fcard_init\n");
#endif
    if (WYSE_MIDWAY) {
	/*RYAN:
	  Based on arch-i386-mm-init.c: Address 0xA0000 to 0x100000(1M)
	  is reserved with PG_reserved set. so I use 0xD4000 to 0xD7FFF (16K) 
	  area for the card system memory window.
     */
		fcard_vaddr = (int) FC_WINDOW_START;
	}
	else {
		fcard_vaddr = (int) vremap(FC_WINDOW_START, FC_WINDOW_SIZE);
		if (fcard_vaddr == 0) {
			printf("fcard_init: vremap failed\n");
			asm("int3");
			return(-ENOMEM);
		}
	}
#ifdef WYSE_KDB
	printf("RYAN:fcard_init fcard_vaddr=0x%x\n", fcard_vaddr);
#endif

	switch(id = fc_getreg(REVID)) {
	case ID1:
	case ID2:
	case ID:
		break;
	default:
		printf("fcard_init: unknown FlashCard ID 0x%x\n", id);
		return(-ENXIO);
	}
#ifdef WYSE_KDB
	printf("RYAN:fcard_init:after get id=%d\n", id);
#endif

	fc_setup();
#ifdef WYSE_KDB
	//RYAN: read the start and end reg back
	data = (FC_DATA*) fcard_vaddr;

	smsl = fc_getreg(SMA0sl);
    smsh = fc_getreg(SMA0sh);
    smel = fc_getreg(SMA0el);
    smeh = fc_getreg(SMA0eh);
	printf("RYAN:fcard_init:sl=0x%x,sh=0x%x,el=0x%x,eh=0x%x\n", smsl,smsh,smel,smeh);
	//fcard_ryan_test();
	
#endif

#ifdef	LATER
	set_window(0x3900, 0x603);	/* init by MARS boot */
#endif

#ifdef	__KERNEL__
	return(fcard_linux_init());	/* register device */
#else
	return(0);
#endif
}

