#ifdef	__KERNEL__
#include <linux/config.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/malloc.h>
#include <linux/ioctl.h>
#include <linux/stat.h>
#include <linux/flash_fs.h>
#include <linux/flash.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>

#define	malloc(size)	kmalloc(size, GFP_KERNEL)  
#define	free(ptr)	kfree(ptr)
typedef volatile unsigned char vubyte;
#include <linux/blk.h>

#else

#include "system.h"
#include "flash_fs.h"
#include "flash.h"
#include "errno.h"

#endif

char	fd_tmpbuf[FFS_SECTSIZE + FLASH_RDNTSIZE];	/* shared */
char	fd_gcbuf[FFS_SECTSIZE + FLASH_RDNTSIZE];	/* shared */
STATIC	struct	ffs_buf	fd_bp;

int uncompress(char *, int *, char *, int);
int compress(char *, int *, char *, int);

FLASHDEV flashs[NFLASHS];

#define	FD_READ			(*fd->read)
#define	FD_READB		(*fd->readb)
#define	FD_WRITE		(*fd->write)
#define	FD_WRITEB		(*fd->writeb)
#define	FD_ERASE		(*fd->erase)
#define	FD_ALLOC		(*fd->alloc)
#define	FD_FREE			(*fd->free)
#define	FD_DELETE		(*fd->delete)

#define	FREEMAP_SET(fd, i)	fd->freemap[i/8] |=  (1 << (i & 0x7))
#define	FREEMAP_CLR(fd, i)	fd->freemap[i/8] &= ~(1 << (i & 0x7))
#define	FREEMAP_TST(fd, i)     (fd->freemap[i/8] &   (1 << (i & 0x7)))

#define	FD_MINFREE		32

void
flash_erase(FLASHDEV *fd)
{
#ifndef	EPROM
	int	blockno, i = 0;

	for (blockno = FLASH_ROOTBLKNO; blockno < fd->blocks; ++blockno) {
		if (FD_ERASE(fd, blockno, 0) == 0)
			continue;
		if (i < (fd->pagesize / sizeof(flash_sect_t)) - 1)
			fd->badblocks[++i] = blockno;
	}
	fd->badblocks[0] = i;
#endif
}

STATIC void
fd_rebuild(FLASHDEV *fd)
{
	int	blockno, pageno;
	AREAB	areab;

#ifndef	EPROM
	for (blockno = FLASH_ROOTBLKNO; blockno < fd->blkend; ++blockno) {
		pageno = blockno << fd->blkshift;
		if (FD_READB(fd, pageno, &areab) == 0) {
			ASSERT(areab.ino == 0xFFFF);
			// areab.inuse = 1;	/* version 1: badblock */
			areab.inuse = 2;	/* version 2: nand sum */
			areab.ino = FLASH_ROOTINO - 1;
			areab.sectno = 0;
			areab.firstno = pageno;
			areab.pageno = pageno;
			areab.nextno = FLASH_INODES;
			areab.size = (fd->blocks * fd->blocksize) / 0x100000;
			areab.sequence = FLASH_ROOTINO;

			FD_WRITE(fd, pageno, (char *) fd->badblocks, &areab);
			return;
		}
	}
#endif
}

STATIC int
fd_badblock(FLASHDEV *fd, int blockno)
{
	int	i, badcount;

	badcount = fd->badblocks[0];
	for (i = 0; i < badcount; ) {
		if (fd->badblocks[++i] == blockno)
			return(1);
	}
	return(0);
}

STATIC int
fd_freecnt(FLASHDEV *fd, int blockno)
{
	int	i, count = 0;
	int	pageno;

	pageno = blockno << fd->blkshift;
	for (i = fd->ppb; i > 0; --i, ++pageno) {
		if (FREEMAP_TST(fd, pageno))
			++count;
	}
	return(count);
}

STATIC int
fd_gc_block(FLASHDEV *fd, int blockno)
{
#ifndef	EPROM
	int	i, pageno, inuse_cnt = 0, deleted_cnt, rcpn = fd->pgend, size;
	AREAB	areab;
	static	u_short	inuse_list[256];	

	deleted_cnt = fd->blkdelcnt[blockno];

	/* Stage 1 - Mark blockno to be reclaimed */
	if (FD_ERASE(fd, fd->blkend, 0) < 0)
		return(0);
	memset(&areab, 0xFF, sizeof(areab));
	areab.checksum = blockno;
	areab.sectno   = blockno;
	areab.ino      = blockno;
	areab.size     = size = fd->ppb - deleted_cnt - fd_freecnt(fd, blockno);
	ASSERT(areab.size > 0 && areab.size < fd->ppb);
	if (FD_WRITEB(fd, rcpn, &areab) != 0)
		return(0);

 	pageno = blockno << fd->blkshift;	/* first page in block */
	for (i = 0; i < fd->ppb; ++i, ++pageno) {
		if (FREEMAP_TST(fd, pageno))	/* skip free page */
			continue;
		if (FD_READB(fd, pageno, &areab) != 0)
			continue;
		if (areab.ino == 0x0000) {	/* deleted will be freed */
			fd->freepages++;
			fd->deletepages--;
			FREEMAP_SET(fd, pageno);
			continue;
		}
		ASSERT(inuse_cnt < sizeof(inuse_list) / sizeof(inuse_list[0]));
		++rcpn;
		if (areab.ino != 0xFFFF) {	/* allocated and written */
			inuse_list[inuse_cnt++] = pageno;
			FD_READ(fd, pageno, fd_gcbuf, &areab);
			FD_WRITE(fd, rcpn, fd_gcbuf, &areab);
		} else {
			inuse_list[inuse_cnt++] = 0xFFFF;	/* skip */
		}
	} 	
	ASSERT(inuse_cnt == size);

	/* Stage 2 - Mark blockno to be reclaim is all copied */
	rcpn = fd->pgend;
	memset(&areab, 0xFF, sizeof(areab));
	areab.checksum = blockno;
	areab.ino      = 0;
	FD_WRITEB(fd, rcpn, &areab);

	FD_ERASE(fd, blockno, 0);	/* now erase the block */
	 
	for (i = 0; i < inuse_cnt; ++i) {
		pageno = inuse_list[i];
		++rcpn;
		if (pageno != 0xFFFF) {
			FD_READ(fd, rcpn, fd_gcbuf, &areab);
			FD_WRITE(fd, pageno, fd_gcbuf, &areab);
		}
	} 	
	
	/* Stage 3: erase the reclaimed block and back to Stage 0 */
	FD_READB(fd, fd->pgend, &areab);
	areab.size = 0;
	if (FD_WRITEB(fd, fd->pgend, &areab) < 0)
		FD_ERASE(fd, fd->blkend, 0);

	fd->blkdelcnt[blockno] = 0;
	return(deleted_cnt);
#else
	return(0);
#endif
}

STATIC int
fd_gc(FLASHDEV *fd)
{
	int	j = 0, blockno;

	for (blockno = FLASH_ROOTBLKNO + 1; blockno < fd->blkend; ++blockno) {
		if (fd->blkdelcnt[blockno] > (fd->ppb / 4)) {
			j += fd_gc_block(fd, blockno);
			if (j >= fd->ppb && fd->freepages > FD_MINFREE)
				return(j);
		}
	}
	for (blockno = FLASH_ROOTBLKNO + 1; blockno < fd->blkend; ++blockno) {
		if (fd->blkdelcnt[blockno] > (fd->ppb / 8)) {
			j += fd_gc_block(fd, blockno);
			if (j >= fd->ppb && fd->freepages > FD_MINFREE)
				return(j);
		}
	}
	for (blockno = FLASH_ROOTBLKNO + 1; blockno < fd->blkend; ++blockno) {
		if (fd->blkdelcnt[blockno] > 0) {
			j += fd_gc_block(fd, blockno);
			if (j >= fd->ppb && fd->freepages > FD_MINFREE)
				return(j);
		}
	}
	return(j);
}

STATIC void
fd_gc_recover(FLASHDEV *fd)
{
#ifndef	EPROM
	int	i, pageno, blockno;
	AREAB	areab;

	if (FD_READB(fd, fd->pgend, &areab) != 0)
		return;

	if (areab.ino == 0xFFFF || areab.checksum == 0xFFFF)
		return;		/* Stage 0 or 3: reclaim block empty */
	if (areab.size == 0 || areab.size >= fd->ppb)
		return;
	if (areab.ino != 0x0000) {
		ASSERT(areab.ino == areab.checksum);
		return;		/* Stage 1: not trashed yet */ 
	}
	ASSERT(areab.sectno == areab.checksum);
	if (areab.sectno != areab.checksum)
		return;

	/*
	 * At this point, the reclaim block has good pages.
	 * We must have crashed during fd_gc() after a 
	 * to-be-reclaimed block has copied to the reclaim block, and
	 * before the fd_gc() completed with its work.
	 * Now, trashed the block and make a copy again.
	*/
	blockno = areab.sectno;
	ASSERT(blockno > FLASH_ROOTBLKNO && blockno < fd->blkend);
	if (blockno <= FLASH_ROOTBLKNO || blockno >= fd->blkend)
		return;

	FD_ERASE(fd, blockno, 0);

	for (i = areab.size, pageno = fd->pgend + 1; i > 0; --i, ++pageno) {
		if (FD_READB(fd, pageno, &areab) != 0)
			continue;
		ASSERT(areab.ino != 0);
		if (areab.ino == 0xFFFF)
			continue;
		FD_READ(fd, pageno, fd_gcbuf, &areab);
		ASSERT((areab.pageno >> fd->blkshift) == blockno);
		FD_WRITE(fd, areab.pageno, fd_gcbuf, &areab); 
	}

	FD_ERASE(fd, fd->blkend, 0);
#endif
}

/*
**	Returns a pageno, clear the bit in the freemap
*/
STATIC int
fd_alloc(FLASHDEV *fd)
{
	int	i, j, pageno, startindx;

	startindx = i = fd->freeindx;
	for ( ;; ) {
		if (fd->freemap[i] != 0) {
			fd->freeindx = i;
			pageno = i * 8;
			for (j = 0; j < 8; ++j, ++pageno) {
				if (fd->freemap[i] & (1 << j)) {
					fd->freemap[i] &= ~(1 << j);
					fd->freepages--;
					return(pageno);
				}
			}
		}
		if (++i >= fd->pgend / 8)
			i = FLASH_ROOTPAGENO / 8;
		if (i != startindx)
			continue;
		if (fd_gc(fd) == 0) {
			printf("fd_alloc: no space\n");
			return(0xFFFF);
		}
	}
}

void
fd_free(FLASHDEV *fd, int pageno)
{
	ASSERT(pageno < fd->pages);
	ASSERT(FREEMAP_TST(fd, pageno) == 0);
	FREEMAP_SET(fd, pageno);
	fd->freepages++;
}

u_short
flash_checksum(u_short *buf, int size)
{
	u_short	sum = 0;

	while(--size >= 0)
		sum += *buf++;
	return(sum);
}

void
flash_free(FLASHDEV *fd, int pageno)
{
	FD_FREE(fd, pageno);
} 

STATIC void
fd_delete(FLASHDEV *fd, int pageno)
{
	int	blockno, i, error;

	blockno = pageno >> fd->blkshift;
	ASSERT(blockno < fd->blocks);
	fd->deletepages++;
#ifndef	EPROM
	fd->blkdelcnt[blockno]++;
	/* if it is the last deleted page in a block, erase and free the blk */
	if (fd->blkdelcnt[blockno] + fd_freecnt(fd, blockno) >= fd->ppb) {
		error = FD_ERASE(fd, blockno, 0);
		if (error == 0) {
			fd->deletepages -= fd->blkdelcnt[blockno];
			fd->blkdelcnt[blockno] = 0;
			pageno = blockno << fd->blkshift;
			for (i = 0; i < fd->ppb; ++i, ++pageno) {
				if (FREEMAP_TST(fd, pageno) == 0)
					FD_FREE(fd, pageno);
			}
		}
	}
#endif
}

void
flash_imap_expand(struct flash_imap *ip, int sectno)
{
	int	oldsize = ip->mapsize, size, newsize;
	flash_sect_t	*newmap;

	/* double the map size each time */
	
	newsize = oldsize;
	while(newsize <= sectno)
		newsize <<= 1;
	ip->mapsize = newsize;
	size = newsize * sizeof(flash_sect_t);
	newmap = (flash_sect_t *) malloc(size); 
	ASSERT(newmap);
	if (newmap) {
		oldsize *= sizeof(flash_sect_t);
		memcpy(newmap, ip->map, oldsize);
		memset(((char *) newmap) + oldsize, 0, size - oldsize);
		free(ip->map);
		ip->map = newmap;
	}
}

int
flash_mount(kdev_t dev)
{
	int	i, j, k, pn, blockno, pageno = 0, size, error;
	AREAB	areab;
	struct 	flash_imap *ip;
	DIRENT	*ep;
	FLASHDEV *fd;
	int	chunkno, inc;

	for (i = 0, fd = flashs; i < NFLASHS; ++i, ++fd) {
		if (fd->dev == dev)
			break;
	}
	if (i >= NFLASHS)
		return(-ENXIO);

	inc = startbar_nitems - startbar_item;
	if (inc <= 0)
		inc = 1;
	inc = chunkno = fd->blkend / inc;

	if (dev == ROOT_DEV) {
		STARTBAR("Kernel: mounting device %s as root ...",
			kdevname(dev));
	}
	
	/* Allocate 256 inodes for every 2MB of space */
	FLASH_INODES = 8 * (fd->blocks * fd->blocksize / 0x10000);

	/*
	 *  Zap the devcie if so requested on the minor device number.
	*/
	if (fd->dev & FLASH_FORMAT) {
		fd_rebuild(fd);
	}
	
	/*
	 *  Pass 0: find super block (inode number == ROOTINODE - 1)
	*/
	for (blockno = FLASH_ROOTBLKNO; blockno < fd->blkend; ++blockno) {
		pageno = blockno << fd->blkshift;
		error = FD_READB(fd, pageno, &areab);
		if (error != 0)
			continue;
		if (areab.ino == FLASH_ROOTINO - 1 &&
		    areab.sequence == FLASH_ROOTINO) {
			if (areab.inuse >= 1)	/* badblock supported */
				FD_READ(fd, pageno, (char *) fd->badblocks, &areab);
			break;
		}
	}

	if (blockno >= fd->blkend) {
		printf("flash_mount: pageno = 0x%x, ino = 0x%x, magic = 0x%x\n",
			pageno, areab.ino, areab.sequence);
		free(fd->badblocks);
		memset(fd, 0, sizeof(*fd));
		return(-ENXIO);
	}

	FLASH_ROOTBLKNO = blockno;
	FLASH_INODES = areab.nextno;	/* actual inodes in the device */
	fd->freeinodes = FLASH_INODES - FLASH_ROOTINO;
	fd->dirinodes	= FLASH_INODES / 8;

	i = fd->blocks + fd->pages / 8 + sizeof(fd->inodes[0]) * FLASH_INODES;
	ASSERT(fd->blkdelcnt == 0);
	fd->blkdelcnt = (unsigned char *) malloc(i);
	if (fd->blkdelcnt == 0)
		return(-ENOMEM);
	memset((char *) fd->blkdelcnt, 0, i);

	fd->freemap = fd->blkdelcnt + fd->blocks;
	fd->inodes = (struct flash_imap *) (fd->freemap + fd->pages / 8);

#ifndef	EPROM
	if ((fd->dev & FLASH_FORMAT) == 0) {
		fd_gc_recover(fd);	/* in case we crashed during gc() */
		if (dev == ROOT_DEV)
			STARTBAR("Kernel: mounting device %s as root ... ",
				kdevname(dev));
	}
#endif

	fd->inoindx = FLASH_DIRINO;
#ifdef	__KERNEL__
	for (i = jiffies & 0xffffff; ; i >>= 1) {
		fd->freeindx = FLASH_ROOTPAGENO / 8 + i;
		if (fd->freeindx < fd->pgend / 8)
			break;
	}
#else
	fd->freeindx = FLASH_ROOTPAGENO / 8;
#endif

	/*
	 *  Pass 1: build index block in memory
	*/
	for (blockno = FLASH_ROOTBLKNO; blockno < fd->blkend; ++blockno) {
		if (fd_badblock(fd, blockno)) {
			fd->badpages += fd->ppb;
			continue;
		}
		pageno = blockno << fd->blkshift;
		for (k = 0; k < fd->ppb; ++pageno, ++k) {

			error = FD_READB(fd, pageno, &areab);
			if (error != 0) {
				fd->badpages++;
				continue;
			}

			if (areab.ino == 0xFFFF) {
				ASSERT(areab.sectno == 0xFFFF);
				ASSERT(areab.nextno == 0xFFFF);
				FD_FREE(fd, pageno);
				continue;
			}
			
			if (areab.ino == 0x0000) {
				FD_DELETE(fd, pageno);
				continue;
			}

			if (areab.ino == FLASH_ROOTINO - 1)
				continue;

			ASSERT(areab.ino < FLASH_INODES);
			ip = flash_getimap(fd, areab.ino);
			ASSERT(ip->map);

			flash_imap_expand(ip, areab.sectno);
			fd->inusepages++;
			if ((pn = ip->map[areab.sectno]) == 0) {
				ip->map[areab.sectno] = areab.firstno;
			} else if (pn != areab.firstno) {
				AREAB	b;
				
				FD_READB(fd, pn, &b);
				ASSERT(b.sequence != areab.sequence);
				b.sequence++;
				if (b.sequence == areab.sequence) {
					/* Prev one was older one */
					/* delete new one, likely bad */
					pn = areab.firstno;
					FD_READB(fd, pn, &b);
				} else {
					b.sequence--;
					areab.sequence++;
					ASSERT(b.sequence == areab.sequence);
					ip->map[areab.sectno] = areab.firstno;
				}

				for ( ;; ) {
					b.ino = 0;
					FD_WRITEB(fd, pn, &b);
					if (pn <= pageno) {
						fd->inusepages--;
						FD_DELETE(fd, pn);
					}
					pn = b.nextno;
					if (pn == 0xFFFF)
						break;
					FD_READB(fd, pn, &b);
				}
			}
		}
		if (blockno != FLASH_ROOTBLKNO && 
	            fd->blkdelcnt[blockno] > (fd->ppb / 4))
			fd_gc_block(fd, blockno);

		if (dev == ROOT_DEV && blockno > chunkno) {
			chunkno += inc;
			STARTBAR("Kernel: mounting device %s as root ... ",
				kdevname(dev));
		}
	}

	fd->inodes[FLASH_ROOTINO - 1].nlink = 1;
	fd->inodes[FLASH_ROOTINO].nlink = 1;
	fd->inodes[FLASH_ROOTINO].parent = FLASH_ROOTINO;

	fd_bp.fd = fd;
	fd_bp.buf = fd_bp.buffer;
	/*
	 *  Pass 2: Search thru all directories for inused inodes 
	*/
	ip = &fd->inodes[FLASH_ROOTINO];
	for (i = FLASH_ROOTINO; i < FLASH_DIRINO; ++i, ++ip) {
		if (ip->map == 0)
			continue;
		for (j = 0; j < ip->mapsize; ++j) {
			if (ip->map[j] == 0)
				continue;
			fd_bp.sectno = j;
			fd_bp.pageno = ip->map[j];
			fd_bp.ino = i;
			if (flash_read(&fd_bp)) {
				ASSERT(0);
				continue;
			}

			for (size = 0; size < FFS_SECTSIZE; size += ep->entsize) {
				ep = (DIRENT *) (fd_bp.buf + size);
				if (ep->entsize == 0xFF)
					break;
				ASSERT(ep->entsize != 0);
				if (ep->ino == 0)
					continue;
				ASSERT(ep->ino < FLASH_INODES);
				if (ep->ino >= FLASH_INODES)
					continue;
				ip->nlink++;
				fd->inodes[ep->ino].parent = i;
				if (fd->inodes[ep->ino].nlink++ == 0) {
					fd->inuseinodes++;
					fd->freeinodes--;
				}
			}
		}
	}

	/*
	 *  Pass 3: Search all inodes for orphans
	*/
	ip = &fd->inodes[FLASH_ROOTINO];
	for (i = FLASH_ROOTINO; i < FLASH_INODES; ++i, ++ip) {
		if (ip->nlink) {	/* referenced inodes */
			if (i < FLASH_DIRINO)
				ip->nlink++;
			continue;
		}
		if (ip->map == 0)	/* empty inode */
			continue;

#ifndef	EPROM
		printf("flash: orphan inode 0x%x with mapsize 0x%x\n", 
			i, ip->mapsize);
		for (j = 0; j < ip->mapsize; ++j) {
			if (ip->map[j] != 0)
				flash_delete(fd, ip->map[j]);
		}
		flash_irelse(fd, i);
#endif
	}

	return(0);
}

int
flash_init(FLASHDEV *fd, int blocks, int pages, int blocksize, int pagesize)
{
	ASSERT(fd->blocks == 0 && blocks > 0);
	fd->blocks = blocks;
	fd->pages = pages;
	fd->blocksize = blocksize;
	fd->pagesize = pagesize;
	fd->rawsize = (pages - fd->rootpageno) * (pagesize + fd->rdntsize);
	fd->zipsize = FFS_SECTSIZE - pagesize;
	fd->ppb = blocksize / pagesize;
	fd->pgend = pages - fd->ppb;
	fd->blkend = blocks - 1;

	if (fd->alloc == 0)
		fd->alloc = fd_alloc;
	if (fd->free == 0)
		fd->free = fd_free;
	if (fd->delete == 0)
		fd->delete = fd_delete;

	switch(fd->ppb) {
	case 8:	  fd->blkshift = 3; break;
	case 16:  fd->blkshift = 4; break;
	case 32:  fd->blkshift = 5; break;
	case 64:  fd->blkshift = 6; break;
	case 128: fd->blkshift = 7; break;
	case 256: fd->blkshift = 8; break;
	default: ASSERT(0); 
	}

	/* Now, fake the ppb for Flash Card support */
	if (blocksize >= 0x8000) {	/* Must be non-nand device */
		fd->ppb = blocksize / (pagesize + sizeof(AREAB));
	}

	fd->badblocks = (flash_sect_t *) malloc(fd->pagesize);
	if (fd->badblocks == 0)
		return(-ENOMEM);
	memset(fd->badblocks, 0, fd->pagesize);

	if (fd->dev & FLASH_FORMAT)
		flash_erase(fd);
	return(0);
}

void
flash_close(FLASHDEV *fd)
{
	struct 	flash_imap *ip;
	int	i;

	if (fd->blkdelcnt) {
		ip = &fd->inodes[FLASH_ROOTINO];
		for (i = FLASH_ROOTINO; i < FLASH_INODES; ++i, ++ip) {
			if (ip->map == 0)	/* empty inode */
				continue;
			free(ip->map);
		}
		free(fd->blkdelcnt);
	}
	if (fd->badblocks)
		free(fd->badblocks);
	memset(fd, 0, sizeof(*fd));
}

FLASHDEV *
flash_device(kdev_t dev)
{
	FLASHDEV	*fd;
	int		i;

	for (i = 0, fd = flashs; i < NFLASHS; ++i, ++fd) {
		if (fd->dev == dev && fd->blkdelcnt != 0)
			return(fd);
	}
	return(0);
}

int
fd_extent(FLASHDEV *fd, int size, flash_sect_t *nextno)
{
	flash_sect_t pageno, *firstno = nextno;

	while(size > 0) {
		size -= fd->pagesize;
		pageno = FD_ALLOC(fd);
		if (pageno == 0xFFFF) {
			while(nextno > firstno) { 	/* put back */
				--nextno;
				FD_FREE(fd, *nextno);
			} 
			return(-ENOSPC);
		}
		*nextno++ = pageno;
	}
	*nextno = 0xFFFF;
	return(0);
}

void
flash_delete(FLASHDEV *fd, int pageno)
{
#ifndef	EPROM
	AREAB	areab;

	do {
		ASSERT(pageno < fd->pages);
		FD_READB(fd, pageno, &areab);
		ASSERT(areab.ino != 0 && areab.ino != 0xFFFF);
		ASSERT(areab.nextno == 0xFFFF || areab.nextno < fd->pages);
		areab.ino = 0;	/* mark old page deleted */
		FD_WRITEB(fd, pageno, &areab);

		FD_DELETE(fd, pageno);
		pageno = areab.nextno;
	} while(pageno != 0xFFFF);
#endif
}

/*
 *	The following flash_read() and flash_write()
 *	are the only routines interfacing with file systems.
 *	The routines have the knowledge of buffer structure,
 *	which defines the pageno, sequence, ino, etc.
*/	
int
flash_read(struct ffs_buf *bp)
{
	int	pageno, error, size, i;
	char	*buf;
	AREAB	areab;
	FLASHDEV *fd = bp->fd;

	ASSERT(bp->pageno < fd->pages);
	pageno = bp->pageno;

	error = FD_READ(fd, pageno, fd_tmpbuf, &areab); 
	if (error)
		return(error);

	size = areab.size;
	ASSERT(size > 0 && size <= FFS_SECTSIZE);
	bp->sequence = areab.sequence + 1;
	if (size == FFS_SECTSIZE) {
		buf = bp->buf;
		memcpy(buf, fd_tmpbuf, fd->pagesize);
	} else {
		buf = fd_tmpbuf;
	}

	for (i = fd->pagesize; i < size; i += fd->pagesize) {
		pageno = areab.nextno;
		ASSERT(areab.sectno == bp->sectno);
		ASSERT(areab.firstno == bp->pageno);
		ASSERT(areab.ino == bp->ino);
		ASSERT(pageno < fd->pages);
		if (pageno == 0xFFFF)
			return(-EIO);
		error = FD_READ(fd, pageno, buf + i, &areab);
		if (error)
			return(error);
	}

	ASSERT(areab.nextno == 0xFFFF);
	if (size <= fd->zipsize) {
		i = FFS_SECTSIZE;
		error = uncompress(bp->buf, &i, fd_tmpbuf, size);
		if (error || i != FFS_SECTSIZE) {
			printf("flash_read: uncompress = %d, size = 0x%x\n",
				error, i);
			return(-EIO);
		}
	}
	return(0);
}

int
flash_write(struct ffs_buf *bp)
{
#ifndef	EPROM
	int	pageno, oldpageno, error, size, i;
	AREAB	areab;
	char	*buf;
	flash_sect_t *n, nextno[FFS_SECTSIZE / 512 + 1];
	FLASHDEV *fd = bp->fd;

	ASSERT(bp->pageno < fd->pages);
	ASSERT(bp->pageno > 0);

	ASSERT(bp->dirty == 1 || bp->dirty == 2);
	/*
	 * If after compress, we can't save even a page, we might
	 * as well dont compress this 4K chunk.  So limit the
	 * output buffer size to be 4K - pagesize.
	*/
	size = fd->zipsize;
	error = compress(fd_tmpbuf, &size, bp->buf, FFS_SECTSIZE);
	if (error) {
		size = FFS_SECTSIZE;
		buf = bp->buf;
	} else {
		buf = fd_tmpbuf;
	}

	areab.inuse = 0;
	areab.size = size; 
	areab.ino = bp->ino;
	areab.sectno = bp->sectno;
	oldpageno = bp->pageno;

	n = nextno;
	if (bp->dirty == 2) {
		if (fd_extent(fd, size, nextno) < 0)
			return(-ENOSPC);
		pageno = *n++;	/* all new pages */
		bp->pageno = pageno;	/* buf identity changed */
		ASSERT(fd->inodes[bp->ino].map[bp->sectno] == oldpageno);
		fd->inodes[bp->ino].map[bp->sectno] = pageno;
	} else {
		if (fd_extent(fd, size - fd->pagesize, nextno) < 0)
			return(-ENOSPC);
		pageno = bp->pageno;	/* new page added */
	}
	areab.sequence = bp->sequence++;
	areab.firstno = pageno;

	for (i = 0; ; ) {
		areab.nextno = *n++;
		areab.pageno  = pageno;
		if (FD_WRITE(fd, pageno, buf + i, &areab) != 0) {
			printf("flash%d: write error on pageno 0x%x\n",
				fd - flashs, pageno);
		}
		i += fd->pagesize;
		if (i >= size) {
			ASSERT(areab.nextno == 0xFFFF);
			break;
		}
		pageno = areab.nextno;
		ASSERT(pageno < fd->pages);
	}

	if (bp->dirty == 2) {		/* now, we can delete old ones */
		flash_delete(fd, oldpageno);
	}
	bp->dirty = 0;
#endif
	return(0);
}

int
flash_alloc(struct ffs_buf *bp)
{
	FLASHDEV *fd = bp->fd;
	flash_sect_t	pageno[2];

	if (fd->freepages < FD_MINFREE && fd->alloc == fd_alloc)
		fd_gc(fd);
	if (fd->freepages < FD_MINFREE || 
	    fd_extent(fd, fd->pagesize, pageno) < 0)
		return(0);
	bp->pageno = pageno[0];
	return(1);
}

int
flash_ialloc(FLASHDEV *fd, int isdir)
{
	int	i, startindx;
	struct 	flash_imap *ip;

	ASSERT(fd->inoindx >= FLASH_DIRINO);
	i = (isdir) ?  FLASH_ROOTINO + 1 : fd->inoindx;
	startindx = i; 

	for ( ;; ) {
		ip = &fd->inodes[i];
		if (ip->nlink == 0) {
			if (!isdir) {
				fd->inoindx = i;
				ip->nlink = 1;
			} else {
				ip->nlink = 2;
			}
			ASSERT(fd->freeinodes > 0);
			fd->freeinodes--;
			fd->inuseinodes++;
			return(i);
		}
		if (isdir) {
			if (++i >= FLASH_DIRINO)
				i = FLASH_ROOTINO + 1;
		} else {
			if (++i >= FLASH_INODES)
				i = FLASH_DIRINO;
		}
		if (i == startindx) {
			return(0);
		}
	}
}

struct flash_imap * 
flash_getimap(FLASHDEV *fd, int ino)
{
	struct 	flash_imap *ip;
	int	size;

	ASSERT(ino < FLASH_INODES);
	ip = &fd->inodes[ino];
	if (ip->map == 0) {
		size = FLASH_IMAPSIZE * sizeof(flash_sect_t);
		ip->map = (flash_sect_t *) malloc(size);
		if (ip->map == 0)
			return(0);
		memset(ip->map, 0, size);
		ip->mapsize = FLASH_IMAPSIZE;
	}
	return(ip);
}

void
flash_irelse(FLASHDEV *fd, int ino)
{
	struct 	flash_imap *ip;

	ip = &fd->inodes[ino];
	if (ip->map)
		free(ip->map);
	ip->map = 0;
	ip->mapsize = 0;
	ip->nlink = 0;
	ASSERT(fd->inuseinodes > 0);
	fd->inuseinodes--;
	fd->freeinodes++;
}

#ifdef	DEBUG
void
flash_idump(FLASHDEV *fd, int ino)
{
	struct 	flash_imap *ip;

	ip = &fd->inodes[ino];
	printf("ip: 0x%x, map: 0x%x, mapsize: 0x%x, nlink: %d\n",
		(int) ip, (int) ip->map, ip->mapsize, ip->nlink);
}
#endif
