#ifdef	__KERNEL__
#include <linux/config.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/ioctl.h>
#include <linux/stat.h>
#include <linux/flash_fs.h>
#include <linux/flash.h>
#include "fnand.h"

#include <linux/delay.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>

/*	Flash Major Device is set to 62 */
#define MAJOR_NR 	FNAND_MAJOR

typedef volatile unsigned char vubyte;
#include <linux/blk.h>

#else

#include "system.h"
#include "errno.h"
#include "flash_fs.h"
#include "flash.h"
#include "fnand.h"

#define	MKDEV(j, n)	n

#endif

#define	FNAND_BASEPART(fd)	(fd == FNAND_DEV0 || fd == FNAND_DEV2)
#define	FNAND_BOOTPART(fd)	(fd == FNAND_DEV1 || fd == FNAND_DEV3)

//	Definitions for MARS platform
#define	MARS_FNAND_ADDR	0x2800000

#define	DATA		(*((vubyte *) (fnand_vaddr + 0x0000)))
#define	ADDRESS		(*((vubyte *) (fnand_vaddr + 0x1000)))
#define	DISABLE		(*((vubyte *) (fnand_vaddr + 0x2000)))
#define	COMMAND		(*((vubyte *) (fnand_vaddr + 0x3000)))

#define	FNAND_BOOTSIZE	(256 * 1024)
int	fnand_vaddr;

//	Definitions for LEO and MIDWAY platforms
//	int	gpio_reg2 = 0xE0;
#define	GPIO_REG2			0xE0

#define MIDWAY_ADDRESS_PORT   0x100
#define MIDWAY_TERMINATE_PORT 0x200
#define MIDWAY_COMMAND_PORT   0x300

#define LEO_ADDRESS_PORT      0x200
#define LEO_TERMINATE_PORT    0x300
#define LEO_COMMAND_PORT      0x400

#define ADDRESS_PORT    (WYSE_MIDWAY ? MIDWAY_ADDRESS_PORT : LEO_ADDRESS_PORT)
#define TERMINATE_PORT  (WYSE_MIDWAY ? MIDWAY_TERMINATE_PORT : LEO_TERMINATE_PORT)
#define COMMAND_PORT    (WYSE_MIDWAY ? MIDWAY_COMMAND_PORT : LEO_COMMAND_PORT)

//#ifdef WYSE_MIDWAY
#define	ENABLE_SUSTAIN()  		{OUTB(GPIO_REG2, INB(GPIO_REG2) | 0x20);udelay(1);}
#define	DISABLE_SUSTAIN() 		{OUTB(GPIO_REG2, INB(GPIO_REG2) & 0xDF);udelay(1);}

//#define	ENABLE_SUSTAIN()  		
//#define	DISABLE_SUSTAIN() 		

#define	DATA_READ(port)	  		    INW(port + 0x000) >> 8
#define	DATA_WRITE(port, val)  		OUTW(port + 0x000, (val) << 8)
#define	ADDRESS_WRITE(port, addr)	OUTW(port + ADDRESS_PORT, (addr) << 8)
#define TERMINATE_IN(port)          INW(port +  TERMINATE_PORT)
#define	COMMAND_SEND(port, cmd) 	OUTW(port + COMMAND_PORT, (cmd) << 8)

#define	ENABLE_WRITE_PROTECT()		{OUTB(GPIO_REG2, INB(GPIO_REG2) | 0x10);udelay(1);}
#define	DISABLE_WRITE_PROTECT()		{OUTB(GPIO_REG2, INB(GPIO_REG2) & 0xEF);udelay(1);}
/*
#else
#define	ENABLE_SUSTAIN()  		{OUTB(GPIO_REG2, INB(GPIO_REG2) | 0x20);udelay(1);}
#define	DISABLE_SUSTAIN() 		{OUTB(GPIO_REG2, INB(GPIO_REG2) & 0xDF);udelay(1);}

#define	DATA_READ(port)	  		    INW(port + 0x000) >> 8
#define	DATA_WRITE(port, val)  		OUTW(port + 0x000, (val) << 8)
#define	ADDRESS_WRITE(port, addr)	OUTW(port + 0x200, (addr) << 8)
#define TERMINATE_IN(port)
#define	COMMAND_SEND(port, cmd) 	OUTW(port + 0x400, (cmd) << 8)

#define	ENABLE_WRITE_PROTECT()		{OUTB(GPIO_REG2, INB(GPIO_REG2) | 0x10);udelay(1);}
#define	DISABLE_WRITE_PROTECT()		{OUTB(GPIO_REG2, INB(GPIO_REG2) & 0xEF);udelay(1);}
#endif
*/

//	End of Platform specific definitions

extern	char	fd_tmpbuf[FFS_SECTSIZE + FLASH_RDNTSIZE];
extern	char	fd_gcbuf[FFS_SECTSIZE + FLASH_RDNTSIZE];

/* 
 * RYAN: HEXDUMP for debugging purpose
 */
#ifdef WYSE_KDB 
void 
hexdump(u_char *data, int len, int offset)
{
	int	i, use_offset, c;
    u_char *p;

	/* init to match code in wt5000-boot-libc-misc.c */
	p = data;
	c = len;

	if (c < 0) {
		use_offset = 1;
		c = -c;
	} else
		use_offset = 0;

	for (;;) {
		if (use_offset)
			printf("%04x: ", offset);
		else
			printf("%08x: ", (unsigned int)p);

		for (i=0; i<16; i++) {
			if (i >= c)
				printf("   ");
			else
				printf("%02x ", p[i]);
		}
		printf("   ");

		for (i=0; i<16; i++) {
			if (i >= c)
				break;
			if (p[i] < ' ' || p[i] > '~')
				printf(".");
			else
				printf("%c", p[i]);
		}
		printf("\n");

		offset += 16;
		p += 16;
		c -= 16;
		if (c <= 0) {
			//printf("\n");
			break;
		}
	}
}
#endif

STATIC int
fn_wait4ready(FLASHDEV *fd)
{
	int	i;
	int flags;

	save_flags(flags); cli();
	for (i = 0; i < 100000; ++i) {
		if ( WYSE_MIDWAY ) {
			if ((INB(GPIO_REG2) & 0x04) != 0) { 
				restore_flags(flags);
				return(1);
			}
		} else 	if (WYSE_LEO) {
			OUTL(0xCF8, 0x80009090);
			if ((INW(0xCFC) & 0x8000) != 0) {
			  	restore_flags(flags);
				return(1);
			}
		} else { 
			OUTB(0x22, 0xa8);
			if ((INB(0x23) & 0x01) != 0) {
			  	restore_flags(flags);
				return(1);
			}
		}
	}
	restore_flags(flags);
	printf("fn_wait4ready: timeout\n");
	return(0);
}

STATIC int 
fn_status(FLASHDEV *fd)
{
  	int flags;
	
	save_flags(flags); cli();
	if (!fn_wait4ready(fd)) {
	  	restore_flags(flags);
		return(-ETIMEDOUT);
	}
 	if (WYSE_LEO) {
		COMMAND_SEND(fd->port, FNAND_CMD_STATUS_READ);
		if (DATA_READ(fd->port) & FNAND_STS_FAIL) {
			printf("flash: status failed\n");
			restore_flags(flags);
			return(-EIO);
		}
		restore_flags(flags);
		return(0);
	} else {
		COMMAND = (u_char) FNAND_CMD_STATUS_READ;
		if (DATA & FNAND_STS_FAIL) {
			printf("flash: status failed\n"); 
			restore_flags(flags);
			return(-EIO);
		}
		restore_flags(flags);
		restore_flags(flags);
		return(0);
	}
}

STATIC void
fn_setup(FLASHDEV *fd, int cmd, int pageno)
{
	int	flags, smi;

	ASSERT(pageno < fd->pages);

	fn_wait4ready(fd);

	if (WYSE_LEO) {
		save_flags(flags); cli();
		OUTB(0x22, 0xC1);
		smi = INB(0x23);
		OUTB(0x22, 0xC1);
		OUTB(0x23, smi & 0xFD);		// ignore SMI
		ENABLE_SUSTAIN();
		COMMAND_SEND(fd->port, cmd);
		if (cmd != FNAND_CMD_ERASE)
			ADDRESS_WRITE(fd->port, 0);
		ADDRESS_WRITE(fd->port, pageno);
		ADDRESS_WRITE(fd->port, pageno >> 8);
		OUTB(0x22, 0xC1);
		OUTB(0x23, smi);
		restore_flags(flags);
	} else {
		COMMAND = cmd;
		if (cmd != FNAND_CMD_ERASE)
			ADDRESS = 0;
		ADDRESS = pageno;
		ADDRESS = pageno >> 8;
	}

	fn_wait4ready(fd);
}

STATIC u_short
fn_input(FLASHDEV *fd, char *buf, int size)
{
	u_short	sum, *p;
	int flags;

	save_flags(flags); cli();
	ASSERT((size % 8) == 0);

	/* RYAN: WYSE_MIDWAY shares with WYSE_LEO here */
	if (WYSE_LEO) {
		for (sum = 0 ; size > 0; size -= 8, buf += 8) {
			buf[0] = DATA_READ(fd->port); 
			buf[1] = DATA_READ(fd->port); 
			buf[2] = DATA_READ(fd->port); 
			buf[3] = DATA_READ(fd->port);
			buf[4] = DATA_READ(fd->port); 
			buf[5] = DATA_READ(fd->port); 
			buf[6] = DATA_READ(fd->port); 
			buf[7] = DATA_READ(fd->port);
			p = (u_short *) buf;
			sum += p[0] + p[1] + p[2] + p[3];
		}
	} else {
		for (sum = 0 ; size > 0; size -= 8, buf += 8) {
			buf[0] = DATA; buf[1] = DATA; 
			buf[2] = DATA; buf[3] = DATA;
			buf[4] = DATA; buf[5] = DATA; 
			buf[6] = DATA; buf[7] = DATA;
			p = (u_short *) buf;
			sum += p[0] + p[1] + p[2] + p[3];
		}
	}
	restore_flags(flags);
	return(sum);
}

STATIC u_short
fn_output(FLASHDEV *fd, char *buf, int size)
{
	u_short	sum, *p;
	int flags;

	save_flags(flags); cli();
	ASSERT((size % 8) == 0);
	if (WYSE_LEO) {
		for (sum = 0; size > 0; size -= 8, buf += 8) {
			DATA_WRITE(fd->port, buf[0]); 
			DATA_WRITE(fd->port, buf[1]); 
			DATA_WRITE(fd->port, buf[2]); 
			DATA_WRITE(fd->port, buf[3]);
			DATA_WRITE(fd->port, buf[4]); 
			DATA_WRITE(fd->port, buf[5]); 
			DATA_WRITE(fd->port, buf[6]); 
			DATA_WRITE(fd->port, buf[7]);
			p = (u_short *) buf;
			sum += p[0] + p[1] + p[2] + p[3];
		}
	} else {
		for (sum = 0; size > 0; size -= 8, buf += 8) {
			DATA = buf[0]; DATA = buf[1]; 
			DATA = buf[2]; DATA = buf[3];
			DATA = buf[4]; DATA = buf[5]; 
			DATA = buf[6]; DATA = buf[7];
			p = (u_short *) buf;
			sum += p[0] + p[1] + p[2] + p[3];
		}
	}
	restore_flags(flags);
	return(sum);
}

STATIC void
fn_memset(FLASHDEV *fd, int size)
{
  	int flags;
  	save_flags(flags); cli();
	if (WYSE_LEO) {
		for ( ; size > 0; size -= 8) {
			DATA_WRITE(fd->port, 0xFF); 
			DATA_WRITE(fd->port, 0xFF); 
			DATA_WRITE(fd->port, 0xFF); 
			DATA_WRITE(fd->port, 0xFF);
			DATA_WRITE(fd->port, 0xFF); 
			DATA_WRITE(fd->port, 0xFF); 
			DATA_WRITE(fd->port, 0xFF); 
			DATA_WRITE(fd->port, 0xFF);
		}
	} else {
		for ( ; size > 0; size -= 8) {
			DATA = 0xFF; DATA = 0xFF; DATA = 0xFF; DATA = 0xFF;
			DATA = 0xFF; DATA = 0xFF; DATA = 0xFF; DATA = 0xFF;
		}
	}
	restore_flags(flags);
}

STATIC int
fn_standby(FLASHDEV *fd)
{
	if (WYSE_MARS)
		return(DISABLE);
	return(0);
}

STATIC int
fnand_readb(FLASHDEV *fd, int pageno, AREAB *b)
{
//#ifdef WYSE_MIDWAY
	int flags;
	save_flags(flags); cli();
	ASSERT(pageno < fd->pages);
    fn_setup(fd, FNAND_CMD_READ_MODE2, pageno);
	fn_input(fd, (char *) b, sizeof(*b));
	//error = fn_status(fd);
    if ( WYSE_MIDWAY ) {
		/* reset back to main data area */
		fn_setup(fd, FNAND_CMD_READ_MODE1, 0);
    
		TERMINATE_IN(fd->port);  //terminate nand 	
    } else {
		if (WYSE_LEO) {
			DISABLE_SUSTAIN();
			COMMAND_SEND(fd->port, 0); 
		} else {
			COMMAND = 0;		/* reset back to main data area */
		}
		fn_standby(fd);
	}
	restore_flags(flags);
	return(0);		
/*
#else
	int flags;
	save_flags(flags); cli();
	ASSERT(pageno < fd->pages);
    fn_setup(fd, FNAND_CMD_READ_MODE2, pageno);
	fn_input(fd, (char *) b, sizeof(*b));
	if (WYSE_LEO) {
		DISABLE_SUSTAIN();
		COMMAND_SEND(fd->port, 0); 
	} else {
		COMMAND = 0;		// reset back to main data area
	}
	fn_standby(fd);
	restore_flags(flags);
	return(0);
#endif
*/
}

STATIC int
fn_read(FLASHDEV *fd, int pageno, char *buf, AREAB *b)
{
	int	flags;

	save_flags(flags); cli();
	ASSERT(pageno < fd->pages);
    fn_setup(fd, FNAND_CMD_READ_MODE1, pageno);
	fn_input(fd, buf, fd->pagesize);
	fn_input(fd, (char *) b, fd->rdntsize);
    if ( WYSE_MIDWAY )
		TERMINATE_IN(fd->port);  //terminate nand 
    else {
		if (WYSE_LEO)
			DISABLE_SUSTAIN();
		fn_standby(fd);
    }
	restore_flags(flags);
	return(0);
}

STATIC int
fnand_read(FLASHDEV *fd, int pageno, char *buf, AREAB *b)
{
	u_short	sum;
	int flags;

	ASSERT(pageno < fd->pages);
	save_flags(flags); cli();
    fn_setup(fd, FNAND_CMD_READ_MODE1, pageno);
	sum =  fn_input(fd, buf, fd->pagesize);
	sum += fn_input(fd, (char *) b, fd->rdntsize);
    if ( WYSE_MIDWAY )
		TERMINATE_IN(fd->port);  //terminate nand 
    else {
		if (WYSE_LEO)
			DISABLE_SUSTAIN();
		fn_standby(fd);
    } 

	if (sum != FNAND_CHECKSUM) {
		restore_flags(flags);
		printf("fnand_read: cksum 0x%x on page 0x%x port 0x%x\n",
		       sum, pageno, fd->port);
		ASSERT(0);
		return(-EIO);
	}
	restore_flags(flags);
	return(0);
}

STATIC int
fnand_writeb(FLASHDEV *fd, int pageno, AREAB *b)
{
	int	error;
	int 	flags;

	save_flags(flags); cli();
	ASSERT(pageno < fd->pages);
    fn_setup(fd, FNAND_CMD_WRITE, pageno);
	fn_memset(fd, fd->pagesize);
	fn_output(fd, (char *) b, sizeof(*b));
	if (WYSE_LEO) {
		COMMAND_SEND(fd->port, FNAND_CMD_AUTO_PROGRAM);
	} else {
		COMMAND = (u_char) FNAND_CMD_AUTO_PROGRAM;
	}

	fn_wait4ready(fd);
	error = fn_status(fd);
    if ( WYSE_MIDWAY )
		TERMINATE_IN(fd->port);  //terminate nand 
    else {
		if (WYSE_LEO)
			DISABLE_SUSTAIN();
		fn_standby(fd);
	}
	restore_flags(flags);
	return(error);
}

int
fn_write(FLASHDEV *fd, int pageno, char *buf, AREAB *b)
{
	int	error;
	u_short	sum;
	int	flags;

	save_flags(flags); cli();
	ASSERT(pageno < fd->pages);
    fn_setup(fd, FNAND_CMD_WRITE, pageno);
	sum = fn_output(fd, buf, fd->pagesize);
	if (b->inuse == 0xfe) {
		b->checksum = 0;
		sum += flash_checksum((u_short *) b, fd->rdntsize / 2);
		b->checksum = FNAND_CHECKSUM - sum;
	}
	fn_output(fd, (char *) b, fd->rdntsize);
	if (fd->rdntsize < sizeof(*b))
		fn_memset(fd, sizeof(*b) - fd->rdntsize);
	if (WYSE_LEO) {
		COMMAND_SEND(fd->port, FNAND_CMD_AUTO_PROGRAM);
	} else {
		COMMAND = (u_char) FNAND_CMD_AUTO_PROGRAM;
	}

	fn_wait4ready(fd);
	error = fn_status(fd);
    if ( WYSE_MIDWAY )
		TERMINATE_IN(fd->port);  //terminate nand 
    else {
		if (WYSE_LEO)
			DISABLE_SUSTAIN();
		fn_standby(fd);
	}
	restore_flags(flags);
	return(error);
}

int
fnand_write(FLASHDEV *fd, int pageno, char *buf, AREAB *b)
{
	int	error;
	u_short	sum;
	int	flags;

	save_flags(flags); cli();
	ASSERT(pageno < fd->pages);
    fn_setup(fd, FNAND_CMD_WRITE, pageno);
	sum = fn_output(fd, buf, fd->pagesize);
	b->checksum = 0;
	sum += flash_checksum((u_short *) b, fd->rdntsize / 2);
	b->checksum = FNAND_CHECKSUM - sum;
	fn_output(fd, (char *) b, fd->rdntsize);
	if (WYSE_LEO) {
		COMMAND_SEND(fd->port, FNAND_CMD_AUTO_PROGRAM);
	} else {
		COMMAND = (u_char) FNAND_CMD_AUTO_PROGRAM;
	}

	if (fn_wait4ready(fd) == 0) {
			printf("fnand_init1: not ready\n");
			return(-ENODEV);
		}
		
	//fn_wait4ready(fd);
	error = fn_status(fd);
    if ( WYSE_MIDWAY )
		TERMINATE_IN(fd->port);  //terminate nand 
    else {
		if (WYSE_LEO)
			DISABLE_SUSTAIN();
		fn_standby(fd);
    }
	restore_flags(flags);
	return(error);
}

int
fnand_erase(FLASHDEV *fd, int blockno, int verify)
{
	int	error = verify;
	int	pageno = blockno << fd->blkshift;
	int	i, j, *p;
	int	flags;

	save_flags(flags); cli();
	ASSERT(pageno < fd->pages);
	if (pageno >= fd->pages) {
	  	restore_flags(flags);
		return(-1);
	}
	if (WYSE_LEO) {
		fn_setup(fd, FNAND_CMD_ERASE, pageno);
		COMMAND_SEND(fd->port, FNAND_CMD_ERASE2);
	} else {
		fn_setup(fd, FNAND_CMD_ERASE, pageno);
		COMMAND = (u_char) FNAND_CMD_ERASE2;
	}

	fn_wait4ready(fd);
	error = fn_status(fd);
    if ( WYSE_MIDWAY )
		TERMINATE_IN(fd->port);  //terminate nand 
    else {
		if (WYSE_LEO)
			DISABLE_SUSTAIN();
		fn_standby(fd);
	}	
	
	if (error) {
	  	restore_flags(flags);
		return(error);
	}

	for (i = 0; i < fd->ppb; ++i, ++pageno) {
		fn_setup(fd, FNAND_CMD_READ_MODE1, pageno);
		fn_input(fd, fd_gcbuf, fd->pagesize);
		fn_input(fd, fd_gcbuf + fd->pagesize, sizeof(AREAB));
	    if ( WYSE_MIDWAY )
			TERMINATE_IN(fd->port);  //terminate nand 
		else {
			if (WYSE_LEO)
				DISABLE_SUSTAIN();
			fn_standby(fd);
		}
		j = (fd->pagesize + sizeof(AREAB)) / sizeof(*p); 
		p = (int *) fd_gcbuf;
		for (j = 0; j < (fd->pagesize + sizeof(AREAB))/sizeof(*p); ++j) {
			if (*p++ == 0xFFFFFFFF)
				continue;
			restore_flags(flags);
			printk("fnand_erase: block 0x%x, pageno 0x%x, offset 0x%x, bad data 0x%x (fd->port 0x%x)\n",
				blockno, pageno, j, *--p, fd->port);
			ASSERT(0);
			return(-ENXIO);
		}
	}
	restore_flags(flags);
	return(0);
}

int 
fnand_open(struct inode * inode, struct file * filp)
{
	int	minor;
	FLASHDEV *fd;
	int	i, blocks, pages, blocksize, pagesize;
	vubyte	id0, id1;


	ASSERT(inode);
	minor = FLASH_MINOR(inode->i_rdev);
	i = (WYSE_MARS) ? NFNANDS / 2 : NFNANDS;
	if (minor >= i)
		return(-ENODEV);
	fd = FNAND_DEV0 + minor;
	if (fd->blocks != 0)
		return(-EBUSY);

	fd->port 	= 0x1000 + (minor / 2) * 0x800;	/* chip2 = 0x1800 */ 
	if (WYSE_LEO) {
		//COMMAND_SEND(fd->port, FNAND_CMD_RESET);
		if (fn_wait4ready(0) == 0) {
			printf("fnand_init1: not ready\n");
			return(-ENODEV);
		}
		ENABLE_SUSTAIN();
		COMMAND_SEND(fd->port, FNAND_CMD_ID_READ);	
		ADDRESS_WRITE(fd->port, 0x00);
		id0 = DATA_READ(fd->port);	/* ID 0 */
		id1 = DATA_READ(fd->port);	/* ID 1 */
		DISABLE_SUSTAIN();
	} else {
		//COMMAND = (u_char) FNAND_CMD_RESET;
		if (fn_wait4ready(0) == 0) {
			printf("fnand_init2: not ready\n");
			return(-ENODEV);
		}
		COMMAND  = FNAND_CMD_ID_READ;	/* Read Flash ID */
		ADDRESS  = 0x00;
		id0 = DATA;			/* ID 0 */
		id1 = DATA;			/* ID 1 */
	}

	fd->read 	= fnand_read;	/* export 5 functions */
	fd->readb	= fnand_readb;
	fd->write	= fnand_write;
	fd->writeb	= fnand_writeb;
	fd->erase	= fnand_erase;  
	fd->dev		= inode->i_rdev;

	switch(id1) {
	case 0x75:		/* TH58V256FT */
		blocks	  = FNAND_BLOCKS * 4 - 1; /* Reduce by 1 block because of BOS limitation */
		pages	  = FNAND_PAGES * 8 - 0x20; /* Reduce by 1 block because of BOS limitation */
		blocksize = FNAND_BLOCK_SIZE * 2;
		pagesize  = FNAND_PAGE_SIZE;
		break;
	case 0x73:		/* TH58V128FT */
		blocks	  = FNAND_BLOCKS * 2;
		pages	  = FNAND_PAGES * 4;
		blocksize = FNAND_BLOCK_SIZE * 2;
		pagesize  = FNAND_PAGE_SIZE;
		break;
	case 0xE6:		/* TC5864AFT */
		blocks    = FNAND_BLOCKS * 2;
		pages     = FNAND_PAGES * 2;
		blocksize = FNAND_BLOCK_SIZE;
		pagesize  = FNAND_PAGE_SIZE;
		break;
	case 0xE5:		/* TC58V32AFT */
	case 0x6B:		/* TC5832AFT */
		blocks    = FNAND_BLOCKS;
		pages     = FNAND_PAGES;
		blocksize = FNAND_BLOCK_SIZE;
		pagesize  = FNAND_PAGE_SIZE;
		break;
	case 0x6E:		/* TC5816AFT */
		blocks    = FNAND_BLOCKS;
		pages     = FNAND_PAGES;
		blocksize = FNAND_BLOCK_SIZE * 2;
		pagesize  = FNAND_PAGE_SIZE * 2;
		break;
	default:
		return(-ENXIO);
	}

	if (FNAND_BOOTPART(fd)) {
		fd->rdntsize    = 8;
		fd->rootblockno	= 0;
		fd->rootpageno	= 0;
		blocks = FNAND_BOOTSIZE / blocksize;
		pages  = FNAND_BOOTSIZE / pagesize;
	} else {
		fd->rdntsize    = sizeof(AREAB);
		fd->rootblockno	= FNAND_BOOTSIZE / blocksize;
		fd->rootpageno	= FNAND_BOOTSIZE / pagesize;
	}

	i = flash_init(fd, blocks, pages, blocksize, pagesize); 
	if (i != 0)		// error return
		return(i);
	if (FNAND_BASEPART(fd))	// non-boot partition
		return(0); 
	if (filp->f_flags & O_WRONLY)
		flash_erase(fd);
	return(0);
}

STATIC int 
fnand_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	FLASHDEV *fd;
	int 	err;
	
	if (!inode || !inode->i_rdev) 	
		return -EINVAL;

	fd = flash_device(inode->i_rdev);
	if (fd == 0)
		return -EIO;

	switch (cmd) {
	case BLKFLSBUF:
		return -EACCES;
		break;
	case BLKGETSIZE:   /* Return device size */
		if (!arg)  return -EINVAL;
		err = verify_area(VERIFY_WRITE, (long *) arg,
				  sizeof(long));
		if (err)
			return err;
		put_user((fd->pages * fd->pagesize) / 512, 
			 (long *) arg);
		return 0;
	}

	return 0;
}

STATIC void 
fnand_close(struct inode *inode, struct file *file)
{
	FLASHDEV *fd = FNAND_DEV0 + FLASH_MINOR(inode->i_rdev);

	sync_dev(inode->i_rdev);
	flash_close(fd);
}

int 
fnand_blkread(struct inode *inode, struct file *filp, char *buf, int count)
{
	int	pageno;
	int	offset = filp->f_pos;
	char	*b0 = buf;
	FLASHDEV *fd = FNAND_DEV0 + FLASH_MINOR(inode->i_rdev);
	int	pgrdsize = fd->pagesize + fd->rdntsize;

	if (offset + count > fd->rawsize) 
		count = fd->rawsize - offset;
	if (count <= 0)
		return(0);
	if ((count % pgrdsize) != 0 || (offset % pgrdsize) != 0) {
		if (FNAND_BASEPART(fd))
			return(count);
	}
	while (count > 0) {
		pageno = offset / pgrdsize;
		fn_read(fd, pageno + fd->rootpageno, (char *) fd_tmpbuf, 
				(AREAB *) (fd_tmpbuf + fd->pagesize));
		memcpy_tofs(buf, fd_tmpbuf, pgrdsize);	/* copy to user space */
		count -= pgrdsize;
		buf += pgrdsize;
		offset += pgrdsize;
	}
	filp->f_pos = offset;
	return(buf - b0);
}

int 
fnand_blkwrite(struct inode *inode, struct file *filp,
	const char *buf, int count)
{
	int	pageno;
	int	offset = filp->f_pos;
	char	*b0 = (char *) buf;
	FLASHDEV *fd = FNAND_DEV0 + FLASH_MINOR(inode->i_rdev);
	int	pgrdsize = fd->pagesize + fd->rdntsize;

	if ((count % pgrdsize) != 0 || (offset % pgrdsize) != 0)
		return(-EIO);
	if (offset + count > fd->rawsize) 
		count = fd->rawsize - offset;
	if (count <= 0)
		return(0);
	while (count > 0) {
		pageno = offset / pgrdsize;
		memcpy_fromfs(fd_tmpbuf, (char *) buf, pgrdsize);
		fn_write(fd, pageno + fd->rootpageno, (char *) fd_tmpbuf, 
				(AREAB *) (fd_tmpbuf + fd->pagesize));
		count -= pgrdsize;
		buf += pgrdsize;
		offset += pgrdsize;
	}
	filp->f_pos = offset;
	return(buf - b0);
}

static struct file_operations fn_fops = {
	NULL,		/* lseek - default */
	fnand_blkread,	/* read - block dev read */
	fnand_blkwrite,	/* write - block dev write */
	NULL,		/* readdir - not here! */
	NULL,		/* select */
	fnand_ioctl, 	/* ioctl */
	NULL,		/* mmap */
	fnand_open,	/* open */
	fnand_close,	/* no special release code... */
	block_fsync	/* fsync */ 
};

/* This is the registration and initialization section of the flash driver */
int 
fnand_linux_init(void)
{
	if (register_blkdev(MAJOR_NR, "fnand", &fn_fops)) {
		printf("fnand: Could not get major %d", MAJOR_NR);
		return -EIO;
	}

	printk( KERN_INFO "Major=%d FlashNand driver initialized - 1 Nand device\n", MAJOR_NR);
	return 0;
}

int
fnand_init(void)
{
  	int flags;

	save_flags(flags); cli();
	if (WYSE_MARS) {
		fnand_vaddr = (int) vremap(MARS_FNAND_ADDR, 0x4000);
		if (fnand_vaddr == 0) {
			printf("fnand_init: vremap failed\n");
			restore_flags(flags);
			return(-ENOMEM);
		}

#ifndef	LATER
		setWindowROMCS(2);	/* Addr Window 2, MMS window D */
		setCSC(0x31, (getCSC(0x31) & 0xF3) | 0x08);	/* enable it */	
#else
		enableFlashWindow();
		selectROMCS(2);
#endif
	}

	if (WYSE_LEO) {
		DISABLE_WRITE_PROTECT();
	}

	restore_flags(flags);
	return(fnand_linux_init());
}

