#ifdef	__KERNEL__
#include <linux/config.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/malloc.h>
#include <linux/ioctl.h>
#include <linux/stat.h>
#include <linux/flash_fs.h>
#include <linux/flash.h>
#include "fnand.h"

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>

#define	malloc(size)	kmalloc(size, GFP_KERNEL)  
typedef volatile unsigned short vushort;

#else

#include "system.h"
#include "errno.h"
#include "flash_fs.h"
#include "flash.h"
#include "fnand.h"
#endif

#define	DATA		(*((vushort *) (fd->port + 0x0000)))
#define	ADDRESS		(*((vushort *) (fd->port + 0x1000)))
#define	DISABLE		(*((vushort *) (fd->port + 0x2000)))
#define	COMMAND		(*((vushort *) (fd->port + 0x3000)))
#define	COMMAND_XCE	(*((vushort *) (fd->port + 0x3000)))

#define	FN_ROOTBLOCKNO	0

STATIC int
fnc_wait4ready(FLASHDEV *fd)
{
	int	i = 0;

	while(fcard_ready() == 0) {		/* wait for ready */
		if (++i >= 100000) {
			//printf("fnc_wait4ready: timeout\n");
			return(0);
		}
	}
	return(1);
}

STATIC int 
fnc_status(FLASHDEV *fd)
{
	if (!fnc_wait4ready(fd)) 
		return(-ETIMEDOUT);

        COMMAND = FNAND_CMD_STATUS_READ;

	if (DATA & FNAND_STS_FAIL) {
		printf("flash: status failed\n"); 
		return(-EIO);
	}
	return(0);
}

STATIC void
fnc_setup(FLASHDEV *fd, int cmd, int pageno)
{
	int	i;

	ASSERT(pageno < fd->pages);

	fnc_wait4ready(fd);

        COMMAND_XCE = cmd;
	if (cmd != FNAND_CMD_ERASE)
		ADDRESS = 0x0000;
	
	i = pageno & 0xFF;
	ADDRESS = i + (i << 8);
	i = pageno >> 8;
	ASSERT(i < 0x100);
	ADDRESS = i + (i << 8);

	fnc_wait4ready(fd);
}

STATIC u_short
fnc_input(FLASHDEV *fd, short *buf, int size)
{
	u_short	sum, *p;

	ASSERT((size % 16) == 0);
	for (sum = 0 ; size > 0; size -= 16, buf += 8) {
		buf[0] = DATA; buf[1] = DATA; buf[2] = DATA; buf[3] = DATA;
		buf[4] = DATA; buf[5] = DATA; buf[6] = DATA; buf[7] = DATA;
		p = (u_short *) buf;
		sum += p[0] + p[1] + p[2] + p[3] + p[4] + p[5] + p[6] + p[7];
	}
	return(sum);
}

STATIC u_short
fnc_output(FLASHDEV *fd, short *buf, int size)
{
	u_short	sum, *p;

	ASSERT((size % 16) == 0);
	for (sum = 0; size > 0; size -= 16, buf += 8) {
		DATA = buf[0]; DATA = buf[1]; DATA = buf[2]; DATA = buf[3];
		DATA = buf[4]; DATA = buf[5]; DATA = buf[6]; DATA = buf[7];
		p = (u_short *) buf;
		sum += p[0] + p[1] + p[2] + p[3] + p[4] + p[5] + p[6] + p[7];
	}
	return(sum);
}

STATIC void
fnc_memset(FLASHDEV *fd, int size)
{
	ASSERT((size % 16) == 0);
	for ( ; size > 0; size -= 16) {
		DATA = 0xFFFF; DATA = 0xFFFF; DATA = 0xFFFF; DATA = 0xFFFF;
		DATA = 0xFFFF; DATA = 0xFFFF; DATA = 0xFFFF; DATA = 0xFFFF;
	}
}

STATIC int
fnc_standby(FLASHDEV *fd)
{
	return(DISABLE);
}

STATIC int
fncard_readb(FLASHDEV *fd, int pageno, AREAB *b)
{
	int	error;

	ASSERT(pageno < fd->pages);
        fnc_setup(fd, FNAND_CMD_READ_MODE2, pageno);
	fnc_input(fd, (short *) b, sizeof(*b));
	error = fnc_status(fd);
	COMMAND = 0x0000;	/* reset back to main data area */
	fnc_standby(fd);
	return(error);
}

STATIC int
fncard_read(FLASHDEV *fd, int pageno, short *buf, AREAB *b)
{
	int	error;
	u_short	sum;

	ASSERT(pageno < fd->pages);
	ASSERT(sizeof(*b) == 16);
        fnc_setup(fd, FNAND_CMD_READ_MODE1, pageno);
	sum =  fnc_input(fd, buf, fd->pagesize);
	sum += fnc_input(fd, (short *) b, sizeof(*b));
	error = fnc_status(fd);
	fnc_standby(fd);
	if (error)
		return(error);
	if (sum != FNAND_CHECKSUM) {
		printf("fncard_read: cksum 0x%x on page 0x%x\n", sum, pageno);
		return(-EIO);
	}
	return(error);
}

STATIC int
fncard_writeb(FLASHDEV *fd, int pageno, AREAB *b)
{
	int	error;

	ASSERT(pageno < fd->pages);
	ASSERT(sizeof(*b) == 16);
        fnc_setup(fd, FNAND_CMD_WRITE, pageno);
	fnc_memset(fd, fd->pagesize);
	fnc_output(fd, (short *) b, sizeof(*b));
	fnc_memset(fd, sizeof(*b));
	COMMAND = FNAND_CMD_AUTO_PROGRAM;

	fnc_wait4ready(fd);
	error = fnc_status(fd);
	fnc_standby(fd);
	return(error);
}

int
fncard_write(FLASHDEV *fd, int pageno, short *buf, AREAB *b)
{
	int	error;
	u_short	sum;

	ASSERT(pageno < fd->pages);
	b->checksum = 0;
        fnc_setup(fd, FNAND_CMD_WRITE, pageno);
	sum = fnc_output(fd, buf, fd->pagesize);
	sum += flash_checksum((u_short *) b, (sizeof(*b) / 2));
	b->checksum = FNAND_CHECKSUM - sum;
	fnc_output(fd, (short *) b, sizeof(*b));
	fnc_memset(fd, sizeof(*b));
	COMMAND = FNAND_CMD_AUTO_PROGRAM;

	fnc_wait4ready(fd);
	error = fnc_status(fd);
	fnc_standby(fd);
	return(error);
}

int
fncard_erase(FLASHDEV *fd, int blockno, int verify)
{
	int	error = verify;
	int	i, j, *p, pageno = blockno << fd->blkshift;
	int	tmp[(FNAND_PAGE_SIZE * 2) / 4 + sizeof(AREAB) / 4];

	ASSERT(pageno < fd->pages);
	if (pageno >= fd->pages)
		return(-1);
	fnc_setup(fd, FNAND_CMD_ERASE, pageno);
	COMMAND_XCE = FNAND_CMD_ERASE2;

	fnc_wait4ready(fd);
	error = fnc_status(fd);
	fnc_standby(fd);

	if (error == 0 && verify != 0) {
		for (i = 0; i < fd->ppb; ++i, ++pageno) {
			fnc_setup(fd, FNAND_CMD_READ_MODE1, pageno);
			fnc_input(fd, tmp, fd->pagesize + sizeof(AREAB));
			error = fnc_status(fd);
			fnc_standby(fd);
			if (error)
				break;
			p = tmp;
			for (j = 0; j < fd->pagesize + sizeof(AREAB); j += 4) {
				if (*p++ == -1) 
					continue;
				printf("fncard_erase: pageno 0x%x failed\n",
					pageno);
				++error;
				break;
			}
		}
	}
	
	return(error);
}

#ifdef	__KERNEL__

extern	short	fd_tmpbuf[(FFS_SECTSIZE + FLASH_RDNTSIZE) / 2];

int 
fncard_blkread(struct inode *inode, struct file *filp, char *buf, int count)
{
	int	c;
	int	pageno;
	int	off, offset = filp->f_pos;
	char	*b0 = buf;
	FLASHDEV *fd;

	c = FLASH_MINOR(inode->i_rdev);
	fd = FCARD_DEV0 + c;

	if (offset + count > fd->rawsize) 
		count = fd->rawsize - offset;
	while (count > 0) {
		pageno = offset / sizeof(fd_tmpbuf);
		off = offset - pageno * sizeof(fd_tmpbuf);
		c = sizeof(fd_tmpbuf) - off;	/* remaining data in a window */
		if (c > count) 
			c = count;

		fnc_setup(fd, FNAND_CMD_READ_MODE1, pageno);
		fnc_input(fd, fd_tmpbuf, sizeof(fd_tmpbuf));
		fnc_status(fd);
		fnc_standby(fd);
		memcpy_tofs(buf, fd_tmpbuf + off, c);	/* copy to user space */

		count -= c;
		buf += c;
		offset += c;
	}
	filp->f_pos = offset;
	return(buf - b0);
}

int 
fncard_blkwrite(struct inode *inode, struct file *filp,
	const char *buf, int count)
{
	int	c;
	int	pageno;
	int	off, offset = filp->f_pos;
	char	*b0 = (char *) buf;
	FLASHDEV *fd;

	c = FLASH_MINOR(inode->i_rdev);
	fd = FCARD_DEV0 + c;

	ASSERT((count & (fd->pagesize - 1)) == 0);
	ASSERT((offset & (fd->pagesize - 1)) == 0);
	if (offset + count > fd->rawsize) 
		count = fd->rawsize - offset;
	while (count > 0) {
		pageno = offset / sizeof(fd_tmpbuf);
		off = offset - pageno * sizeof(fd_tmpbuf);
		c = sizeof(fd_tmpbuf) - off;	/* remaining data in a window */
		if (c > count) 
			c = count;

		memset(fd_tmpbuf, 0xFF, sizeof(fd_tmpbuf));
		memcpy_fromfs(fd_tmpbuf + off, (char *) buf, c);
		fnc_setup(fd, FNAND_CMD_WRITE, pageno);
		fnc_output(fd, fd_tmpbuf, sizeof(fd_tmpbuf));
		COMMAND = FNAND_CMD_AUTO_PROGRAM;
		fnc_wait4ready(fd);
		fnc_status(fd);
		fnc_standby(fd);

		count -= c;
		buf += c;
		offset += c;
	}
	filp->f_pos = offset;
	return(buf - b0);
}
#endif

int
fncard_open(FLASHDEV *fd)
{
	vushort	id0, id1;

	COMMAND = FNAND_CMD_RESET;
	if (fnc_wait4ready(fd) == 0) {
		//printf("fncard_open: not ready\n");
		return(-ENODEV);
	}

	COMMAND  = FNAND_CMD_ID_READ;	/* Read Flash ID */
	ADDRESS  = 0x00;
	id0 = DATA;			/* ID 0 */
	id1 = DATA;			/* ID 1 */

	//FFS_PRINTF("fncard_open: 0x%x  0x%x\n", id0, id1);

	fd->read 	= fncard_read;	/* export 5 functions */
	fd->readb	= fncard_readb;
	fd->write	= fncard_write;
	fd->writeb	= fncard_writeb;
	fd->erase	= fncard_erase;  

	fd->rootblockno	= 0;
	fd->rootpageno	= 0;

	switch(id1) {
	case 0xE6E6:		/* Dual TC58V64AFT */
		return(flash_init(fd, FNAND_BLOCKS * 2, FNAND_PAGES * 2, 
				  FNAND_BLOCK_SIZE * 2, FNAND_PAGE_SIZE * 2));
	}
	return(-ENXIO);
}
