#ifdef	__KERNEL__
#include <linux/config.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/ioctl.h>
#include <linux/module.h>
#include <linux/stat.h>
#include <linux/flash_fs.h>
#include <linux/flash.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>

/*	Flash Card Major Device is set to 63 */
#define MAJOR_NR 	FNOR_MAJOR

#include <linux/debug.h>
#include <linux/blk.h>

typedef volatile unsigned short vushort;

#else

#include "system.h"
#include "errno.h"
#include "flash_fs.h"
#include "flash.h"

#endif

#define	FNOR_SIZE	0x200000
#ifndef	LATER
#define	FNOR_ADDR	0x2C00000
#else
#define	FNOR_ADDR	0x2800000
#endif

#define	FX_ROOTBLOCKNO	0
#define	FX_BLOCKSIZE	0x0010000	
#define	FX_PAGESIZE	0x0000200	
#define	FX_BLKSHIFT	7
#define	FX_BLKMASK	0x7F
#define	FX_BLOCKS	(FNOR_SIZE / FX_BLOCKSIZE)
#define	FX_PAGES	(FNOR_SIZE / FX_PAGESIZE)

int	fnor_vaddr;
int	fnor_id;
u_short	fnor_cksum;

#define	fx_setshort(off, value)	*((vushort *) (off + fnor_vaddr)) = value
#define	fx_getshort(off)	*((vushort *) (off + fnor_vaddr))

#define	FX_PGOFST(pg)		((pg >> FX_BLKSHIFT) * FX_BLOCKSIZE + \
			         (pg & FX_BLKMASK) * (FX_PAGESIZE + 16))

#define	FX_PGOFST_B(pg)		FX_PGOFST(pg) + FX_PAGESIZE
				 
/*
 *	Static internal routines
*/

STATIC int 
fx_test(int offset, u_short write_data)
{
	int	i, j;

	for (i = 0; ; ++i) {
		if (i >= 0x3000000) {
			fx_setshort(offset, 0xFFFF);
			printf("fx_test: write 0x%x, read 0x%x\n",
				write_data, fx_getshort(offset));
			ASSERT(0);
			break;
		}
		/* read data */
		if ((j = fx_getshort(offset)) & 0x80) {	/* status done? */
			fx_setshort(offset, 0xFFFF);
			if (j & 0x20) {
				printf("fx_test: program error 0x%x\n", j);
				ASSERT(0);
				break;
			}
			return(0);
		}
	}

	return(-ENXIO);
}

STATIC void
fx_input(int offset, u_short *buf, int count)
{
	while(count > 0) { 
		*buf = fx_getshort(offset);
		fnor_cksum += *buf++;
		offset += sizeof(*buf);
		count -= sizeof(*buf);
	}
}

STATIC int 
fx_output(int offset, u_short *buf, int count)
{       
	u_short	s;

	while(count > 0) { 
		if (*buf != 0xFFFF) {
			s = fx_getshort(offset);
			if (s != *buf) {
				fx_setshort(offset, 0x5050); /* clear status */
				fx_setshort(offset, 0x4040);/* write command */
				fx_setshort(offset, *buf);
				if (fx_test(offset, *buf))
					return(-EIO);
				if ((s = fx_getshort(offset)) != *buf) {
					printf("fx_output: wrote 0x%x, read back 0x%x\n",
						*buf, s);
					ASSERT(0);
					return(-ENXIO);
				}
			}
		}
		fnor_cksum += *buf++;
		offset += sizeof(*buf);
		count -= sizeof(*buf);
	}
	return(0);
}

STATIC int
fnor_readb(FLASHDEV *fd, int pageno, AREAB *b)
{
	int	offset = FX_PGOFST_B(pageno);

	ASSERT(pageno < fd->pages);
	fx_input(offset, (u_short *) b, sizeof(*b));
	return(0);
}

STATIC int
fnor_read(FLASHDEV *fd, int pageno, char *buf, AREAB *b)
{
	int	offset = FX_PGOFST(pageno);

	ASSERT(pageno < fd->pages);
	fnor_cksum = 0;
	fx_input(offset, (u_short *) buf, fd->pagesize);
	fx_input(offset + FX_PAGESIZE, (u_short *) b, sizeof(*b));
	if (fnor_cksum != FLASH_CHECKSUM) {
		printf("fnor_read: cksum 0x%x on page 0x%x\n", 
			fnor_cksum, pageno);
		ASSERT(0);
		return(-EIO);
	}
	return(0);
}

STATIC int
fnor_writeb(FLASHDEV *fd, int pageno, AREAB *b)
{
	int	offset = FX_PGOFST_B(pageno);

	ASSERT(pageno < fd->pages);
	return(fx_output(offset, (u_short *) b, sizeof(*b)));
}

int
fnor_write(FLASHDEV *fd, int pageno, char *buf, AREAB *b)
{
	int	offset = FX_PGOFST(pageno);

	ASSERT(pageno < fd->pages);
	fnor_cksum = 0;
	if (fx_output(offset, (u_short *) buf, fd->pagesize) < 0)
		return(-EIO);
	b->checksum = 0;
	fnor_cksum += flash_checksum((u_short *) b, (sizeof(*b) / 2));
	b->checksum = FLASH_CHECKSUM - fnor_cksum;
	return(fx_output(offset + FX_PAGESIZE, (u_short *) b, sizeof(*b)));
}

int
fnor_erase(FLASHDEV *fd, int blockno, int verify)
{
	int	i, pageno = blockno << fd->blkshift;
	int	offset = FX_PGOFST(pageno);
	vushort	*p = (vushort *) (offset + fnor_vaddr);

	ASSERT(blockno < fd->blocks);
	fx_setshort(offset, 0x5050);		/* clear status */
	fx_setshort(offset, 0x2020);		/* erase command */
	fx_setshort(offset, 0xd0d0);		/* erase confirmation */
	if (fx_test(offset, 0xFFFF) < 0)
		return(-ENXIO);

	for (i = 0; i < FX_BLOCKSIZE; i += 2, ++p) { 
		if (*p == 0xFFFF)
			continue;
		printk("fnor_erase: block 0x%x, offset 0x%x, bad data 0x%x\n",
			blockno, offset, *p);
		ASSERT(0);
		return(-ENXIO);
	}
	return(0);
}

int 
fnor_open(struct inode * inode, struct file * filp)
{
	FLASHDEV *fd = FNOR_DEV0;

	ASSERT(inode);
	if (fd->blocks != 0)
		return(-EBUSY);

	fd->read 	= fnor_read;	/* export 5 functions */
	fd->readb	= fnor_readb;
	fd->write	= fnor_write;
	fd->writeb	= fnor_writeb;
	fd->erase	= fnor_erase;  
	fd->dev		= inode->i_rdev;

	fd->rootblockno	= FX_ROOTBLOCKNO;
	fd->rootpageno	= FX_ROOTBLOCKNO * (FX_BLOCKSIZE / FX_PAGESIZE); 

	switch(fnor_id) {
	case 0x66a0:		/* Intel 28F016SV */
#ifdef	LATER
		selectROMCS(1);
#endif
		return(flash_init(fd, FX_BLOCKS, FX_PAGES, 
				       FX_BLOCKSIZE, FX_PAGESIZE));
	}
	return(-ENXIO);
}

#ifdef	__KERNEL__

extern	char	fd_tmpbuf[];	/* borrow from flash.c */

STATIC void 
fnor_close(struct inode *inode, struct file *file)
{
	sync_dev(inode->i_rdev);
	flash_close(FNOR_DEV0);
}

int 
fnor_blkread(struct inode *inode, struct file *filp, char *buf, int count)
{
	int	c;
	int	offset = filp->f_pos;
	char	*b0 = buf;
	FLASHDEV *fd = FNOR_DEV0;

	/* To read flash card, count must be in even number for short read */
	ASSERT((offset & 1) == 0);
	ASSERT((count & 1) == 0);
	if (offset + count > fd->rawsize) 
		count = fd->rawsize - offset;
	while (count > 0) {
		c = count;
		if (c > FFS_SECTSIZE)
			c = FFS_SECTSIZE;
		fx_input(offset, (u_short *) fd_tmpbuf, c); /* copy data in */
		memcpy_tofs(buf, fd_tmpbuf, c);		/* copy to user space */
		count -= c;
		buf += c;
		offset += c;
	}
	filp->f_pos = offset;
	return(buf - b0);
}

int 
fnor_blkwrite(struct inode *inode, struct file *filp,
	const char *buf, int count)
{
	int	c;
	int	offset = filp->f_pos;
	char	*b0 = (char *) buf;
	FLASHDEV *fd = FNOR_DEV0;

	ASSERT((count & 1) == 0);
	ASSERT((offset & 1) == 0);
	if (offset + count > fd->rawsize) 
		count = fd->rawsize - offset;
	while (count > 0) {
		c = count;
		if (c > FFS_SECTSIZE)
			c = FFS_SECTSIZE;
		memcpy_fromfs(fd_tmpbuf, (char *) buf, c);
		if (fx_output(offset, (u_short *) fd_tmpbuf, c) < 0)
			break;
		count -= c;
		buf += c;
		offset += c;
	}
	filp->f_pos = offset;
	return(buf - b0);
}

STATIC int 
fnor_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int 	err;
	FLASHDEV *fd = FNOR_DEV0;
	
	if (!inode || !inode->i_rdev) 	
		return -EINVAL;

	switch (cmd) {
	case BLKFLSBUF:
		if (!suser()) 
			return -EACCES;
		invalidate_buffers(inode->i_rdev);
		break;
	case BLKGETSIZE:   /* Return device size */
		if (!arg)  
			return -EINVAL;
		err = verify_area(VERIFY_WRITE, (long *) arg, sizeof(long));
		if (err)
			return err;
		put_user(fd->rawsize / 512, (long *) arg);
		return 0;
	}

	return 0;
}

static struct file_operations fn_fops = {
	NULL,		/* lseek - default */
	fnor_blkread,	/* read - block dev read */
	fnor_blkwrite,	/* write - block dev write */
	NULL,		/* readdir - not here! */
	NULL,		/* select */
	fnor_ioctl, 	/* ioctl */
	NULL,		/* mmap */
	fnor_open,	/* open */
	fnor_close,	/* release */
	block_fsync	/* fsync */ 
};

/* This is the registration and initialization section of the card driver */
STATIC int 
fnor_linux_init(FLASHDEV *fd)
{
	if (register_blkdev(MAJOR_NR, "fnor", &fn_fops)) {
		printf("fnor: Could not get major %d", MAJOR_NR);
		return -EIO;
	}

	printf("FlashNor driver initialized - 1 Nor device\n");

	return 0;
}
#endif

int 
fnor_init(void)
{
	FLASHDEV *fd = FNOR_DEV0;

	if (WYSE_LEO) 
		return(-ENODEV);

	fnor_vaddr = (int) vremap(FNOR_ADDR, FNOR_SIZE);
	if (fnor_vaddr == 0) {
		printf("fnor_init: vremap failed\n");
		return(-ENOMEM);
	}

#ifndef	LATER
	// setMMSwindow(1, FNOR_ADDR, FNOR_SIZE);
	// should have moved the block of code to boot block 
	set3E0(0x58, 0x00);	/* start address bits 19:12 = 00 */ 
	set3E0(0x59, 0x2C);	/* start address bits 25:20 = 2C */ 
	set3E0(0x5A, 0xFF);	/* stop address bits 19:12 = 00 */ 
	set3E0(0x5B, 0x2F);	/* stop address bits 25:20 = 2F */ 
	set3E0(0x5C, 0x00);	/* offset bits 19:12 = 00 */ 
	set3E0(0x5D, 0x00);	/* offset bits 25:20 = 04 */ 

	setWindowROMCS(1);	/* Addr Window 1, MMS window C */
	setCSC(0x31, (getCSC(0x31) & 0xFC) | 0x01);	/* enable it */	
#else
	enableFlashWindow();
	selectROMCS(1);
#endif

	fx_setshort(0, 0x9090);	
	fnor_id = fx_getshort(2);
	fx_setshort(0, 0xFFFF);		/* reset */	

	switch(fnor_id) {
	case 0x66a0:		/* Intel 28F016SV */
		break;
	default:
		printf("fnor_init: unknown Nor Flash ID 0x%x\n", fnor_id);
		return(-ENODEV);
	}
	
#ifdef	__KERNEL__
	return(fnor_linux_init(fd));
#else
	return(0);
#endif
}

