#include <linux/config.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/ioctl.h>
#include <linux/module.h>
#include <linux/stat.h>
#include <linux/flash_fs.h>
#include <linux/flash.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>

/*	Flash Card Major Device is set to 63 */
#define MAJOR_NR 	FRAM_MAJOR

#include <linux/debug.h>
#include <linux/blk.h>

#define	FR_SIZE		0x1000000
#define	FR_ROOTBLOCKNO	0
#define	FR_BLOCKSIZE	0x008000	
#define	FR_PAGESIZE	0x000200	
#define	FR_BLKSHIFT	6
#define	FR_BLKMASK	0x3F

#define	FR_PGOFST(pg)		((pg & FR_BLKMASK) * (FR_PAGESIZE + 16))
#define	FR_PGOFST_B(pg)		FR_PGOFST(pg) + FR_PAGESIZE
				 
#define	malloc(size)	kmalloc(size, GFP_KERNEL)  
#define	free(ptr)	kfree(ptr)

extern	void		fd_free(FLASHDEV *fd, int pageno);
/*
 *	Static internal routines
 */
static	AREAB		fram_areab;
u_char			**fr_blockmap;

STATIC int
fram_readb(FLASHDEV *fd, int pageno, AREAB *b)
{
	int	blockno = pageno >> fd->blkshift;
	int	offset = FR_PGOFST_B(pageno);
	u_char	*devptr;

	ASSERT(pageno < fd->pages);
	devptr = fr_blockmap[blockno];
	if (pageno == fd->rootpageno)
		*b = fram_areab;
	else if (devptr)
		memcpy((char *) b, devptr + offset, sizeof(*b)); 
	else
		return(-EIO);
	return(0);
}

STATIC int
fram_read(FLASHDEV *fd, int pageno, char *buf, AREAB *b)
{
	int	blockno = pageno >> fd->blkshift;
	int	offset = FR_PGOFST(pageno);
	u_char	*devptr;

	ASSERT(pageno < fd->pages);
	if (pageno == fd->rootpageno)
		*b = fram_areab;
	else if ((devptr = fr_blockmap[blockno]) != 0) {
		memcpy(buf, devptr + offset, fd->pagesize); 
		memcpy(b, devptr + offset + FR_PAGESIZE, sizeof(*b)); 
	} else
		return(-EIO);
	return(0);
}

STATIC int
fram_writeb(FLASHDEV *fd, int pageno, AREAB *b)
{
	int	blockno = pageno >> fd->blkshift;
	int	offset = FR_PGOFST_B(pageno);

	ASSERT(pageno < fd->pages);
	if (pageno == fd->rootpageno)
		fram_areab = *b;
	else {
		ASSERT(fr_blockmap[blockno] != 0);
		memcpy(fr_blockmap[blockno] + offset, (char *) b, sizeof(*b)); 
	}
	return(0);
}

int
fram_write(FLASHDEV *fd, int pageno, char *buf, AREAB *b)
{
	int	blockno = pageno >> fd->blkshift;
	int	offset = FR_PGOFST(pageno);

	ASSERT(pageno < fd->pages);
	if (pageno == fd->rootpageno)
		fram_areab = *b;
	else {
		ASSERT(fr_blockmap[blockno] != 0);
		memcpy(fr_blockmap[blockno] + offset, buf, fd->pagesize); 
		memcpy(fr_blockmap[blockno] + offset + FR_PAGESIZE, b, sizeof(*b)); 
	}
	return(0);
}

int
fram_erase(FLASHDEV *fd, int blockno, int verify)
{
	return(0);
}

int
fram_alloc(FLASHDEV *fd)
{
	int	i, j, blockno, pageno;

	for (i = FR_ROOTBLOCKNO, blockno = -1; i < fd->blocks; ++i) {
		if (fd->blkdelcnt[i] > 0)	/* found a block */
			break;
		if (blockno < 0 && fr_blockmap[i] == 0)
			blockno = i;
	}

	if (i >= fd->blocks) {		/* temp out */
		if (blockno < 0)
			return(0xFFFF);
		fr_blockmap[blockno] = (u_char *) vmalloc(FR_BLOCKSIZE);
		if (fr_blockmap[blockno] == 0) {
			printk("fram: out of memory\n");
			return(0xFFFF);
		}

		memset(fr_blockmap[blockno], 0xFF, FR_BLOCKSIZE);
		i = (blockno == FR_ROOTBLOCKNO) ? 1 : 0;
		fd->blkdelcnt[blockno] = fd->ppb - i;
		fd->freepages -= fd->ppb + i;
		pageno = i + (blockno << fd->blkshift);
		for  (; i < fd->ppb; ++i)
			fd_free(fd, pageno++);	
		i = blockno;
	} else
		blockno = i;

	i <<= (fd->blkshift - 3);
	for ( ;; ) {
		if (fd->freemap[i] != 0) {
			pageno = i * 8;
			for (j = 0; j < 8; ++j, ++pageno) {
				if (fd->freemap[i] & (1 << j)) {
					fd->freemap[i] &= ~(1 << j);
					fd->freepages--;
					fd->blkdelcnt[blockno]--;
					return(pageno);
				}
			}
		}
		if (++i >= fd->pages / 8)
			break;
	}
	ASSERT(0);
	return(0xFFFF);
}

void
fram_free(FLASHDEV *fd, int pageno)
{
	int	blockno = pageno >> fd->blkshift;

	ASSERT(fr_blockmap[blockno] != 0);
	fd_free(fd, pageno);

	fd->blkdelcnt[blockno]++;
	if (fd->blkdelcnt[blockno] >= fd->ppb) {
		ASSERT(blockno != FR_ROOTBLOCKNO);
		vfree((void *) fr_blockmap[blockno]);
		fr_blockmap[blockno] = 0;
		fd->blkdelcnt[blockno] = 0;
		pageno = blockno << fd->blkshift;
		pageno >>= 3;
		memset(fd->freemap+pageno, 0, (FR_BLOCKSIZE / FR_PAGESIZE)/8);
	}
}

int 
fram_open(struct inode * inode, struct file * filp)
{
	FLASHDEV *fd = FRAM_DEV0;
	int	blocks, pages, i;

	ASSERT(inode);
	if (fd->blocks) 
		return(-EBUSY);
	if ((MINOR(inode->i_rdev) & FLASH_FORMAT) == 0)
		return(-ENXIO);

	fd->read 	= fram_read;	/* export 5 functions */
	fd->readb	= fram_readb;
	fd->write	= fram_write;
	fd->writeb	= fram_writeb;
	fd->erase	= fram_erase;  
	fd->alloc	= fram_alloc;
	fd->free	= fram_free;
	fd->delete	= fram_free;
	fd->dev		= inode->i_rdev;

	fd->rootblockno	= FR_ROOTBLOCKNO;
	fd->rootpageno	= FR_ROOTBLOCKNO * (FR_BLOCKSIZE / FR_PAGESIZE); 

	blocks = (FR_SIZE / FR_BLOCKSIZE);
	pages  = (FR_SIZE / FR_PAGESIZE);
	fr_blockmap = (u_char **) malloc(blocks * sizeof(fr_blockmap[0]));
	if (fr_blockmap == 0)
		return(-ENOMEM);
	memset(fr_blockmap, 0, blocks * sizeof(fr_blockmap[0]));
	i = flash_init(fd, blocks, pages, FR_BLOCKSIZE, FR_PAGESIZE);
	if (i >= 0) 
		fd->freepages = blocks * fd->ppb - 1;
	return(i);
}

#ifdef	__KERNEL__

extern	char	fd_tmpbuf[];	/* borrow from flash.c */

STATIC void 
fram_close(struct inode *inode, struct file *file)
{
	FLASHDEV *fd = FRAM_DEV0;
	int	i;

	sync_dev(inode->i_rdev);
	if (fr_blockmap) {
		for (i = 0; i < FR_SIZE / FR_BLOCKSIZE; ++i) {
			if (fr_blockmap[i])
				vfree(fr_blockmap[i]);
		}
		free((void *) fr_blockmap);
		fr_blockmap = 0;
	}
	flash_close(fd);
	memset((char *) &fram_areab, 0xFF, sizeof(fram_areab));
}

STATIC int 
fram_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int 	err;
	FLASHDEV *fd = FRAM_DEV0;
	
	if (!inode || !inode->i_rdev) 	
		return -EINVAL;

	switch (cmd) {
	case BLKFLSBUF:
		if (!suser()) 
			return -EACCES;
		invalidate_buffers(inode->i_rdev);
		break;
	case BLKGETSIZE:   /* Return device size */
		if (!arg)  
			return -EINVAL;
		err = verify_area(VERIFY_WRITE, (long *) arg, sizeof(long));
		if (err)
			return err;
		put_user(fd->rawsize / 512, (long *) arg);
		return 0;
	}

	return 0;
}

static struct file_operations fr_fops = {
	NULL,		/* lseek - default */
	NULL,		/* read - block dev read */
	NULL,		/* write - block dev write */
	NULL,		/* readdir - not here! */
	NULL,		/* select */
	fram_ioctl, 	/* ioctl */
	NULL,		/* mmap */
	fram_open,	/* open */
	fram_close,	/* release */
	block_fsync	/* fsync */ 
};

/* This is the registration and initialization section of the card driver */
STATIC int 
fram_linux_init(FLASHDEV *fd)
{
	if (register_blkdev(MAJOR_NR, "fram", &fr_fops)) {
		printf("fram: Could not get major %d", MAJOR_NR);
		return -EIO;
	}

	printf("FlashRam driver initialized - 1 Ram device\n");

	memset((char *) &fram_areab, 0xFF, sizeof(fram_areab));
	return 0;
}
#endif

int 
fram_init(void)
{
	FLASHDEV *fd = FRAM_DEV0;

#ifdef	__KERNEL__
	return(fram_linux_init(fd));
#else
	return(0);
#endif
}

