
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <linux/delay.h>
#include <linux/errno.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

#include <linux/debug.h>
#include "cs8900.h"
/*
 * The name of the card. Is used for messages and in the requests for
 * io regions, irqs and dma channels
 */
static const char* cardname = 
	"Ethernet: Crystal LAN (tm) CS8900 Ethernet Adapter";

/* A zero-terminated list of I/O addresses to be probed. */
static unsigned int cs_portlist[] = { CS_IOBASE, 0};

/* use 0 for production, 1 for verification, >2 for debug */
   
/* The number of low I/O ports used by the ethercard. */
#define CS_IO_EXTENT	16

/* Information that need to be kept for each board. */
struct cs_local {
	struct enet_statistics stats;
	long open_time;			/* Useless example local info. */
};

/* Index to functions, as function prototypes. */

extern void print_eth(char *, int len);
extern int cs8900_probe(struct device *dev);

STATIC int cs_probe1(struct device *dev, int ioaddr);
STATIC int cs_open(struct device *dev);
STATIC int cs_send_packet(struct sk_buff *skb, struct device *dev);
STATIC void cs_interrupt(int irq, void *dev_id, struct pt_regs *regs);
STATIC void cs_rx(struct device *dev);
STATIC int cs_close(struct device *dev);
STATIC struct enet_statistics *cs_get_stats(struct device *dev);
STATIC void set_multicast_list(struct device *dev);

/* Example routines you must write ;->. */
#define tx_done(dev) 1
extern void	hardware_send_packet(short ioaddr, char *buf, int length);
extern void chipset_init(struct device *dev, int startp);

static	unsigned char ea_addr[] = { 0x00, 0x80, 0x64, 0x10, 0x12, 0x7c };
/*
 * Check for a network adaptor of this type, and return '0' iff one exists.
 * If dev->base_addr == 0, probe all likely locations.
 * If dev->base_addr == 1, always return failure.
 * If dev->base_addr == 2, allocate space for the device and return success
 * (detachable devices only).
 */
#ifdef HAVE_DEVLIST
/*
 * Support for a alternate probe manager,
 * which will eliminate the boilerplate below.
 */
struct netdev_entry cs_drv =
	{cardname, cs_probe1, CS_IO_EXTENT, cs_portlist};
#else
int
cs8900_probe(struct device *dev)
{
	int i;
	int base_addr = dev ? dev->base_addr : 0;

#ifdef WYSE
	if (!WYSE_MARS) return -ENODEV;	/* Sometimes autodetection returns a false positive on Leo. */
#endif

	if (base_addr > 0x1ff)    /* Check a single specified location. */
		return cs_probe1(dev, base_addr);
	else if (base_addr != 0)  /* Don't probe at all. */
		return -ENXIO;

	for (i = 0; cs_portlist[i]; i++) {
		int ioaddr = cs_portlist[i];
		if (check_region(ioaddr, CS_IO_EXTENT))
			continue;
		if (cs_probe1(dev, ioaddr) == 0)
			return 0;
	}

	return -ENODEV;
}
#endif

static int
getPacketPage(int PageOffset)
{
	OUTW(PortPacketPagePtr, PageOffset);
	return(INW(PortPacketPageData));
}

static void
waitOnBusy(void)
{
	int	i;

	for (i = 0; i < 100000; ++i) { 
		if ((getPacketPage(SelfST) & NSS_SIBUSY) == 0)
			return;
	}
	printk("cs: timeout on wait for ready\n");
}

static void
setPacketPage(int PageOffset, ushort value)
{
	OUTW(PortPacketPagePtr, PageOffset);
	OUTW(PortPacketPageData, value);
	waitOnBusy();
}

void
ReadEEPROM(unsigned short *buf)
{
	int	i;

	for (i = 0; i < EEPROM_SIZE; ++i) {
		setPacketPage(EEPROMcmd, EEread | i);
		*buf++ = getPacketPage(EEPROMdata);
	}
}

void 
WriteEEPROM(unsigned short *buf)
{
	int	i;

	setPacketPage(EEPROMcmd, EEenable);
	for (i = 0; i < EEPROM_SIZE; ++i) {
		setPacketPage(EEPROMdata, *buf++);
		setPacketPage(EEPROMcmd, EEwrite | i);
	}
	setPacketPage(EEPROMcmd, EEdisable);
}

static void 
eeprom_init(void)
{
	unsigned char buf[EEPROM_SIZE * 2];
	
	ReadEEPROM((unsigned short *) buf);
#ifdef	WYSE
	if (WYSE_MARS) {
		if (memcmp(buf + 0x38, ea_addr, 3) == 0) 
			memcpy(ea_addr + 3, buf + 0x38 + 3, 3);
		else
			printk("cs8900: bad MAC prefix %02x-%02x-%02x\n",
			       buf[0x38], buf[0x39], buf[0x40]);
	}
#endif
}

static int
cs_reset(void)
{
	short x;
	int   i;

	for (i = 0; i < 10; ++i) {
		setPacketPage(SelfCTL, 0x100);		/* Reset Chip */
		x = getPacketPage(SelfST);
		if ((x & NSS_INITD) != 0) {	/* Wait till INITD */
			udelay(10000);
			/* turn off receiver/transmitter */
			setPacketPage(LineCTL, 0x13); /* default */
			if (x & (NSS_EEPROMEXIST | NSS_EEPROMOK))
				eeprom_init();
			return(0);
		}
	}
	printk("cs: SelfST = 0x%x\n", x);
	return(-1);				/* no card */
}

void
configReceiver(void)
{
	/* RxOKiE(100) CRCerroriE(1000) ExtraiE(4000) RuntiE(2000) */
	/* setPacketPage(RxCFG,0x0103); */
	setPacketPage(RxCFG,0x7103);

	/* Prom(80) BroadcastA(800) & individualA(400) & good CRC(100) */
	setPacketPage(RxCTL, 0x7d05);
}

void
configTransmitter(void)
{
	setPacketPage(TxCFG, 0x107);	/* TxOKie */
	setPacketPage(BufCFG, 0x10B); 	/* Rdy4TxiE */
}

void
configEA(unsigned char *ea) 
{
	short 	x;

	x = (short) ((ea_addr[1] << 8) + ea_addr[0]);
	setPacketPage(ChipEA0, x);

	x = (short) ((ea_addr[3] << 8) + ea_addr[2]);
	setPacketPage(ChipEA2, x);
	
	x = (short) ((ea_addr[5] << 8) + ea_addr[4]);
	setPacketPage(ChipEA4, x);
	
	memcpy(ea, ea_addr, 6);
}

/*
 * This is the real probe routine. Linux has a history of friendly device
 * probes on the ISA bus. A good device probes avoids doing writes, and
 * verifies that the correct device exists and functions.
 */
STATIC int 
cs_probe1(struct device *dev, int ioaddr)
{
	int i;

	if (cs_reset())
		return(-ENODEV);

	/* Allocate a new 'dev' if needed. */
	if (dev == NULL) {
		/*
		 * Don't allocate the private data here, it is done later
		 * This makes it easier to free the memory when this driver
		 * is used as a module.
		 */
		dev = init_etherdev(0, 0);
		if (dev == NULL)
			return -ENOMEM;
	}

	printk(KERN_INFO "%s: %s found at %#3x, ", dev->name, cardname, ioaddr);

	/* Fill in the 'dev' fields. */
	dev->base_addr = ioaddr;

	configEA((unsigned char *) dev->dev_addr);
	/* Retrieve and print the ethernet address. */
	for (i = 0; i < 6; i++)
		printk(" %2.2x", dev->dev_addr[i]);
	printk("\n");

#define jumpered_interrupts
#ifdef 	jumpered_interrupts
	/*
	 * If this board has jumpered interrupts, allocate the interrupt
	 * vector now. There is no point in waiting since no other device
	 * can use the interrupt, and this marks the irq as busy. Jumpered
	 * interrupts are typically not reported by the boards, and we must
	 * used autoIRQ to find them.
	 */
	printk("cs: probe with irq 0x%x\n", dev->irq);
	if (dev->irq == 0xff)
		;	/* Do nothing: a user-level program will set it. */
	else {
		int	irqval;

		dev->irq = CS_IRQ;
		irqval = request_irq(dev->irq, &cs_interrupt, 0, cardname, dev);
		if (irqval) {
			printk("%s: unable to get IRQ %d (irqval=%d).\n",
				   dev->name, dev->irq, irqval);
			return -EAGAIN;
		}
	}
#endif	/* jumpered interrupt */

	/* Initialize the device structure. */
	if (dev->priv == NULL) {
		dev->priv = kmalloc(sizeof(struct cs_local), GFP_KERNEL);
		if (dev->priv == NULL)
			return -ENOMEM;
	}

	memset(dev->priv, 0, sizeof(struct cs_local));

	/* Grab the region so that no one else tries to probe our ioports. */
	request_region(ioaddr, CS_IO_EXTENT, cardname);

	dev->open		= cs_open;
	dev->stop		= cs_close;
	dev->hard_start_xmit = cs_send_packet;
	dev->get_stats	= cs_get_stats;
	dev->set_multicast_list = &set_multicast_list;

	/* Fill in the fields of the device structure with ethernet values. */
	ether_setup(dev);

	return 0;
}

/*
 * Open/initialize the board. This is called (in the current kernel)
 * sometime after booting when the 'ifconfig' program is run.
 *
 * This routine should set everything up anew at each open, even
 * registers that "should" only need to be set once at boot, so that
 * there is non-reboot way to recover if something goes wrong.
 */
STATIC int
cs_open(struct device *dev)
{
	struct cs_local *lp = (struct cs_local *)dev->priv;
	int ioaddr = dev->base_addr;

	/* Reset the hardware here. Don't forget to set the station address. */

	configReceiver();
	configTransmitter();

	setPacketPage(IObase, ioaddr);
	setPacketPage(IRQnum, CS_INTRQ);	
	setPacketPage(LineCTL, 0x0d3); 	/* SerTxON SerRxON Auto(200) */
	setPacketPage(TestCTL, 0x0099); /* DisableLT */
	setPacketPage(BusCTL, 0x8017);  /* Enable IRQ */
#if 0
	printk("Crystal Line Status = 0x%x\n", getPacketPage(LineST));
	if ((getPacketPage(LineST) & 0x80) == 0) /* LineOK */
		printk("Crystal: No link pause\n");
#endif

	lp->open_time = jiffies;

	dev->tbusy = 0;
	dev->interrupt = 0;
	dev->start = 1;

	MOD_INC_USE_COUNT;

	return 0;
}

STATIC int
cs_send_packet(struct sk_buff *skb, struct device *dev)
{
	if (dev->tbusy) {
		/*
		 * If we get here, some higher level has decided we are broken.
		 * There should really be a "kick me" function call instead.
		 */
		int tickssofar = jiffies - dev->trans_start;
		if (tickssofar < 5)
			return 1;
		printk(KERN_WARNING "%s: transmit timed out, %s?\n", dev->name,
			   tx_done(dev) ? "IRQ conflict" : "network cable problem");
		/* Try to restart the adaptor. */
		dev->tbusy=0;
		dev->trans_start = jiffies;
	}
	/*
	 * If some higher layer thinks we've missed an tx-done interrupt
	 * we are passed NULL. Caution: dev_tint() handles the cli()/sti()
	 * itself.
	 */
	if (skb == NULL) {
		printk("cs: dev_tint\n");
		dev_tint(dev);
		return 0;
	}
	/*
	 * Block a timer-based transmit from overlapping. This could better be
	 * done with atomic_swap(1, dev->tbusy), but set_bit() works as well.
	 */
	if (set_bit(0, (void *) &dev->tbusy) != 0)
		printk(KERN_WARNING "%s: Transmitter access conflict.\n", dev->name);
	else {
		short x, length = ETH_ZLEN < skb->len ? skb->len : ETH_ZLEN;

		OUTW(PortTxCMD, 0x00C9); /* full frame */
		OUTW(PortTxLength, length);

		for ( ;; ) {
			x = getPacketPage(BusST);
			if (x & Rdy4TxNow)
				break;
		}
		OUTW(PortPacketPagePtr, TxFrame);
		outsw(PortData, skb->data, (length + 1) >> 1);
		dev->trans_start = jiffies;
#if 	0
		print_eth((char *) skb->data, length);
#endif
	}
	dev_kfree_skb (skb, FREE_WRITE);

	return 0;
}

/*
 * The typical workload of the driver:
 *   Handle the network interface interrupts.
 */
STATIC void
cs_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	struct device *dev = (struct device *)dev_id;
	struct cs_local *lp;
	int event, status;

	if (dev == NULL) {
		printk(KERN_WARNING "%s: irq %d.\n", cardname, irq);
		return;
	}
	dev->interrupt = 1;

	lp = (struct cs_local *)dev->priv;

	while((event = INW(PortISQ)) != 0) {
		switch(event & 0x3F) {
		case RxEventReg:
			if (event & RxOK)
				cs_rx(dev);
			else { 
				lp->stats.rx_errors++;
				printk("cs: RxOK?\n");
			}
			break;
		case TxEventReg:
			lp->stats.tx_packets++;
			dev->tbusy = 0;
			mark_bh(NET_BH);
			break;
		case BufEventReg:
			printk("Buf event 0x%x\n", event);
			lp->stats.tx_window_errors++;
			break;
		case RxMissEventReg:
			status = getPacketPage(RxMissCount);
			printk("cs: RxMiss event 0x%x = %d\n", event, status);
			lp->stats.rx_fifo_errors++;
			break;
		case TxCOLEventReg:
			status = getPacketPage(TxColCount);
			lp->stats.tx_fifo_errors++;
			lp->stats.collisions++;
			break;
		default:
			goto out;
		}
	}
out:
	dev->interrupt = 0;
	return;
}

/* We have a good packet(s), get it/them out of the buffers. */
STATIC void
cs_rx(struct device *dev)
{
	struct 	cs_local *lp = (struct cs_local *)dev->priv;
        int 	rxLength;
        int 	rxStatus;

	OUTW(PortPacketPagePtr, RxStatus);
	rxStatus = INW(PortData);
	rxLength = INW(PortData);

	/*
	** Check for packets that are too small or too large
	*/
	if (rxLength < 60 || rxLength > 1514) {
		lp->stats.rx_frame_errors++;
		printk("cs: rxLength = %d\n", rxLength);
	} else { 
		/* Malloc up new buffer. */
		struct sk_buff *skb;

		rxLength++;
		rxLength &= 0xfffffffe;
		skb = dev_alloc_skb(rxLength);
		if (skb == NULL) {
			printk(KERN_NOTICE "%s: dropping packet.\n",
					  dev->name);
			lp->stats.rx_dropped++;
		} else {
			unsigned char *buf;

			skb->dev = dev;
			buf = skb_put(skb, rxLength);
			insw(PortData, buf, rxLength >> 1);
#if 	0
			print_eth((char *) buf, rxLength);
#endif
			skb->protocol = eth_type_trans(skb, dev);
			netif_rx(skb);
			lp->stats.rx_packets++;
		}
	}

	/*
	 * If any worth-while packets have been received, dev_rint()
	 * has done a mark_bh(NET_BH) for us and will work on them
	 * when we get to the bottom-half routine.
	 */
	return;
}

/* The inverse routine to cs_open(). */
STATIC int
cs_close(struct device *dev)
{
	struct cs_local *lp = (struct cs_local *)dev->priv;

	lp->open_time = 0;

	dev->tbusy = 1;
	dev->start = 0;

	free_irq(dev->irq, NULL);

	/* Update the statistics here. */

	MOD_DEC_USE_COUNT;

	return 0;
}

/*
 * Get the current statistics.
 * This may be called with the card open or closed.
 */
STATIC struct enet_statistics *
cs_get_stats(struct device *dev)
{
	struct cs_local *lp = (struct cs_local *)dev->priv;

	return &lp->stats;
}

/*
 * Set or clear the multicast filter for this adaptor.
 * num_addrs == -1	Promiscuous mode, receive all packets
 * num_addrs == 0	Normal mode, clear multicast list
 * num_addrs > 0	Multicast mode, receive normal and MC packets,
 *			and do best-effort filtering.
 */
STATIC void
set_multicast_list(struct device *dev)
{
#if 0
	if (dev->flags&IFF_PROMISC)
	{
		/* Enable promiscuous mode */
	}
	else if((dev->flags&IFF_ALLMULTI) || dev->mc_count > HW_MAX_ADDRS)
	{
		/* Disable promiscuous mode, use normal mode. */
	}
	else if(dev->mc_count)
	{
		/* Walk the address list, and load the filter */
	}
#endif
}

