#include <linux/module.h>

#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>

#include <linux/flash_fs.h>

#include <asm/segment.h>

STATIC int
mapdirect(struct inode *inode, int index, int alloc)
{
	int	sectno;
	struct 	ffs_dirent *de;
	struct 	flash_device *fd = flash_device(inode->i_dev);

	de = ffs_get_dirent(inode);
	ASSERT(de);
	if (de == 0)
		return(-1);
	sectno = de->map[index];
	if (sectno == 0xFFFF) {
		if (alloc == 0)
			return(-1);
		if ((sectno = flash_alloc(fd, fd->pagesize)) < 0) 
			return(-1);
		de->map[index] = sectno;
		flash_write(fd, ino_sectno(inode->i_ino), 
			de2buf(de, ino_offset(inode->i_ino)), fd->pagesize);
	}
	return(sectno);
}

STATIC int
mapindex(struct inode *inode, int index, int indexsectno, int alloc)
{
	int	sectno;
	u_short	*indexbuf;
	struct 	flash_device *fd = flash_device(inode->i_dev);

	if (flash_read(fd, indexsectno, (char **) &indexbuf, fd->pagesize))
		return(-1); 
	if ((sectno = indexbuf[index]) == 0xFFFF) {
		if (alloc == 0)
			return(-1);
		if ((sectno = flash_alloc(fd, fd->pagesize)) < 0) 
			return(-1);
		indexbuf[index] = sectno;
		flash_write(fd, indexsectno, (char *) indexbuf, fd->pagesize);
	}
	return(sectno);
}

int
ffs_bmap(struct inode *inode, int offset, int alloc)
{
	int	nindex, i, j, sectno;
	struct 	flash_device *fd = flash_device(inode->i_dev);

	/* index page, directory pages are fixed pagesize */
	i = SECTNO(inode->i_blksize, offset);

	if (i < DIRMAP0)
		return(mapdirect(inode, i, alloc));

	/* nindex = number of pointers in an index page */
	nindex = fd->pagesize / sizeof(ffs_sect_t);
	i -= DIRMAP0;
	if (i < nindex) {	/* one level of index */
		sectno = mapdirect(inode, DIRMAP0, alloc);
		if (sectno < 0)
			return(-1);
		return(mapindex(inode, i, sectno, alloc));
	}

	i -= nindex;
	sectno = mapdirect(inode, DIRMAP0 + 1, alloc);
	if (sectno < 0)
		return(-1);
	j = i % nindex;
	i = i / nindex;
	ASSERT(i < nindex);
	sectno = mapindex(inode, i, sectno, alloc);
	if (sectno < 0)
		return(-1);
	return(mapindex(inode, j, sectno, alloc));
}

