#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/flash_fs.h>
#include <linux/stat.h>
#include <linux/string.h>

#include <asm/segment.h>

STATIC int ffs_readdir(struct inode *, struct file *, void *, filldir_t);

STATIC int ffs_dir_read(struct inode * inode, struct file * filp, char * buf, int count)
{
	return -EISDIR;
}

static struct file_operations ffs_dir_operations = {
	NULL,			/* lseek - default */
	ffs_dir_read,		/* read */
	NULL,			/* write - bad */
	ffs_readdir,		/* readdir */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	NULL,			/* mmap */
	NULL,			/* no special open code */
	NULL,			/* no special release code */
	file_fsync		/* default fsync */
};

/*
 * directories can handle most operations...
 */
struct inode_operations ffs_dir_iops = {
	&ffs_dir_operations,	/* default directory file-ops */
	ffs_create,		/* create */
	ffs_lookup,		/* lookup */
	ffs_link,		/* link */
	ffs_unlink,		/* unlink */
	ffs_symlink,		/* symlink */
	ffs_mkdir,		/* mkdir */
	ffs_rmdir,		/* rmdir */
	ffs_mknod,		/* mknod */
	ffs_rename,		/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* readpage */
	NULL,			/* writepage */
	NULL,			/* bmap */
	ffs_truncate,		/* truncate */
	NULL			/* permission */
};

STATIC int ffs_readdir(struct inode * inode, struct file * filp,
	void * dirent, filldir_t filldir)
{
	unsigned int offset, i;
	int	sectno;
	struct 	ffs_dirent *de;
	struct 	flash_device *fd = flash_device(inode->i_dev);
	char	*sectbuf;

	if (!inode || !inode->i_sb || !S_ISDIR(inode->i_mode))
		return -EBADF;
	offset = filp->f_pos;
	for ( ;; ) {
		sectno = ffs_bmap(inode, offset, 0);
		if (sectno < 0)
			break;
		if (flash_read(fd, sectno, &sectbuf, fd->pagesize))
			return -EIO;
		i = offset & (fd->pagesize - 1);
		offset -= i;
		for ( ;; ) {
			de = (struct ffs_dirent *) (sectbuf + i); 
			if (de->entsize == 0xFF) {
				offset += fd->pagesize;
				break;		/* end of sector */
			}
			if (de->mode == 0) {	/* deleted entry */
				i += de->entsize;
				continue;
			}
			if (filldir(dirent, de->name, strlen(de->name), 
				    filp->f_pos, mk_ino(sectno, i)) < 0)
				return 0;
			i += de->entsize;
			ASSERT(i <= fd->pagesize);
			filp->f_pos = offset + i;
		}
	}
	return 0;
}

struct ffs_dirent *
ffs_get_dirent(struct inode *inode)
{
	char	*buf;
	int	offset;
	struct 	ffs_dirent *de;
	struct 	flash_device *fd = flash_device(inode->i_dev);

	ASSERT(inode);

	if (flash_read(fd, ino_sectno(inode->i_ino), &buf, fd->pagesize))
		return(0);
	offset = ino_offset(inode->i_ino);
	de = (struct ffs_dirent *) (buf + offset);
	
	ASSERT(de->entsize > 0 && de->entsize < 0xFF);
#ifdef	LATER
	ASSERT(de->mode == inode->i_mode);
	ASSERT(de->nlink == inode->i_nlink);
#endif
	return(de);
}
