#ifndef _LINUX_FLASH_H
#define _LINUX_FLASH_H

/*
; Flash Control Signals
; (Note WP and CE are active low)
*/

#define	FLASH_DAT        0
#define	FLASH_CLE        1
#define	FLASH_ALE        2
#define	FLASH_WP         4
#define	FLASH_CE         8
#define	FLASH_CE2        16

#define	FLASH_DATA       FLASH_DAT
#define	FLASH_DATAW      FLASH_DAT + FLASH_WP
#define	FLASH_COMMAND    FLASH_CLE
#define	FLASH_COMMANDW   FLASH_CLE + FLASH_WP
#define	FLASH_ADDRESS    FLASH_ALE
#define	FLASH_ADDRESSW   FLASH_ALE + FLASH_WP
#define	FLASH_STANDBY    FLASH_CE + FLASH_CE2

#define	FLASH_MASK       0xe0

/*
; Flash Status Signals
*/

#define	FLASH_READY0     0x02
#define	FLASH_READY1     0x01
#define	FLASH_READYMASK  FLASH_READY0 + FLASH_READY1

/*
; Flash Commands
*/

#define	FLASH_CMD_READ_MODE1	0x00
#define	FLASH_CMD_READ_MODE2    0x50
#define	FLASH_CMD_WRITE         0x80
#define	FLASH_CMD_STATUS_READ   0x70
#define	FLASH_CMD_ID_READ       0x90
#define	FLASH_CMD_AUTO_PROGRAM  0x10
#define	FLASH_CMD_ERASE    	0x60
#define	FLASH_CMD_ERASE2   	0xd0
#define	FLASH_CMD_SUSPEND       0xb0
#define	FLASH_CMD_RESUME        0xd0
#define	FLASH_CMD_RESET         0xff


/*
; Flash Status Bits
; (The write protect bit is active low)
*/

#define	FLASH_STS_FAIL           0x01
#define	FLASH_STS_SUSPEND        0x20
#define	FLASH_STS_READY          0x40
#define	FLASH_STS_WRITE_PROTECT  0x80

/*
; Flash Error Return Codes
*/
#define	FLASH_ERR_NONE           0
#define	FLASH_ERR_BADCKS         -1
#define	FLASH_ERR_DELETED        -2
#define	FLASH_ERR_CHECKSUM       -3
#define	FLASH_ERR_TIMEOUT        -4
#define	FLASH_ERR_STATUS       	 -5
#define	FLASH_ERR_EOF       	 -6

/*
; Flash Sizes
*/
#define	FLASH_PAGE_SIZE          512
#define	FLASH_RDNT_SIZE          8
#define	FLASH_PAGES_PER_BLOCK    16
#define	FLASH_BLOCK_SIZE         (FLASH_PAGE_SIZE * FLASH_PAGES_PER_BLOCK)

#define	FLASH_BLOCKS             512
#define	FLASH_PAGES              (FLASH_BLOCKS * FLASH_PAGES_PER_BLOCK)

#define	FLASH_BLOCK2PAGE(blkno)	(blkno * FLASH_PAGES_PER_BLOCK)

#define	FLASH_SECTSIZE	4096
#define	NFLASHS		1
#define	FLASH_NCACHES	8

typedef	struct	flash_cache {
	char	*buf;
	ushort	sectsize;
	ushort	sectno;
	int	access;
	int	dirty;
} FDCACHE;

typedef	struct	flash_device {
	int	blocks, pages, blocksize, pagesize;
	int	badpages, freepages, inusepages, deletepages;
	int	sectindx, freeindx;
	int	ppb, sectsize, rdntsize, initok;
	u_char	*blkdelcnt;		/* deleted page count in a block */
	u_char	*freemap;		/* free logical sector map */
	u_short	*mapping;		/* from logical sector -> pageno */
	FDCACHE	cache[FLASH_NCACHES];
} FLASHDEV;

#define	SBSECTNO	2
#define	SBBLOCKNO	2
#define	SBPAGENO	(fd->ppb * SBBLOCKNO)

int 	flash_alloc(FLASHDEV *fd, int size);
void 	flash_delete(FLASHDEV *fd, int sectno);
void	flash_del_index(FLASHDEV *fd, int sectno, int level);
int 	flash_remap(FLASHDEV *fd, int sectno);
int 	flash_read(FLASHDEV *fd, int sectno, char **buf, int sectsize);
int 	flash_write(FLASHDEV *fd, int sectno, char *buf, int sectsize);
void	flash_sync(FLASHDEV *fd);
void 	flash_invalidate(FLASHDEV *fd, int sectno);
FLASHDEV * flash_device(int dev);

#endif
