#ifndef _LINUX_FLASH_FS_H
#define _LINUX_FLASH_FS_H

#include <linux/flash.h>
#include <linux/debug.h>

#define	FLASH_NAMELEN	127
#define	FLASH_DIRSIZE	256

#define	DIRMAP0	5		/* flash_fs_i.h depends on this */

typedef	u_short	ffs_sect_t;

typedef	struct 	ffs_dirent {
	u_char		entsize;
	u_char		noff;
	u_short		mode;
	int		size;
	kdev_t		rdev;		/* for BLK or CHR node */
	u_short		nlink;		/* meaningful on for non-directory */
	u_short 	unuse;
	ffs_sect_t	map[DIRMAP0+3];
	char		name[1];
} DIRENT;

#define	de2buf(de, offset)	((char *) (((int) de) - offset))

static inline ino_t mk_ino(int secno, int offset)
{
	return (secno << 16) | offset;
}

static inline ino_t ino_sectno(ino_t ino)
{
	return ino >> 16;
}

static inline int ino_offset(ino_t ino)
{
	return ino & 0xFFFF;
}

#define	SECTNO(sectsize, off)	((off) / sectsize)
#define	SECTOFF(sectsize, off)	((off) % sectsize)

/*
 * Function prototypes
 */
extern int init_flash_fs(void);
extern int ffs_lookup(struct inode * dir,const char * name, int len,
	               struct inode ** result);
extern int ffs_create(struct inode * dir,const char * name, int len, int mode,
	struct inode ** result);
extern int ffs_mkdir(struct inode * dir, const char * name, int len, int mode);
extern int ffs_rmdir(struct inode * dir, const char * name, int len);
extern int ffs_unlink(struct inode * dir, const char * name, int len);
extern int ffs_symlink(struct inode * inode, const char * name, int len,
	const char * symname);
extern int ffs_link(struct inode * oldinode, struct inode * dir, const char * name, int len);
extern int ffs_mknod(struct inode * dir, const char * name, int len, int mode, int rdev);
extern int ffs_rename(struct inode * old_dir, const char * old_name, int old_len,
	struct inode * new_dir, const char * new_name, int new_len, int must_be_dir);
extern struct inode * ffs_new_inode(struct inode * dir, const char *, mode_t mode);
extern int ffs_free_inode(struct inode * inode);

extern int ffs_bmap(struct inode *, int, int);

extern int ffs_file_read(struct inode *, struct file *, char *, int);
extern int ffs_file_write(struct inode *, struct file *, const char *, int);

extern void ffs_truncate(struct inode *);
extern void ffs_put_super(struct super_block *);
extern struct super_block *ffs_read_super(struct super_block *,void *,int);
extern int init_ffs_fs(void);
extern void ffs_write_super(struct super_block *);
extern void ffs_read_inode(struct inode *);
extern int ffs_notify_change(struct inode *, struct iattr *);
extern void ffs_write_inode(struct inode *);
extern void ffs_read_inode(struct inode *);
extern void ffs_put_inode(struct inode *);
extern void ffs_statfs(struct super_block *, struct statfs *, int);
extern int ffs_sync_inode(struct inode *);
extern int ffs_sync_file(struct inode *, struct file *);
extern int ffs_mmap(struct inode *, struct file *, struct vm_area_struct *);
extern struct ffs_dirent *ffs_get_dirent(struct inode *inode); 
extern void ffs_set_iops(struct inode *inode);

extern	struct inode_operations	ffs_dir_iops, ffs_file_iops, ffs_symlink_iops;

#endif
