#ifdef	__KERNEL__
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/malloc.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>

#include <linux/flash_fs.h>
#include <linux/flash.h>

#include <asm/segment.h>
#else
#include "system.h"
#include "errno.h"
#include "flash_fs.h"
#include "flash.h"
#endif

#define	FFS_BUFFERS 	12

int	ffs_timestamp;
struct	ffs_buf	ffs_buffers[FFS_BUFFERS];

struct ffs_buf *
ffs_balloc(struct inode *inode, int sectno)
{
	int	i;
	struct 	ffs_buf *bp, *lp;
	
	for (lp = bp = ffs_buffers, i = 0; i < FFS_BUFFERS; ++i, ++bp) {
		ASSERT(bp->fd != inode->i_ffs_fd ||
		       bp->ino != inode->i_ino ||
		       bp->sectno != sectno);
		if (bp->access < lp->access)	/* least recent used */
			lp = bp;
	}

	if (lp->dirty)
		flash_write(lp);

	ASSERT(inode->i_ffs_fd);
	lp->buf = lp->buffer;
	lp->fd  = inode->i_ffs_fd;
	lp->ino = inode->i_ino;
	ASSERT(lp->ino);
	if (flash_alloc(lp) == 0) {
		lp->fd = 0;
		lp->access = 0;
		lp->dirty = 0;
		return(0);
	}
	inode->i_ffs_ip->map[sectno] = lp->pageno;
	lp->dirty = 1;
	lp->sectno = sectno;
	lp->access = ++ffs_timestamp;
	lp->sequence = 1;
	if (S_ISDIR(inode->i_mode))
		memset(lp->buf, 0xFF, FFS_BLKSIZE(inode));
	else
		memset(lp->buf, 0x00, FFS_BLKSIZE(inode));

	return(lp);
}

struct ffs_buf *
ffs_bread(struct inode *inode, int sectno, int pageno)
{
	int	i;
	struct 	ffs_buf *bp, *lp;
	
	for (lp = bp = ffs_buffers, i = 0; i < FFS_BUFFERS; ++i, ++bp) {
		if (bp->fd == inode->i_ffs_fd &&
		    bp->ino == inode->i_ino &&
		    bp->sectno == sectno) {
			bp->access = ++ffs_timestamp; 
			return(bp);
		}
		if (bp->access < lp->access)	/* least recent used */
			lp = bp;
	}

	if (lp->dirty)
		flash_write(lp);

	lp->buf = lp->buffer;
	lp->fd  = inode->i_ffs_fd;
	lp->ino = inode->i_ino;
	lp->sectno = sectno;
	lp->pageno = pageno;
	lp->dirty = 0;
	
	if (flash_read(lp)) {
		lp->fd = 0;
		lp->access = 0;
		return(0);
	}

	lp->access = ++ffs_timestamp;
	ASSERT(!S_ISDIR(inode->i_mode) || lp->buf[0] != 0);

	return(lp);
}

#ifdef	__KERNEL__
struct ffs_buf *
ffs_bpage(struct inode *inode, int sectno, int pageno)
{
	int	i;
	struct 	ffs_buf *bp, *lp;
	extern	char	*ffs_pagebuf;
	
	for (lp = bp = ffs_buffers, i = 0; i < FFS_BUFFERS; ++i, ++bp) {
		if (bp->fd == inode->i_ffs_fd &&
		    bp->ino == inode->i_ino &&
		    bp->sectno == sectno) {
			bp->access = ++ffs_timestamp; 
			memcpy(ffs_pagebuf, bp->buf, FFS_BLKSIZE(inode));
			return(bp);
		}
		if (bp->access < lp->access)	/* least recent used */
			lp = bp;
	}

	if (lp->dirty)
		flash_write(lp);

	lp->fd  = inode->i_ffs_fd;
	lp->ino = inode->i_ino;
	lp->sectno = sectno;
	lp->pageno = pageno;
	lp->buf = ffs_pagebuf;
	lp->dirty = 0;
	lp->access = 0;
	
	if (flash_read(lp)) {
		lp->fd = 0;
		return(0);
	}
	lp->fd = 0;
	return(lp);
}
#endif

void
ffs_bwrite(struct ffs_buf *bp)
{
	ASSERT(bp->dirty);
	bp->access = 0;
	flash_write(bp);	/* write out at end of buf */
}

struct ffs_buf *
ffs_bremap(struct inode *inode, int sectno, int pageno)
{
	struct ffs_buf *bp;

	bp = ffs_bread(inode, sectno, pageno);
	/* if we found dirty buf, we can keep writting to it */
	/* this is where a lot of remap are reduced for small writes */
	/* dirty buf get flushed on sync(), close() or least-recent-used */
	if (bp == 0 || bp->dirty)
		return(bp);

	bp->dirty = 2;		/* flag to write new & delete old */
	return(bp);
}

void
ffs_bsync(struct inode *inode)
{
	int	i;
	struct 	ffs_buf *bp;
	
	for (bp = ffs_buffers, i = 0; i < FFS_BUFFERS; ++i, ++bp) {
		if (bp->dirty == 0) {
			if (inode == 0) {
				bp->access = 0;
				bp->fd = 0;
			}
			continue;
		}
		bp->access = 0;
		if (inode == 0) {		/* called from umount */
			flash_write(bp);
			bp->fd = 0;		/* invalid all buffers */
			bp->dirty = 0;
			continue;
		}
		if (inode->i_ffs_fd == bp->fd)
			flash_write(bp);
	}
}

void
ffs_bdelete(struct inode *inode, int sectno, int pageno)
{
	int	i, dirty;
	struct 	ffs_buf *bp;
	
	for (bp = ffs_buffers, i = 0; i < FFS_BUFFERS; ++i, ++bp) {
		if (bp->fd == inode->i_ffs_fd &&
		    bp->ino == inode->i_ino &&
		    bp->sectno == sectno) {
			dirty = bp->dirty;
			bp->fd = 0;	/* invalidate buf */
			bp->access = 0;	/* first one to reuse */
			bp->dirty = 0;
			if (dirty == 1) { /* never been written */
				flash_free(inode->i_ffs_fd, pageno);
				return;
			}
			break;
		}
	}
	flash_delete(inode->i_ffs_fd, pageno);
}

#ifdef	WYSE_DEBUG
void
ffs_bdump(void)
{
	int	i;
	struct 	ffs_buf *bp;

	for (bp = ffs_buffers, i = 0; i < FFS_BUFFERS; ++i, ++bp) {
		PRINTF("buf %x, fd %x, ino %x, sect %x, page %x, dirt %x, %x\n",
			(int) bp, (int) bp->fd, bp->ino, 
			bp->sectno, bp->pageno, bp->dirty, bp->access);
	}
}
#endif
