#ifdef	__KERNEL__
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/flash_fs.h>
#include <linux/flash.h>
#include <linux/stat.h>
#include <linux/string.h>

#include <asm/segment.h>

STATIC int ffs_readdir(struct inode *, struct file *, void *, filldir_t);

STATIC int 
ffs_dir_read(struct inode * inode, struct file * filp, char * buf, int count)
{
	return -EISDIR;
}

static struct file_operations ffs_dir_operations = {
	NULL,			/* lseek - default */

	ffs_dir_read,		/* read */
	NULL,			/* write - bad */
	ffs_readdir,		/* readdir */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	NULL,			/* mmap */
	NULL,			/* no special open code */
	NULL,			/* no special release code */
	file_fsync		/* default fsync */
};

/*
 * directories can handle most operations...
 */
struct inode_operations ffs_dir_iops = {
	&ffs_dir_operations,	/* default directory file-ops */
	ffs_create,		/* create */
	ffs_lookup,		/* lookup */
	ffs_link,		/* link */
	ffs_unlink,		/* unlink */
	ffs_symlink,		/* symlink */
	ffs_mkdir,		/* mkdir */
	ffs_rmdir,		/* rmdir */
	ffs_mknod,		/* mknod */
	ffs_rename,		/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* readpage */
	NULL,			/* writepage */
	NULL,			/* bmap */
	NULL,			/* ffs_truncate	*/
	NULL			/* permission */
};

STATIC int 
ffs_readdir(struct inode * inode, struct file * filp,
	void * dirent, filldir_t filldir)
{
	unsigned int offset, i;
	struct 	ffs_dirent *de;
	struct	ffs_buf	*bp;

	ASSERT(inode != 0 && S_ISDIR(inode->i_mode));

	offset = filp->f_pos;
	switch (offset) {
	case 0:
		if (filldir(dirent, ".", 1, 0, inode->i_ino) < 0)
			return 0; 
		filp->f_pos = 1;	/* drop thru */
	case 1:
		if (filldir(dirent, "..", 2, 1, inode->i_ffs_ip->parent) < 0)
			return 0; 
		filp->f_pos = 2;
	case 2:
		offset = 0;
		break;
	}

	for ( ;; ) {
		bp = ffs_getblk(inode, offset, 0);
		if (bp == 0)
			break;
		i = offset & (FFS_BLKSIZE(inode) - 1);
		offset -= i;
		for ( ;; ) {
			de = (struct ffs_dirent *) (bp->buf + i); 
			ASSERT(de->entsize > 0);
			if (de->entsize == 0xFF) {
				break;		/* end of sector */
			}
			if (de->ino == 0) {	/* deleted entry */
				i += de->entsize;
				continue;
			}
			filp->f_pos = offset + i;
			if (filldir(dirent, DE_NAME(de), strlen(DE_NAME(de)), 
				    filp->f_pos, de->ino) < 0) {
				if (filp->f_pos <= 0)
					filp->f_pos = 2;
				return 0;
			}
			i += de->entsize;
			if (i >= FFS_BLKSIZE(inode)) {
				ASSERT(i == FFS_BLKSIZE(inode));
				break;
			}
		}
		offset += FFS_BLKSIZE(inode);
		filp->f_pos = offset;
	}
	return 0;
}
#else
#include "system.h"
#include "errno.h"
#include "flash_fs.h"
#include "flash.h"
#endif

struct ffs_dirent *
ffs_get_dirent(struct inode *inode, int dirty)
{
	unsigned int offset, i;
	struct 	ffs_dirent *de;
	struct	ffs_buf	*bp;
	static	struct	inode	parent;

	ASSERT(inode);
	ASSERT(inode->i_ino);
	ASSERT(inode->i_ino != FLASH_ROOTINO);
	ASSERT(inode->i_ffs_ip && inode->i_ffs_ip->parent);
	parent.i_ino = inode->i_ffs_ip->parent;
	parent.i_ffs_fd = inode->i_ffs_fd;
	parent.i_blksize = inode->i_blksize;
	parent.i_ffs_ip = flash_getimap(parent.i_ffs_fd, parent.i_ino);

	for (offset = 0; ; offset += FFS_BLKSIZE(&parent))  {
		bp = ffs_getblk(&parent, offset, 0);
		if (bp == 0)
			break;
		for (i = 0; i < FFS_BLKSIZE(&parent); i += de->entsize) {
			de = (struct ffs_dirent *) (bp->buf + i); 
			ASSERT(de->entsize != 0);
			if (de->entsize == 0xFF)
				break;		/* end of sector */
			if (de->ino == inode->i_ino) {
				if (bp->dirty == 0)
					bp->dirty = dirty;
				return(de);
			}
		}
	}
	return(0);
}

struct ffs_dirent *
ffs_find_dirent(struct inode *dir, const char *name, int len, int dirty)
{
	int	offset, off;
	struct	ffs_buf *bp;
	struct	ffs_dirent *de;

	for (offset = 0; ; offset += FFS_BLKSIZE(dir)) {
		bp = ffs_getblk(dir, offset, 0);
		if (bp == 0)		/* EOF on dir */
			break;;	/* NOENT */
		for (off = 0; off < FFS_BLKSIZE(dir); off += de->entsize) {
			de = (struct ffs_dirent *) (bp->buf + off);
			if (de->entsize == 0xFF) 
				break;
			ASSERT(de->entsize != 0);
			ASSERT(de->mode != 0xFFFF);
			if (de->ino == 0x0000)	/* deleted entry */
				continue;
			if (DE_NAME(de)[len])
				continue;
			if (memcmp(DE_NAME(de), name, len) != 0)
				continue;
			if (bp->dirty == 0)
				bp->dirty = dirty;
			return(de);
		}
	}
	return(0);
}

void
ffs_update_dirent(struct inode *inode)
{
	struct 	ffs_dirent *de;

	if (inode->i_ino == FLASH_ROOTINO)	/* never update root inode */
		return;
	de = ffs_get_dirent(inode, 2);	/* set dirty flag for re-write */
	if (de == 0)
		return;
	de->size = inode->i_size;
	ASSERT(inode->i_ffs_ip->nlink == inode->i_nlink);
	de->mode = inode->i_mode;
	de->uid = inode->i_uid;
	de->gid = inode->i_gid;
	ASSERT(inode->i_nlink > 0);
}

struct ffs_dirent *
ffs_empty_dirent(struct inode *dir, int entsize)
{
	unsigned int off, offset;
	struct 	ffs_dirent *de;
	struct	ffs_buf	*bp;

	if (entsize & 1)
		entsize++;
	ASSERT(entsize < 0xFF);

	for (offset = 0; ; offset += FFS_BLKSIZE(dir)) {
		bp = ffs_getblk(dir, offset, 1);
		if (bp == 0)		/* ENOSPC on dir */
			break;

		for (off = 0; off < FFS_BLKSIZE(dir); off += de->entsize) {
			de = (struct ffs_dirent *) (bp->buf + off);
			ASSERT(de->entsize != 0);
			if (de->entsize != 0xFF) {
				if (de->ino || de->entsize < entsize)
					continue;
			} else {
				if (FFS_BLKSIZE(dir) - off < entsize)
					break;
				ASSERT(de->size == -1);
				ASSERT(de->mode == 0xFFFF);
				ASSERT(DE_NAME(de)[0] == -1);
				de->entsize = entsize;
			}
			if (bp->dirty == 0)
				bp->dirty = 2;
			return(de);
		}
	}
	return(0);
}

int
ffs_rename_dirent(struct inode *old_dir, ino_t ino, 
		  struct inode *new_dir, const char *new_name, int new_len)
{
	unsigned int offset, i, entsize;
	struct 	ffs_dirent *old_de, *new_de;
	struct	ffs_buf	*bp;

	ASSERT(ino);
	ASSERT(ino != FLASH_ROOTINO);

	for (offset = 0; ; offset += FFS_BLKSIZE(old_dir))  {
		bp = ffs_getblk(old_dir, offset, 0);
		ASSERT(bp);
		if (bp == 0)
			break;
		for (i = 0; i < FFS_BLKSIZE(old_dir); i += old_de->entsize) {
			old_de = (struct ffs_dirent *) (bp->buf + i); 
			ASSERT(old_de->entsize != 0);
			if (old_de->entsize == 0xFF)
				break;		/* end of sector */
			if (old_de->ino == ino) {
				ASSERT(new_len == strlen(new_name));
				entsize = new_len + sizeof(*new_de);
				new_de = ffs_empty_dirent(new_dir, entsize); 
				if (new_de == 0)
					return(0);
				entsize = new_de->entsize;
				*new_de = *old_de;
				new_de->entsize = entsize;
				strcpy(DE_NAME(new_de), new_name);
				if (bp->dirty == 0)
					bp->dirty = 2;
				old_de->ino = 0;
				return((int) new_de);
			}
		}
	}
	return(0);
} 
