#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/pagemap.h>

#include <asm/segment.h>
#include <asm/io.h>
#include <asm/bitops.h>

#include <linux/fs.h>
#include <linux/flash_fs.h>
#include <linux/flash.h>

STATIC int ffs_readpage(struct inode * inode, struct page * page);
/*
 * We have mostly NULL's here: the current defaults are ok for
 * the coh filesystem.
 */
static struct file_operations ffs_file_operations = {
	NULL,			/* lseek - default */
	ffs_file_read,		/* read */
	ffs_file_write,		/* write */
	NULL,			/* readdir - bad */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	generic_file_mmap,	/* mmap */
	NULL,			/* no special open is needed */
	NULL,			/* release */
	ffs_sync_file		/* fsync */
};

struct inode_operations ffs_file_iops = {
	&ffs_file_operations,	/* default file operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	ffs_readpage,		/* readpage */
	NULL,			/* writepage */
	NULL,			/* bmap */
	NULL,			/* ffs_truncate	*/
	NULL			/* permission */
};

STATIC void
memset_tofs(char *addr, char val, int count)
{
	static	char	zeros[16];

	while(count > sizeof(zeros)) {
		memcpy_tofs(addr, zeros, sizeof(zeros));
		addr += sizeof(zeros);
		count -= sizeof(zeros);
	}
	if (count > 0)
		memcpy_tofs(addr, zeros, count);
}

int 
ffs_file_read(struct inode * inode, struct file * filp, char * buf, int count)
{
	char	*b0 = buf;
	int	size, c, off, offset;
	struct	ffs_buf  *bp;

	ASSERT(inode);
	ASSERT(S_ISREG(inode->i_mode));

	offset = filp->f_pos;
	size = inode->i_size - offset;
	if (count > size)
		count = size;

	while(count > 0) {
		off = SECTOFF(offset);
		c = FFS_BLKSIZE(inode) - off;
		if (c > count)
			c = count;
		bp = ffs_getblk(inode, offset, 0);
		if (bp) 
			memcpy_tofs(buf, bp->buf + off, c);
		else 
			memset_tofs(buf, 0, c);
		buf += c;
		count -= c;
		offset += c;
	}
	filp->f_pos = offset;
	
	return(buf - b0);
}
	
int
ffs_file_write(struct inode * inode, struct file * filp, const char * buf, int count)
{
	const char *b0 = buf;
	int	c, off, offset;
	struct	ffs_buf  *bp;

	ASSERT(inode);
	ASSERT(S_ISREG(inode->i_mode));

	if (filp->f_flags & O_APPEND)
		offset = inode->i_size;
	else
		offset = filp->f_pos;
	ASSERT(offset >= 0);

	while(count > 0) {
		off = SECTOFF(offset);
		c = FFS_BLKSIZE(inode) - off;
		if (c > count)
			c = count;
		bp = ffs_getblk(inode, offset, 1);
		if (bp == 0)
			return -ENOSPC;
		memcpy_fromfs(bp->buf + off, buf, c);

		buf += c;
		count -= c;
		offset += c;

		if (offset > inode->i_size) {
			if ((off + c) >= FFS_BLKSIZE(inode))
				ffs_bwrite(bp);
			ASSERT(inode->i_nlink <= 1);
			inode->i_dirt = 1;
			inode->i_size = offset;
		} else {
			ffs_bwrite(bp);
		}
	}

	filp->f_pos = offset;
	return(buf - b0);
}

char	*ffs_pagebuf;

STATIC int 
ffs_readpage(struct inode * inode, struct page * page)
{
	struct	ffs_buf  *bp;
	int	offset;

	ffs_pagebuf = (char *) page_address(page);
	clear_bit(PG_error, &page->flags);

	offset = page->offset;
	ASSERT(SECTOFF(offset) == 0);

	bp = ffs_getblk(inode, offset, -1);
	if (bp == 0) {  
		printk("ffs_readpage(%d): filesize 0x%x hole at 0x%x\n",
			(int) inode->i_ino, inode->i_size, offset);
		memset(ffs_pagebuf, 0, PAGE_SIZE);
	}
	
	set_bit(PG_uptodate, &page->flags);
	return(0);
}

int 
ffs_sync_file(struct inode * inode, struct file * file)
{
	if (!(S_ISREG(inode->i_mode) || S_ISDIR(inode->i_mode) ||
	     S_ISLNK(inode->i_mode)))
		return -EINVAL;
	ffs_write_inode(inode);
	return(0);
}
