#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/flash_fs.h>
#include <linux/stat.h>

#include <asm/segment.h>

static int ffs_readlink(struct inode *, char *, int);
static int ffs_follow_link(struct inode *, struct inode *, int, int, struct inode **);

/*
 * symlinks can't do much...
 */
struct inode_operations ffs_symlink_iops = {
	NULL,			/* no file-operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	ffs_readlink,		/* readlink */
	ffs_follow_link,	/* follow_link */
	NULL,			/* readpage */
	NULL,			/* writepage */
	NULL,			/* bmap */
	NULL,			/* truncate */
	NULL			/* permission */
};

static int ffs_follow_link(struct inode * dir, struct inode * inode,
	int flag, int mode, struct inode ** res_inode)
{
	int 	error;
	struct	ffs_buf	*bp;

	*res_inode = NULL;
	if (!dir) {
		dir = current->fs->root;
		dir->i_count++;
	}
	if (!inode) {
		iput(dir);
		return -ENOENT;
	}
	if (!S_ISLNK(inode->i_mode)) {
		iput(dir);
		*res_inode = inode;
		return 0;
	}
	if (current->link_count > 5) {
		iput(inode);
		iput(dir);
		return -ELOOP;
	}

	bp = ffs_getblk(inode, 0, 0);
	if (bp == 0) {
		iput(inode);
		iput(dir);
		return -EIO;
	}
	iput(inode);
	current->link_count++;
	error = open_namei(bp->buf,flag,mode,res_inode,dir);
	current->link_count--;
	return error;
}

static int ffs_readlink(struct inode * inode, char * buffer, int buflen)
{
	struct	ffs_buf	*bp;
	int 	i;

	if (!S_ISLNK(inode->i_mode)) {
		iput(inode);
		return -EINVAL;
	}
	if (buflen > FFS_BLKSIZE(inode))
		buflen = FFS_BLKSIZE(inode);
	bp = ffs_getblk(inode, 0, 0);
	iput(inode);
	if (!bp)
		return 0;
	i = 0;
	while (i < buflen && bp->buf[i]) {
		put_user(bp->buf[i], buffer++);
		i++;
	}
	if (i < buflen)
		put_user(0, buffer++);
	return i;
}

