#ifndef _LINUX_FLASH_H
#define _LINUX_FLASH_H

//	Fram, 2 * (nand + boot), 2 * fcard, fnor
#define	NFLASHS			8	
#define	FLASH_FORMAT		0x80	/* bit in the minor device */
#define	FLASH_AUTOFMT		0x40
#define	FLASH_MINOR(dev)	((dev) & 0x3F)

#define	NFRAMS			1
#define	FRAM_DEV0		&flashs[0]

#define	NFNANDS			4
#define	FNAND_DEV0		&flashs[1]	// on-board nand dev 0
#define	FNAND_DEV1		&flashs[2]	// on-board dev 0, partition 1
#define	FNAND_DEV2		&flashs[3]	// on-board nand dev 1
#define	FNAND_DEV3		&flashs[4]	// on-board dev 1, partition 1

#define	NFCARDS			2
#define	FCARD_DEV0		&flashs[5]
#define	FCARD_DEV1		&flashs[6]

#define	NFNORS			1
#define	FNOR_DEV0		&flashs[7]

#define	FLASH_INODES		fd->allinodes
#define	FLASH_DIRINO		fd->dirinodes
#define	FLASH_ROOTBLKNO		fd->rootblockno
#define	FLASH_ROOTPAGENO	fd->rootpageno

#define	FLASH_ROOTINO		2
#define	FLASH_IMAPSIZE		8
#define	FLASH_CHECKSUM		0xA55A

#define	FLASH_RDNTSIZE		16

typedef	u_short	flash_sect_t;

struct  flash_super_block {
	char			signature[4];	/* Wyse */		
	int			version;	/* 0 */
	int			reserved[2];
	u_short			badblocks;
	u_short			badlist[1];
};

struct	flash_imap {
	u_short			nlink;
	u_short 		mapsize;
	ino_t			parent;
	flash_sect_t 		*map;
};

typedef	struct areab {
	u_char	inuse;		/* Mars inuse flag */
	u_char	sequence;	/* actual sequence */
	u_short	sectno;		/* file logical sector number */
	u_short	nextno;		/* next physical page in this block */
	u_short	checksum;	/* so far, Mars compatible */	

	u_short	pageno;		/* actual physicall pageno of this page */
	u_short	ino;		/* file id, 0xFFFF if unused, 0x0000 if del */
	u_short firstno;	/* pageno of the first page in the sector */
	u_short	size;		/* block size after compressed */
} AREAB;

typedef	struct	flash_device {
	int			blocks, pages, blocksize, pagesize, rdntsize;
	int			rawsize, zipsize;
	int			badpages, freepages, inusepages, deletepages;
	int			inuseinodes, freeinodes;
	int			allinodes, dirinodes, rootblockno, rootpageno;
	int			inoindx, freeindx;
	int			ppb, pgend, blkend, blkshift;
	int			port;
	kdev_t			dev;
	u_char			*blkdelcnt;	
	u_char			*freemap;	
	flash_sect_t		*badblocks;
	struct 	flash_imap 	*inodes;
	int	(*read)		(struct flash_device *fd, int pageno, char *buf, AREAB *b);
	int	(*readb) 	(struct flash_device *fd, int pageno, AREAB *b);
	int	(*write)	(struct flash_device *fd, int pageno, char *buf, AREAB *b);
	int	(*writeb) 	(struct flash_device *fd, int pageno, AREAB *b);
	int	(*erase)	(struct flash_device *fd, int blockno, int verify);
	int	(*alloc)	(struct flash_device *fd);
	void	(*free)		(struct flash_device *fd, int pageno);
	void	(*delete)	(struct flash_device *fd, int pageno);
} FLASHDEV;

extern	struct	flash_device	flashs[];

void	flash_imap_expand(struct flash_imap *ip, int sectno);
int	flash_init(FLASHDEV *fd, int blocks, int pages, int blocksize, int pagesize);
u_short flash_checksum(u_short *buf, int size);
int 	flash_read(struct ffs_buf *bp);
int 	flash_write(struct ffs_buf *bp);
int 	flash_alloc(struct ffs_buf *bp);

void 	flash_free(FLASHDEV *fd, int pageno);
void 	flash_delete(FLASHDEV *fd, int pageno);
int	flash_ialloc(FLASHDEV *fd, int isdir);
void	flash_irelse(FLASHDEV *fd, int ino);

void	selectROMCS(int which);

struct flash_imap * flash_getimap(FLASHDEV *fd, int ino);

int	flash_mount(kdev_t dev);
FLASHDEV *flash_device(kdev_t dev);
void	flash_close(FLASHDEV *fd);

#ifdef	__KERNEL__
#define	printf			printk
#define	OUTB(port, val)		outb(val, port)
#define	INB(port)		inb(port)
#define	OUTW(port, val)		outw(val, port)
#define	INW(port)		inw(port)
#define	OUTL(port, val)		outl(val, port)
#define	INL(port)		inl(port)
#else
#define	vremap(addr, size)	addr
#define	OUTB(port, val)		outb(port, val)
#define	INB(port)		inb(port)
#endif

#endif
