#ifndef _LINUX_FLASH_FS_H
#define _LINUX_FLASH_FS_H

#ifdef	__KERNEL__
#include <linux/debug.h>

#define	i_ffs_ip	u.ffs_i.ip
#define	i_ffs_fd	u.ffs_i.fd
#endif
	
#define	FFS_SECTSIZE	4096
#define	FFS_NAMELEN	127
#define	FFS_DIRSIZE	256

typedef	struct 	ffs_dirent0 {		/* version 0 format */
	u_char		entsize;
	u_char		noff;
	u_short		mode;
	u_short		ino;
	u_short		rdev;		/* for BLK or CHR node */
	int		size;
	char		name[0];
} DIRENT0;

typedef	struct 	ffs_dirent {
	u_char		entsize;
	u_char		noff;
	u_short		mode;
	u_short		ino;
	u_short		rdev;		/* for BLK or CHR node */
	int		size;
	u_short		uid;
	u_short		gid;
	char		name[1];
} DIRENT;

#define	DE_VERSION(de)	(de->noff == sizeof(DIRENT0) ? 0 : 1) 
#define	DE_NAME(de)	(((char *) de) + de->noff)

struct	ffs_buf {
	char	buffer[FFS_SECTSIZE];
	char	*buf;
	struct	flash_device	*fd;
	int	access;
	u_short	ino;
	u_short	sectno;

	u_short	pageno;
	u_char	dirty;
	u_char	sequence;
};

#define	SECTNO(off)	((off) / FFS_SECTSIZE)
#define	SECTOFF(off)	((off) & (FFS_SECTSIZE - 1))

#define	FFS_BLKSIZE(inode)	FFS_SECTSIZE

/*
 * Function prototypes
 */
#ifdef	__KERNEL__
extern int ffs_lookup(struct inode * dir,const char * name, int len,
	               struct inode ** result);
#else
struct inode *ffs_lookup(struct inode *root_inode, char *path, struct inode **parent, char **name);
#endif
extern int init_flash_fs(void);
extern int ffs_create(struct inode * dir,const char * name, int len, int mode,
	struct inode ** result);
extern int ffs_mkdir(struct inode * dir, const char * name, int len, int mode);
extern int ffs_rmdir(struct inode * dir, const char * name, int len);
extern int ffs_unlink(struct inode * dir, const char * name, int len);
extern int ffs_symlink(struct inode * inode, const char * name, int len,
	const char * symname);
extern int ffs_link(struct inode * oldinode, struct inode * dir, const char * name, int len);
extern int ffs_mknod(struct inode * dir, const char * name, int len, int mode, int rdev);
extern int ffs_rename(struct inode * old_dir, const char * old_name, int old_len,
	struct inode * new_dir, const char * new_name, int new_len, int must_be_dir);
extern struct inode * ffs_new_inode(struct inode * dir, const char *, mode_t mode, kdev_t rdev);
extern int ffs_free_inode(struct inode * inode);

extern int ffs_file_read(struct inode *, struct file *, char *, int);
extern int ffs_file_write(struct inode *, struct file *, const char *, int);

extern void ffs_truncate(struct inode *);

#ifdef	__KERNEL__
extern void ffs_put_super(struct super_block *);
extern struct super_block *ffs_read_super(struct super_block *,void *,int);
extern int init_ffs_fs(void);
extern void ffs_write_super(struct super_block *);
extern int ffs_notify_change(struct inode *, struct iattr *);
extern void ffs_statfs(struct super_block *, struct statfs *, int);
extern int ffs_mmap(struct inode *, struct file *, struct vm_area_struct *);
extern void ffs_set_iops(struct inode *inode);
#endif
extern void ffs_read_inode(struct inode *);
extern void ffs_write_inode(struct inode *);
extern void ffs_read_inode(struct inode *);
extern void ffs_put_inode(struct inode *);
extern int ffs_sync_file(struct inode *, struct file *);

extern struct ffs_dirent *ffs_get_dirent(struct inode *inode, int dirty); 
extern struct ffs_dirent *ffs_find_dirent(struct inode *dir, const char *name, int len, int dirty); 
extern void ffs_update_dirent(struct inode *inode);

extern struct ffs_buf *ffs_getblk(struct inode *inode, int offset, int alloc);
extern struct ffs_buf *ffs_balloc(struct inode *inode, int sectno);
extern struct ffs_buf *ffs_bremap(struct inode *inode, int sectno, int pageno);
extern struct ffs_buf *ffs_bread(struct inode *inode, int sectno, int pageno);
extern struct ffs_buf *ffs_bpage(struct inode *inode, int sectno, int pageno);
extern void   ffs_bwrite(struct ffs_buf *bp);
extern void   ffs_bdelete(struct inode *inode, int sectno, int pageno);
extern void   ffs_bsync(struct inode *inode);
extern void   ffs_setname(ino_t ino, char *name, int len);
extern int    ffs_rename_dirent(struct inode *old_dir, ino_t ino, struct inode *new_dir, const char *new_name, int new_len);
extern struct ffs_dirent *ffs_empty_dirent(struct inode *dir, int entsize);

extern struct inode_operations	ffs_dir_iops, ffs_file_iops, ffs_symlink_iops;

#endif
