#include "font.h"
#include "colors.h"
#ifdef	__KERNEL__
#include <asm/io.h>
#define	OUTB(port, val)	outb(val, port)
#define	INB(port)	inb(port)
#else
#define	OUTB	outb
#define	INB	inb
#endif

extern unsigned char *vbase;
extern const Font fa16Font;

#define SEQ_I   0x3C4		/* Sequencer Index */
#define SEQ_D   0x3C5		/* Sequencer Data Register */

#define GRA_I   0x3CE		/* Graphics Controller Index */
#define GRA_D   0x3CF		/* Graphics Controller Data Register */

#ifdef	__KERNEL__
int	video_width = 640;
int	video_height = 480;
#endif
int	fg_color = BLACK;
int	bg_color = WHITE;

#define SCRBYTE(x, y)   (vbase + ((y) * VIDEO_WIDTH + (x) >> COLOR_SHIFT))

#ifdef	MARS_MODE5

void 
outSR(int index, int val)
{
    	OUTB(SEQ_I, index);
    	OUTB(SEQ_D, val);
}

int 
inSR(int index)
{
    	OUTB(SEQ_I, index);
    	return(INB(SEQ_D));
}

void 
outGR(int index, int val)
{
    	OUTB(GRA_I, index);
    	OUTB(GRA_D, val);
}

int 
inGR(int index)
{
    	OUTB(GRA_I, index);
    	return(INB(GRA_D));
}
#endif

wmode5()
{
#ifdef	MARS_MODE5
    	outGR(0x0b, 0x04 | 0x02);		/* Extend write, BY8 addr */
	outGR(0x05, 0x45); 			/* extend write mode 5 */
#endif
}

wmode0()
{
#ifdef	MARS_MODE5
	outGR(0x0b, 0x04);
	outGR(0x05, 0x40);
#endif
}

set_bg_fg(bg, fg)
{
	bg_color = bg;
	fg_color = fg;
#ifdef	MARS_MODE5
	outGR(0x00, bg);
	outGR(0x01, fg);
#endif
}

text_init()
{
	wmode5();
	set_bg_fg(WHITE, BLACK);
}

void
text_out(int v, int h, char *buf)
{
    	unsigned char c;
    	char *bits;
    	unsigned char *vaddr;
    	int 	left, top, j;

    	for ( ; (c = *buf++) != 0; ++h) {
	    	if (c < fa16Font.fnt_minchar || c > fa16Font.fnt_maxchar)
			c = ' ';

	    	c -= fa16Font.fnt_minchar;
	    	bits = fa16Font.fnt_bits + c * fa16Font.fnt_chrsize;
	    	left = h * FONT_WIDTH;
	    	top  = v * FONT_HEIGHT;

	    	vaddr = SCRBYTE(left, top);

	    	for (j = FONT_HEIGHT; j > 0; --j) {
#ifdef	MARS_MODE5
			*vaddr = *bits++;
			vaddr += (VIDEO_WIDTH >> COLOR_SHIFT);
#else
			int	b, w, s, ss;

			b = *bits++;
			ss = 0x80;
			for (w = FONT_WIDTH, s = ss; w > 0; --w, s >>= 1)
				*vaddr++ = (s & b) ? fg_color : bg_color;
			vaddr += VIDEO_WIDTH - FONT_WIDTH;
#endif
	    	}
	}
}

void
text_out2(int v, int h, int bg, int fg, char *buf)
{
	set_bg_fg(bg, fg);
	text_out(v, h, buf);
	set_bg_fg(WHITE, BLACK);
}

int	current_line = 10;
int	current_col  = 0;
int	cursor_state = 0;

void
text_putchar(char c)
{
	static	char	buf[2];

	if (cursor_state) {
		cursor_state = 0;
		if (c == '\n' || c == 8) {
			buf[0] = ' ';
			text_out(current_line, current_col, buf);
		}
	}

	switch(c) {
	
	case 8:		/* backspace */
		if (current_col > 0)
			current_col--;
		break;
	case '\r':
		current_col = 0;
		break;
	default:
		buf[0] = c;
		text_out(current_line, current_col, buf);
		if (++current_col < LINE_WIDTH)
			break;
	case '\n':
		current_col = 0;
		if (current_line < LINE_HEIGHT - 1)
			current_line++;
		else {
#ifdef	MARS_MODE5
			wmode0();
			memcpy(vbase, vbase + LINE_SIZE,
				VIDEO_WIDTH * VIDEO_HEIGHT - LINE_SIZE);
			wmode5();
			memset(vbase + (LINE_SIZE * (LINE_HEIGHT - 1) >> COLOR_SHIFT), 
				0, LINE_SIZE >> COLOR_SHIFT);
#else
			memcpy(vbase, vbase + LINE_SIZE,
				VIDEO_WIDTH * VIDEO_HEIGHT - LINE_SIZE);
			memset(vbase + LINE_SIZE * (LINE_HEIGHT - 1), 
				bg_color, LINE_SIZE);
#endif
		}
		break;
	}

	cursor_state = 1;
	buf[0] = '_';
	text_out(current_line, current_col, buf);
}

#ifdef	EPROM
int
text_getchar(void)
{
	int	c;

	if ((c = kbdgetchar()) < 0)
		return(c);
	return(keyboardRint(c));
}
#endif
